//---------------------------------------------------------------------------

#include <clx.h>
#pragma hdrstop

#include "Main.h"
#include "About.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.xfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------

const int
  RBTOP = 58,
  RBMIDDLE = 74,
  RBBOTTOM = 90,
  RBLEFT = 36,
  TACKLETOP = 59,
  TACKLEMIDDLE = 75,
  TACKLEBOTTOM = 91,
  TACKLEROW1 = 104,
  TACKLEROW2 = 148,
  TACKLEROW3 = 236;


void __fastcall TMainForm::ResetGame()
{
  YardsToGo = 10;
  Down = 1;
  FieldPosition = 20;
  Home = 0;
  Visitor = 0;
  Quarter = 1;
  TimeLeft = 15;
  TimeDisplay->Caption = FloatToStrF(TimeLeft, ffGeneral, 4, 4);
  HomeDisplay->Caption = IntToStr(Home);
  VisitorDisplay->Caption = IntToStr(Visitor);
  Randomize;
  ResetField();
};


void __fastcall TMainForm::ResetField()
{
  int x,y;

  TogglePlay(false);
  LastSack = 0;

  // empty the field
  for (y = 0; y < 3; y++)
    for (x = 0; x < 10; x++)
    {
        field[x][y].player = 0;
        field[x][y].rb = false;
    };

  // initial locations of players
  field[0][1].player = Runningback;
  field[0][1].rb = true;
  rbx = 0;
  rby = 1;
  field[3][0].player = Tackler1;
  field[3][1].player = Tackler2;
  field[3][2].player = Tackler3;
  field[5][1].player = Tackler4;
  field[9][1].player = Tackler5;

  Runningback->Left = RBLEFT;
  Runningback->Top = RBMIDDLE;
  Tackler1->Left = TACKLEROW1;
  Tackler1->Top = TACKLETOP;
  Tackler2->Left = TACKLEROW1;
  Tackler2->Top = TACKLEMIDDLE;
  Tackler3->Left = TACKLEROW1;
  Tackler3->Top = TACKLEBOTTOM;
  Tackler4->Left = TACKLEROW2;
  Tackler4->Top = TACKLEMIDDLE;
  Tackler5->Left = TACKLEROW3;
  Tackler5->Top = TACKLEMIDDLE;

  ShowField(true);
};

void __fastcall TMainForm::Sacked( TLabel *Player )
{
  running = false;
  LastSack = Player;
  if (YardsToGo <=0)
  {
    Down = 1;
    YardsToGo = 10;
  }
  else
  {
     Down++ ;
    if (Down > 4)
    {
      Down = 1;  // First down for visitor
      YardsToGo = 10;
      Computer->Visible = true;
    };
  };
};

void __fastcall TMainForm::ShowDisplay(bool visible)
{
  DownLabel->Visible = visible;
  FieldPosLabel->Visible = visible;
  YTGLabel->Visible = visible;
  DownDisplay->Visible = visible;
  FieldPosDisplay->Visible = visible;
  YTGDisplay->Visible = visible;
};

void __fastcall TMainForm::ShowField(bool visible)
{
  Runningback->Visible = visible;
  Tackler1->Visible = visible;
  Tackler2->Visible = visible;
  Tackler3->Visible = visible;
  Tackler4->Visible = visible;
  Tackler5->Visible = visible;
};

void __fastcall TMainForm::TogglePlay(bool toggle)
{
  running = toggle;
  Timer->Enabled = toggle;
  Clock->Enabled = toggle;
};


__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Exit1Click(TObject *Sender)
{
  Application->Terminate();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::About1Click(TObject *Sender)
{
  AboutForm->ShowModal();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormKeyPress(TObject *Sender, char &Key)
{
  switch (Key){
    case  'i':        MoveUpClick(Sender); break;
    case  'k':        MoveDownClick(Sender); break;
    case  'l':        MoveForwardClick(Sender); break;
    case  'j':        MoveBackClick(Sender); break;

  };
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MoveUpClick(TObject *Sender)
{
  if (OnOffSwitch->Position == 2 || LastSack != 0)
    return;

  if (! running)
    TogglePlay(true);

  if (Runningback->Top == RBTOP)
    return; // already as up as it gets

  if (field[rbx][rby-1].player != 0)
  {
    Sacked(field[rbx][rby-1].player);
    return;
  };

  if (Runningback->Top == RBBOTTOM)
    Runningback->Top = RBMIDDLE;
  else if (Runningback->Top == RBMIDDLE)
    Runningback->Top = RBTOP;

  field[rbx][rby].player = 0;
  field[rbx][rby].rb = false;
  rby-- ;
  field[rbx][rby].player = Runningback;
  field[rbx][rby].rb = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MoveBackClick(TObject *Sender)
{
  if (OnOffSwitch->Position == 2 || LastSack != 0)
    return;

  if (! running)
    TogglePlay(true);
  if(FieldPosition > 0)
  {
    if (Runningback->Left > RBLEFT + 2)
    {
      if (field[rbx-1][rby].player != 0)
      {
        Sacked(field[rbx-1][rby].player);
        return;
      };
      field[rbx][rby].player = 0;
      field[rbx][rby].rb = false;
      Runningback->Left = Runningback->Left - 22;
      rbx-- ;
    }
    else
    {
      if (field[9][rby].player != 0)
      {
        Sacked(field[9][rby].player);
        return;
      };
      field[rbx][rby].player = 0;
      field[rbx][rby].rb = false;
      Runningback->Left = TACKLEROW3;
      rbx = 9;
    };

    YardsToGo++ ;
    FieldPosition--;

    field[rbx][rby].player = Runningback;
    field[rbx][rby].rb = true;
  };
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MoveForwardClick(TObject *Sender)
{
  if (OnOffSwitch->Position == 2 || LastSack != 0)
    return;

  if (! running)
    TogglePlay(true);

  if (Runningback->Left < TACKLEROW3 - 2)
  {
    if (field[rbx+1][rby].player != NULL)
    {
      Sacked(field[rbx+1][rby].player);
      return;
    };
    field[rbx][rby].player = 0;
    field[rbx][rby].rb = false;
    Runningback->Left = Runningback->Left + 22;
    rbx++;
  }
  else
  {
    if (field[0][rby].player != 0)
    {
        Sacked(field[0][rby].player);
        return;
    };
    field[rbx][rby].player = 0;
    field[rbx][rby].rb = false;
    Runningback->Left = RBLEFT;
    rbx = 0;
  };

  YardsToGo--;
  FieldPosition++;
  if (FieldPosition == 100)
  {
    running = false;
    ShowField(false);
    Home+= 7;
    HomeDisplay->Caption = IntToStr(Home);
    FieldPosition=80;  // Starting position for visitor
    Down = 1;
    YardsToGo = 10;
    Computer->Visible = true;
    return;
  };
  field[rbx][rby].player = Runningback;
  field[rbx][rby].rb = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MoveDownClick(TObject *Sender)
{
    if (OnOffSwitch->Position == 2 || LastSack != 0)
        return;

    if (! running)
        TogglePlay(true);

    if (Runningback->Top == RBBOTTOM)
        return; // already as down as it gets

    if (field[rbx][rby+1].player != 0)
    {
        Sacked(field[rbx][rby+1].player);
        return;
    };

    if (Runningback->Top == RBTOP)
        Runningback->Top = RBMIDDLE;
    else if (Runningback->Top == RBMIDDLE)
        Runningback->Top = RBBOTTOM;

    field[rbx][rby].player = 0;
    field[rbx][rby].rb = false;
    rby++;
    field[rbx][rby].player = Runningback;
    field[rbx][rby].rb = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ComputerClick(TObject *Sender)
{
  ShowField(false);
  FieldPosition -= (int) (99.0*rand()/(RAND_MAX+1.0));
  if (FieldPosition <= 0)
  {
    Visitor+= 7;
    VisitorDisplay->Caption = IntToStr(Visitor);
    FieldPosition = 20;
  };

  Computer->Visible = false;
  LastSack = Runningback; // hack to keep movement keys disabled
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ClockTimer(TObject *Sender)
{
  if (! running)
    return;

  TimeLeft = TimeLeft - 0.1;
  TimeDisplay->Caption = FloatToStrF(TimeLeft, ffGeneral, 4, 4);

  if (TimeLeft <= 0)
  {
    Quarter++ ;
    TimeLeft = 15;
  };
  if (Quarter >= 5) 
  {
    // game over
    LastSack = Runningback; // hack to keep movement keys disabled
    TogglePlay(false);
    ShowField(false);
    ShowDisplay(true);
  };
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TimerTimer(TObject *Sender)
{
  int  x, y, newx, newy, direction ;


//  newy = random()*3;
  newy = (int) (2.0*rand()/(RAND_MAX+1.0));
  y = newy;
  newx = (int) (9.0*rand()/(RAND_MAX+1.0));
  x = newx;
  direction = (int) (1.0*rand()/(RAND_MAX+1.0)); // 0 is for x, 1 is for y

  if (! running)
  {
    if (LastSack != 0)
      LastSack->Visible = ! LastSack->Visible;
    return;
  };

  if (field[x][y].rb)
    return; // can't move the runningback!

  if (field[x][y].player == NULL)
    return; // no tacker at this spot

  if (field[x][y].player->Left <= RBLEFT)
    return; // already at the endzone

  if (direction == 0)  // we're moving horizontal
  {
    if (x < rbx)
      newx = x + 1;
    else if (x > rbx)
      newx = x - 1;
    else
      return;
    //  we're already horizontally lined up with rb
  }
  else if (direction == 1)  // we're moving vertical
  {
    if (y < rby)
      newy = y + 1;
    else if (y > rby)
      newy = y - 1;
    else
      return;
    //  we're already vertically lined up with rb
  };

  if (field[newx][newy].rb)  // got him!
  {
    Sacked(field[x][y].player);
    return;
  };

  if (field[newx][newy].player == 0) // not blocked
  {
    field[x][y].player->Left = field[x][y].player->Left - (22 * (x - newx));
    field[x][y].player->Top = field[x][y].player->Top - (16 * (y - newy));
    field[newx][newy].player = field[x][y].player;
    field[x][y].player = 0;
  };
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ScoreMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  if (OnOffSwitch->Position == 2 || running)
    return;

  ShowDisplay(false);
  if (!Computer->Visible) 
    ResetField();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ScoreMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  if (OnOffSwitch->Position == 2 || running)
    return;

  if (Quarter = 5) 
    return;

  YTGDisplay->Caption = IntToStr(YardsToGo);

  if (FieldPosition <= 50) 
  {
    FieldPosDisplay->Caption = IntToStr(FieldPosition) + ' <';
  }
  else
  {
    FieldPosDisplay->Caption = "> " + IntToStr(100 - FieldPosition);
  };

  DownDisplay->Caption = IntToStr(Down);

  ShowField(false);
  ShowDisplay(true);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::KickClick(TObject *Sender)
{
  if (running || Down != 4)
    return;  // button only valid just before fourth down

  ShowField(false);
  FieldPosition += (int) (99.0*rand()/(RAND_MAX+1.0));
  if (FieldPosition >= 100)
  {
    Home+= 3 ;
    HomeDisplay->Caption = IntToStr(Home);
    FieldPosition = 80;  // Starting position for visitor
  };

  Down = 1;  // first down for visitor
  YardsToGo = 10;
  Computer->Visible = true;
  LastSack = Runningback; // hack to keep movement keys disabled
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::OnOffSwitchChange(TObject *Sender)
{
  switch (OnOffSwitch->Position)
  {
    case   1:
    {
      Timer->Interval = 250;
      HomeDisplay->Visible = true;
      VisitorDisplay->Visible = true;
      TimeDisplay->Visible = true;
      ResetGame();
      break;
    };
    case 2:
    {
      ShowField(false);
      HomeDisplay->Visible = false;
      VisitorDisplay->Visible = false;
      TimeDisplay->Visible = false;
      break;
    };
    case 3:
    {
      Timer->Interval = 100;
      HomeDisplay->Visible = true;
      VisitorDisplay->Visible = true;
      TimeDisplay->Visible = true;
      ResetGame();
      break;
    };
  };
}
//---------------------------------------------------------------------------

