/****************************************************************************/
//
// maplistmgmt.h
//
// OVERVIEW
// ~~~~~~~~
// maplistmgmt.h is used to maintain a list of numbers.  This module can
// be instaniated for each list of numbers.  The conversions for the
// numbers to different number bases is included.
//
/****************************************************************************/
//---------------------------------------------------------------------------

#ifndef maplistmgmtH
#define maplistmgmtH
//---------------------------------------------------------------------------

#include <clx.h>
#include <map>


#define NUM_MGMT_NOT_IN_LIST 0
#define NUM_MGMT_UNFILTERED  1
#define NUM_MGMT_FILTERED    2

#define NUM_MGMT_MAX_LIST           2
#define NUM_MGMT_UNFILTERED_LIST    0
#define NUM_MGMT_FILTERED_LIST      1

#define NUM_MGMT_DECIMAL     0
#define NUM_MGMT_OCTAL       1
#define NUM_MGMT_HEXADECIMAL 2

typedef struct {
   char Decimal[8];
   char Octal[8];
   char Hexadecimal[8];
} NumberFormats;

using namespace std;

typedef map<int, NumberFormats, less<int> > NUM_Mgmt_List_Type;
typedef NUM_Mgmt_List_Type::value_type NUM_Mgmt_Entry;
typedef NUM_Mgmt_List_Type::iterator NUM_Mgmt_Iterator_Type;

class MapListMgmt {
   private:
      int NumberBase;
      // define iterators for stepping through the lists when other modules
      // want to copy the list
      NUM_Mgmt_Iterator_Type FilterIter, UnFilterIter;

      // conversion calls
      void DecimalToOctal(int DecimalIn, char *OctalOutput);
      void DecimalToHexadecimal(int DecimalIn, char *HexOutput);
      int OctalToDecimal(char *OctalIn);
      int HexadecimalToDecimal(char *HexadecimalIn);
      int DecimalToDecimal(char *DecimalIn);
      int ChangeToInteger(char *ThisNumber);
   public:
      MapListMgmt(int mode);
      ~MapListMgmt();

      // create the lists for Number Management
      NUM_Mgmt_List_Type NUM_Mgmt_List[NUM_MGMT_MAX_LIST];

      int FilterStatus(int ThisNumber);
      void UnfilterAll();
      void FilterAll();
      void AddToFilter(TListBox *UnfilteredList);
      void AddToUnfilter(TListBox *FilteredList);
      void ChangeMode(TListBox *ListToUpdate, int newbase,
                      int Use_List=NUM_MGMT_UNFILTERED_LIST);
      void AddNewNumber(int DecimalNumber, bool filter=false);
      int RemoveNumber(char *ThisNumber);
      void CopyList(NUM_Mgmt_List_Type ToBeCopied,
                    int Use_List=NUM_MGMT_UNFILTERED_LIST);
      int GetListSize(int Use_List=NUM_MGMT_UNFILTERED_LIST);
      bool GetListItem(int *ItemValue, int Use_List, bool reset);
};
#endif
