//---------------------------------------------------------------------------

// Kylix C++Builder maplistmgmt example

// Contributed by Neal Semich
#pragma hdrstop

#include "maplistmgmtform.h"
#include <stdexcept>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.xfm"
TListForm *ListForm;

//---------------------------------------------------------------------------
__fastcall TListForm::TListForm(TComponent* Owner)
        : TForm(Owner)
{
   NumberBase = NUM_MGMT_DECIMAL;

   // this initializes the lists
   listmanager = new MapListMgmt(NumberBase);

}

//---------------------------------------------------------------------------
void __fastcall TListForm::FilterClick(TObject *Sender)
{
   // click filter button to move selected numbers to filtered list
   listmanager->AddToFilter(UnfilteredList);
   // unfiltered list does not need updated because AddToFilter will
   // delete the entries from the TListBox which will update the display
   // update the filtered list to show the changes
   listmanager->ChangeMode(FilteredList, NumberBase, NUM_MGMT_FILTERED_LIST);
}

//---------------------------------------------------------------------------
void __fastcall TListForm::UnfilterClick(TObject *Sender)
{
   // click unfilter button to move selected numbers to unfiltered list
   listmanager->AddToUnfilter(FilteredList);
   // filtered list does not need updated because AddToUnfilter will
   // delete the entries from the TListBox which will update the display
   // update the unfiltered list to show the changes
   listmanager->ChangeMode(UnfilteredList, NumberBase, NUM_MGMT_UNFILTERED_LIST);
}

//---------------------------------------------------------------------------
void __fastcall TListForm::FilterAllClick(TObject *Sender)
{
   // filter all button clicked so move all numbers to filtered list
   listmanager->FilterAll();
   // need to update both lists because FilterAll only modified map lists
   listmanager->ChangeMode(FilteredList, NumberBase, NUM_MGMT_FILTERED_LIST);
   listmanager->ChangeMode(UnfilteredList, NumberBase, NUM_MGMT_UNFILTERED_LIST);
}

//---------------------------------------------------------------------------
void __fastcall TListForm::UnfilterAllClick(TObject *Sender)
{
   // unfilter all button clicked so move all numbers to unfiltered list
   listmanager->UnfilterAll();
   // need to update both lists because UnfilterAll only modified map lists
   listmanager->ChangeMode(FilteredList, NumberBase, NUM_MGMT_FILTERED_LIST);
   listmanager->ChangeMode(UnfilteredList, NumberBase, NUM_MGMT_UNFILTERED_LIST);
}

//---------------------------------------------------------------------------
void __fastcall TListForm::AddNumberClick(TObject *Sender)
{
   int NumberToAdd;

   try {
      NumberToAdd = StrToInt(AnsiString(InputNumber->Text));
      // add the number the user entered
      listmanager->AddNewNumber(NumberToAdd, Filtered->Checked);
   }
   catch (EConvertError &E) {
      // a valid number was not entered
      ReturnWindow->Text = "Invalid number";
   }
   
   // since AddNewNumber does not return a status we just update the list the
   // number should have been added to
   if (Filtered->Checked) {
      listmanager->ChangeMode(FilteredList, NumberBase, NUM_MGMT_FILTERED_LIST);
   }
   else {
      listmanager->ChangeMode(UnfilteredList, NumberBase,
                              NUM_MGMT_UNFILTERED_LIST);
   }
}

//---------------------------------------------------------------------------
void __fastcall TListForm::DeleteNumberClick(TObject *Sender)
{
   // delete the number the user entered
   int outcome;
   AnsiString NumberToDelete;

   NumberToDelete = AnsiString(InputNumber->Text);
   outcome = listmanager->RemoveNumber(NumberToDelete.c_str());
   // number was not in either list
   if (outcome == NUM_MGMT_NOT_IN_LIST) {
      ReturnWindow->Text = "Not In List";
   }
   else {
      // update the list that the number was deleted from
      if (outcome == NUM_MGMT_FILTERED) {
         listmanager->ChangeMode(FilteredList, NumberBase,
                                 NUM_MGMT_FILTERED_LIST);
      }
      else {
         listmanager->ChangeMode(UnfilteredList, NumberBase,
                                 NUM_MGMT_UNFILTERED_LIST);
      }
      ReturnWindow->Text = "Deleted";
   }
}

//---------------------------------------------------------------------------
void __fastcall TListForm::FindNumberClick(TObject *Sender)
{
   // find the number the user entered
   int outcome = listmanager->FilterStatus(StrToInt(AnsiString(InputNumber->Text)));
   switch (outcome) {
      case NUM_MGMT_NOT_IN_LIST  :
         ReturnWindow->Text = "Not In List";
         break;
      case NUM_MGMT_FILTERED :
         ReturnWindow->Text = "Filtered";
         break;
      case NUM_MGMT_UNFILTERED :
         ReturnWindow->Text = "Unfiltered";
         break;
   }
}


//---------------------------------------------------------------------------
void __fastcall TListForm::RecalculateClick(TObject *Sender)
{
   // change mode
   int mode = 0;
   if (Decimal->Checked)
      mode = NUM_MGMT_DECIMAL;
   if (Octal->Checked)
      mode = NUM_MGMT_OCTAL;
   if (Hexadecimal->Checked)
      mode = NUM_MGMT_HEXADECIMAL;
   NumberBase = mode;
   listmanager->ChangeMode(UnfilteredList, mode, NUM_MGMT_UNFILTERED_LIST);
   listmanager->ChangeMode(FilteredList, mode, NUM_MGMT_FILTERED_LIST);
}
//---------------------------------------------------------------------------
void __fastcall TListForm::FormCreate(TObject *Sender)
{
  int j, newnum;
  for (j=1; j < 10; j++)
  {
    newnum = (40.0*rand()/(RAND_MAX+1.0));
    UnfilteredList->Items->Add(IntToStr((int)newnum));
    listmanager->AddNewNumber(newnum, false);
  }
}
//---------------------------------------------------------------------------

