unit ClockDisplay;

{
  clock - a Kylix demo.
  This demo shows you how to do a masked form and how to use command line params.
}

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, CircularForms, QComCtrls, QExtCtrls, QTypes;
  
type
  TForm1 = class(TCircularForm)
    Timer1: TTimer; // Timer to tell us the time
    Image1: TImage; // Image to hold the face of the clock.
    procedure FormPaint(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    // Need some variables to hold the time inbformation
    FHours: Word;
    FMins: Word;
    FSecs: Word;
    FMSecs: Word;
  end;

var
  Form1: TForm1;

implementation

{$R *.xfm}

uses
  Qt;

procedure TForm1.FormPaint(Sender: TObject);
var
  R: TRect;
  Angle: Extended;
  Center: TPoint;

  // This routine draws our clock hands.
  procedure DrawHand(const Radius, BackRadius, AWidth: Integer; const AColor: TColor; Angle: Real);
  var
    X, Y: Integer;
  begin
    Canvas.Pen.Width := AWidth;
    Canvas.Pen.Color := AColor;
    Angle := (Angle + 3 * PI / 2);
    X := Center.X - Round(BackRadius * Cos(Angle));
    Y := Center.Y - Round(BackRadius * Sin(Angle));
    Canvas.MoveTo(Center.X, Center.Y);
    Canvas.LineTo(X, Y);
    X := Center.X + Round(Radius * Cos(Angle));
    Y := Center.Y + Round(Radius * Sin(Angle));
    Canvas.MoveTo(Center.X, Center.Y);
    Canvas.LineTo(X, Y);
  end;

begin
  R := BoundsRect;
  AdjustForClient(R);
  // Draw the clock face
  bitBlt(QWidget_to_QPaintDevice(Handle), 0, 0, Image1.Picture.Bitmap.Handle, 0, 0, R.Right-R.Left,
  R.Bottom-R.Top, RasterOp_CopyROP, False);

  // Calculate the center of the clock
  Center := Point(Width div 2, Height div 2);

  // Calculate the angles of the hands
  Angle := 2 * PI * (FHours + FMins / 60) / 12;
  DrawHand(Center.X * 70 div 100, 0, 2, clBlack, Angle);
  Angle := 2 * PI * FMins / 60;
  DrawHand(Center.X * 90 div 100, 0, 2, clBlack, Angle);
  Angle := 2 * PI * FSecs / 60;
  DrawHand(Center.X, 10, 1, clRed, Angle);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  DecodeTime(Now, FHours, FMins, FSecs, FMsecs);
  Invalidate;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  { Do we have at least two params? }
  if ParamCount > 2 then
  begin
    Top := StrToIntDef(ParamStr(1), Top);
    Left := StrToIntDef(ParamStr(2), Left);
  end else
    { Is the user asking for help? }
    if (ParamCount = 1) and (ParamStr(1) = 'h') then
    begin
      // Can't assume that in Windows it is compiled with {$APPTYPE CONSOLE}
      {$IFDEF LINUX}
      WriteLn('clock - A demo application for Kylix.');
      WriteLn('');
      WriteLn('Usage: clock [X Y]');
      WriteLn('where X and Y are screen cordinates.');
      Application.Terminate;
      {$ENDIF}
    end;
end;

end.


