unit ShapedForms;

interface

uses Qt, Types, Classes, QForms, QGraphics, QControls;

type
  TShapedForm = class(TForm)
  protected
    procedure DrawMask(ACanvas: TCanvas); virtual;
  public
    procedure InitWidget; override;
    function WidgetFlags: Integer; override;
  end;

implementation

function TShapedForm.WidgetFlags: Integer;
begin
  // We don't want to erase before we paint.
  Result := inherited WidgetFlags and not Integer(WidgetFlags_WRepaintNoErase);
end;

procedure TShapedForm.InitWidget;
var
  QB: QBitmapH;
  QP: QPainterH;
  VCanvas: TCanvas;
begin
  inherited InitWidget;
  // Setup our DrawMask
  QB := QBitmap_create(Width, Height, True, QPixmapOptimization_DefaultOptim);
  try
    QP := QPainter_create(QB, Handle);
    try
      VCanvas := TCanvas.Create;
      try
        VCanvas.Start(False);
        VCanvas.Handle := QP;
        DrawMask(VCanvas);
        VCanvas.Stop;
        VCanvas.ReleaseHandle;
      finally
        VCanvas.Free;
      end;
      QWidget_setMask(Handle, QB);
    finally
      QPainter_destroy(QP);
    end;
  finally
    QBitmap_destroy(QB);
  end;
end;

procedure TShapedForm.DrawMask(ACanvas: TCanvas);
var
  R: TRect;
begin
  // Initially we don't mask anything. In CircularForms we actually do a mask
  R := ClientRect;
  ACanvas.Brush.Color := clDontMask;
  ACanvas.FillRect(R);
end;

end.
