unit ConvertItUnit;

interface

uses
  Classes,
  QComCtrls,
  QControls,
  QExtCtrls,
  QForms,
  QGraphics,
  QStdCtrls,
  SysUtils;

type
  Tfrm_Main = class(TForm)
    StatusBar: TStatusBar;
    ConvFamilies: TTabControl;
    ConvTypes: TListBox;
    ConvResults: TListBox;
    Splitter: TSplitter;
    Panel1: TPanel;
    ConvValue: TEdit;
    procedure FormShow(Sender: TObject);
    procedure ConvFamiliesChange(Sender: TObject);
    procedure ConvTypesClick(Sender: TObject);
    procedure ConvValueChange(Sender: TObject);
    procedure ConvValueKeyPress(Sender: TObject; var Key: Char);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frm_Main: Tfrm_Main;

implementation

{$R *.xfm}

uses
  ConvUtils,
  EuroConv,
  StdConvs;

procedure Tfrm_Main.FormShow(Sender: TObject);
var
  LFamilies: TConvFamilyArray;
  I: Integer;
begin
  ConvFamilies.Tabs.Clear;
  GetConvFamilies(LFamilies);
  with ConvFamilies do
    for I := 0 to Length(LFamilies) - 1 do
      Tabs.Add(ConvFamilyToDescription(LFamilies[I]));
  ConvFamiliesChange(Sender);
end;
procedure Tfrm_Main.ConvFamiliesChange(Sender: TObject);
var
  LFamily: TConvFamily;
  LTypes: TConvTypeArray;
  I: Integer;
begin
  if DescriptionToConvFamily
       (ConvFamilies.Tabs.Items[ConvFamilies.TabIndex].Caption, LFamily) then
  with ConvTypes, Items do
  begin
    BeginUpdate;
    Clear;
    GetConvTypes(LFamily, LTypes);
    for I := 0 to Length(LTypes) - 1 do
      AddObject(ConvTypeToDescription(LTypes[I]), TObject(LTypes[I]));
    ItemIndex := 0;
    EndUpdate;
  end;
  ConvTypesClick(Sender);
end;

procedure Tfrm_Main.ConvTypesClick(Sender: TObject);
begin
  ConvValueChange(Sender);
end;

procedure Tfrm_Main.ConvValueChange(Sender: TObject);
var
  LValue: Double;
  LBaseType, LTestType: TConvType;
  I: Integer;
begin
  with ConvResults, Items do
  try
    BeginUpdate;
    Clear;
    try
      LValue := StrToFloatDef(ConvValue.Text, 0);
      if ConvTypes.ItemIndex <> -1 then
      begin
        LBaseType := TConvType(ConvTypes.Items.Objects[ConvTypes.ItemIndex]);
        for I := 0 to ConvTypes.Items.Count - 1 do
        begin
          LTestType := TConvType(ConvTypes.Items.Objects[I]);
          Add(Format('%n %s', [Convert(LValue, LBaseType, LTestType),
                               ConvTypeToDescription(LTestType)]));
        end;
      end
      else
        Add('No base type');
    except
      Add('Cannot parse value');
    end;
  finally
    EndUpdate;
  end;
end;

procedure Tfrm_Main.ConvValueKeyPress(Sender: TObject; var Key: Char);
const
  backspace: char = #8;
begin
  if not(Key in [DecimalSeparator, backspace, '0'..'9']) then
    Key := #0
  else
    if (Key = DecimalSeparator) and
       (Pos(DecimalSeparator, ConvValue.Text) <> 0) then
      Key := #0;
end;


end.
