unit SortTestUnit;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls;

type
  TForm1 = class(TForm)
    Button1: TButton;
    Edit1: TEdit;
    Edit2: TEdit;
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.xfm}

{
Collation Problems with DBCS routines
-------------------------------------

CLX MBCS routines rely on glibc to define collating order
under Linux. Unfortunately, some widely-used distributions of
glibc contain incorrect collating tables. One symptom of this
problem is that accented characters are not collated next to
their unaccented versions.

The following CLX routines are affected:

    AnsiCompareText
    AnsiCompareStr
    AnsiCompareFileName
    AnsiStrComp
    AnsiStrIComp
    AnsiStrLComp
    AnsiStrLIComp

CLX programmers can work around this problem by defining a Pascal
routine that overrides the glibc comparison routine. To use this
feature:

    1. Define a unit named "CustomSortUnit"

    2. In the unit interface, define a function with this
       signature:

        function CustomStrColl(S1, S2: PChar): Integer;
            exports CustomStrColl name 'strcoll';

    3. Implement CustomStrColl so that it correctly compares MBCS
       strings in S1 and S2 according to the current locale.
       CustomStrColl must return a positive value if S1 > S2, a
       negative value if S1 < S2, and 0 if S1 = S2.

    4. Add CustomSortUnit to the uses clause of any unit that
       uses the above MBCS routines.
}

uses
  CustomSortUnit;

procedure TForm1.Button1Click(Sender: TObject);
begin
  {
    Normally "AnsiCompareStr" calls "strcoll" implemented in Libc.
    The "strcoll" that "AnsiCompareStr" calls here is implemented in" CustomSortUnit".
    Open the CustomSortUnit unit and look for "strcoll"...
  }
  ShowMessage(Format('AnsiCompareStr = %d', [AnsiCompareStr(Edit1.Text, Edit2.Text)]));
end;

end.
