unit Main;

interface

uses
   SysUtils, Classes, Types, QGraphics, QControls, QForms, QDialogs, QExtCtrls;

type
  TForm1 = class(TForm)
    procedure FormPaint(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    FGameOver : Boolean;
    bmp, backbmp: TBitmap;
  public
    { Public declarations }
    backgroundImage: TImage;
    spriteImage: TImage;
    paddle: TImage;
    backgroundCanvas: TCanvas;
    workCanvas: TCanvas;
    backgroundRect, spriteRect, changeRect, paddleRect, changePaddleRect :TRect;
    x, y, xDir, yDir, paddleX, paddleY, paddleCenter, Angle : integer;

    procedure IdleLoop( Sender: TObject; var Done: Boolean );
  end;

var
  Form1: TForm1;

implementation
uses
  Qt;

{$R *.xfm}

procedure TForm1.FormPaint(Sender: TObject);
begin
  Canvas.CopyRect(backgroundRect, workCanvas, backgroundRect);
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  //Animates and moves paddle
  paddleCenter := X;
  if(paddleCenter < paddle.Picture.Width div 2) then
      paddleCenter := paddle.Picture.Width div 2;
  if(paddleCenter > ClientWidth - (paddle.Picture.Width div 2)) then
      paddleCenter := ClientWidth - (paddle.Picture.Width div 2);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  //Assign idle time function
  Application.OnIdle := IdleLoop;

  if(Application.MessageBox('Would you like to play with Earth?', 'Hello Earthling', [smbOK, smbCancel]) = smbOK) then
  begin
      x := 0;
      y := 0;
      FGameOver := false;
      Cursor := crNone;
  end
  else
      Application.Terminate;
end;

procedure TForm1.IdleLoop(Sender: TObject; var Done: Boolean);
var
  choice: Integer;
  SideDef, TopDef, PaddleDifference: integer;
begin
  //keeps loop going
  done := false;
  //slows down action
  Sleep(1);

  if Assigned(spriteImage) then
  begin
    changeRect := spriteRect;
    spriteRect.Left := x;
    spriteRect.Top := y;
    spriteRect.Right := x + spriteImage.Picture.Width;
    spriteRect.Bottom := y + spriteImage.Picture.Height;

    workCanvas.CopyRect(paddleRect, backgroundCanvas, paddleRect);

    changePaddleRect := paddleRect;
    paddleRect.Left := paddleCenter - ((paddle.Picture.Width) div 2);
    paddleX := paddleRect.Left;
    paddleRect.Top := paddleY;
    paddleRect.Right := paddleX + paddle.Picture.Width;
    paddleRect.Bottom := paddleY + paddle.Picture.Height;

    SideDef := changeRect.Left - spriteRect.Left;
    // If SideDiff < 0 the paddle is to the right
    if(SideDef < 0) then
    begin
      changeRect.Right := spriteRect.Right;
    end
    else
    begin
      changeRect.Left := spriteRect.Left;
    end;

    TopDef := changeRect.Top - spriteRect.Top;
    // If SideDiff < 0 the paddle is to the Down
    if(TopDef < 0) then
    begin
      changeRect.Bottom := spriteRect.Bottom;
    end
    else
    begin
      changeRect.Top := spriteRect.Top;
    end;

    workCanvas.CopyRect(spriteRect, backgroundCanvas, spriteRect);

    //ChangeRectCalcs
    if (y <= 0) then
    begin
      yDir := 5;
    end;
    if (y >= ClientHeight - 16) then
    begin
      FGameOver := true;
      Cursor := crArrow;
      Application.OnIdle := nil;
      choice := MessageDlg('You lost Earth', 'Try Again?', mtConfirmation,  [mbRetry, MBCANCEL], 0);
      if(choice = mrRetry) then
      begin
        x := 0;
        y := 0;
        Cursor := crNone;
        Application.OnIdle := IdleLoop;
      end
      else
        Application.Terminate;
    end;

    if ( (spriteRect.Bottom - 16) >= (paddleRect.Top) )
     and ( (spriteRect.Bottom - 16) <= (paddleRect.Top + 5) )
     and ( (spriteRect.Right) >= (paddleRect.Left) )
     and ( (spriteRect.Left) <= (paddleRect.Right) ) then
    begin
      yDir := -5;
    end;

    if (x <= 0) then
    begin
      xDir := 5;
    end;

    if(x >= ClientWidth - 16) then
    begin
      xDir := -5;
    end;

    inc (x, xDir);
    inc (y, yDir);

    PaddleDifference := changePaddleRect.Left - paddleRect.Left;
    // If PaddleDiff < 0 the paddle is to the right
    if(PaddleDifference < 0) then
    begin
      changePaddleRect.Right := paddleRect.Right;
    end
    else
    begin
      changePaddleRect.Left := paddleRect.Left;
    end;

    //Perform dirty rectangle animation on memory and Form canvas
    workCanvas.Draw(x, y, spriteImage.Picture.Icon);
    workCanvas.Draw(paddleX, paddleY, paddle.Picture.Icon);
    Canvas.CopyRect(changeRect, workCanvas, changeRect);
    Canvas.CopyRect(changePaddleRect, workCanvas, changePaddleRect);
  end;

  Application.ProcessMessages;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  backgroundImage := TImage.Create( Self );
  spriteImage := TImage.Create( Self );
  paddle := TImage.Create( Self );
  workCanvas := TControlCanvas.Create;
  backgroundCanvas := TControlCanvas.Create;

  Angle := 1;

  spriteImage.Picture.LoadFromFile('earth.ico');
  backgroundImage.Picture.LoadFromFile('androm.bmp');
  paddle.Picture.LoadFromFile('paddle.ico');

  backgroundRect.Top := 0;
  backgroundRect.Left := 0;
  backgroundRect.Right :=  ClientWidth;
  backgroundRect.Bottom :=  ClientHeight;


  spriteRect.Top := 0;
  spriteRect.Left := 0;
  spriteRect.Right := spriteImage.Picture.Width;
  spriteRect.Bottom := spriteImage.Picture.Height;

  //Set up backgroundCanvas
  backbmp := TBitmap.Create;
  backbmp.Width := ClientWidth;
  backbmp.Height := ClientHeight;
  backgroundCanvas := backbmp.Canvas;
  backgroundCanvas.StretchDraw(backgroundRect, backgroundImage.Picture.Bitmap);


  //Set up workCanvas
  bmp := TBitmap.Create;
  bmp.Width := ClientWidth;
  bmp.Height := ClientHeight;

  workCanvas := bmp.Canvas;
  workCanvas.CopyRect(backgroundRect,  backgroundCanvas, backgroundRect);
  workCanvas.Draw( 0, 0, spriteImage.Picture.Icon);
  paddleX := ClientWidth div 2;
  paddleY := ClientHeight - 50;
  workCanvas.Draw( paddleX, paddleY, paddle.Picture.Icon);

  paddleRect.Left := paddleX - paddle.Width;
  paddleRect.Right := paddleX + paddle.Width;
  paddleRect.Top	:= paddleY;
  paddleRect.Bottom := paddleY + paddle.Height;

  Canvas.CopyRect(backgroundRect, workCanvas, backgroundRect);
end;



procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FreeAndNil(bmp);
  FreeAndNil(backbmp);
  FreeAndNil(backgroundImage);
  FreeAndNil(spriteImage);
  FreeAndNil(paddle);
end;

end.
