unit GraphicForm;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QComCtrls, QButtons, QMenus, QExtCtrls, QTypes, QImgList;

type
  TDrawShape = (dsLine, dsRect, dsCircle, dsRoundRect);
  TMainForm = class(TForm)
    MainMenu1: TMainMenu;
    MainPanel: TPanel;
    PenPanel: TPanel;
    BrushPanel: TPanel;
    ScrollBox1: TScrollBox;
    Image1: TImage;
    File1: TMenuItem;
    New1: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    Edit1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    Cut1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    LineSpeedBtn: TSpeedButton;
    RectSpeedBtn: TSpeedButton;
    CircleSpeedBtn: TSpeedButton;
    RoundRectSpeedBtn: TSpeedButton;
    PenSpeedBtn: TSpeedButton;
    BrushSpeedBtn: TSpeedButton;
    SolidLineSpeedBtn: TSpeedButton;
    DashLineSpeedBtn: TSpeedButton;
    DotLineSpeedBtn: TSpeedButton;
    DashDotLineSpeedBtn: TSpeedButton;
    DashDot2LineSpeedBtn: TSpeedButton;
    ClearLineSpeedBtn: TSpeedButton;
    ColorLineSpeedBtn: TSpeedButton;
    LineSpinEdit: TSpinEdit;
    SolidRectSpeedBtn: TSpeedButton;
    ClearRectSpeedBtn: TSpeedButton;
    ColorRectSpeedBtn: TSpeedButton;
    HoriRectSpeedBtn: TSpeedButton;
    VertRectSpeedBtn: TSpeedButton;
    FDiagRectSpeedBtn: TSpeedButton;
    BDiagRectSpeedBtn: TSpeedButton;
    CrossRectSpeedBtn: TSpeedButton;
    DiagCrossRectSpeedBtn: TSpeedButton;
    ColorDialog1: TColorDialog;
    SaveDialog1: TSaveDialog;
    OpenDialog1: TOpenDialog;
    ImageList1: TImageList;
    StatusBar1: TStatusBar;
    procedure PenSpeedBtnClick(Sender: TObject);
    procedure BrushSpeedBtnClick(Sender: TObject);
    procedure LineSpinEditChanged(Sender: TObject; NewValue: Integer);
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Image1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure SaveAs1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure New1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Cut1Click(Sender: TObject);
    procedure SetDrawShape(Sender: TObject);
    procedure SetPenStyle(Sender: TObject);
    procedure SetBrushStyle(Sender: TObject);
    procedure SetColor(Sender: TObject);
  private
    { Private declarations }
  public
    MDrawing: Boolean;
    Origin,MovePt: TPoint;
    BrushColor,PenColor: TColor;
    CurrentFile: String;
    MDrawShape: TDrawShape;
    procedure DrawShape(TopLeft, BottomRight: TPoint; AMode: TPenMode);
    procedure SetOriginal;
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.xfm}

uses QClipbrd, BmpDlg;

procedure TMainForm.DrawShape(TopLeft, BottomRight: TPoint; AMode: TPenMode);
begin
  with Image1.Canvas do
  begin
    Pen.Mode := AMode;
    case MDrawShape of
      dsLine:
        begin
          MoveTo(TopLeft.X,TopLeft.Y);
          LineTo(BottomRight.X,BottomRight.Y);
        end;
      dsRect:
        Rectangle(TopLeft.X,TopLeft.Y,BottomRight.X,BottomRight.Y);
      dsCircle:
        Ellipse(TopLeft.X,TopLeft.Y,BottomRight.X,BottomRight.Y);
      dsRoundRect:
        RoundRect(TopLeft.X,TopLeft.Y,BottomRight.X,BottomRight.Y,
                Abs((TopLeft.X-BottomRight.X) div 2),Abs((TopLeft.Y-BottomRight.Y) div 2));
    end;
  StatusBar1.Panels[2].Text := 'Modified';
  end;
end;

procedure TMainForm.SetOriginal;
begin
  LineSpeedBtn.Down := True;
  LineSpeedBtn.Click;
  SolidLineSpeedBtn.Down := True;
  SolidLineSpeedBtn.Click;
  SolidRectSpeedBtn.Down := True;
  SolidRectSpeedBtn.Click;
  LineSpinEdit.Value := 1;
  Image1.Canvas.Pen.Width := 1;
  Image1.Canvas.Brush.Color := BrushColor;
  Image1.Canvas.Pen.Color := PenColor;
  StatusBar1.Panels[0].Text := EmptyStr;
  StatusBar1.Panels[1].Text := EmptyStr;
  StatusBar1.Panels[2].Text := EmptyStr;
end;

procedure TMainForm.SetDrawShape(Sender: TObject);
begin
  if Sender = LineSpeedBtn then MDrawShape := dsLine
  else if Sender = RectSpeedBtn then MDrawShape := dsRect
  else if Sender = CircleSpeedBtn then MDrawShape := dsCircle
  else if Sender  = RoundRectSpeedBtn then MDrawShape := dsRoundRect;
end;

procedure TMainForm.SetPenStyle(Sender: TObject);
begin
  with Image1.Canvas.Pen do
  begin
    if Sender = SolidLineSpeedBtn then Style := psSolid
    else if Sender = DashLineSpeedBtn then Style := psDash
    else if Sender = DotLineSpeedBtn then Style := psDot
    else if Sender = DashDotLineSpeedBtn then Style := psDashDot
    else if Sender = DashDot2LineSpeedBtn then Style := psDashDotDot
    else if Sender = ClearLineSpeedBtn then Style := psClear;
  end;
end;

procedure TMainForm.SetBrushStyle(Sender: TObject);
begin
  with Image1.Canvas.Brush do
  begin
    if Sender = SolidRectSpeedBtn then Style := bsSolid
    else if Sender = ClearRectSpeedBtn then Style := bsClear
    else if Sender = HoriRectSpeedBtn then Style := bsHorizontal
    else if Sender = VertRectSpeedBtn then Style := bsVertical
    else if Sender = FDiagRectSpeedBtn then Style := bsFDiagonal
    else if Sender = BDiagRectSpeedBtn then Style  := bsBDiagonal
    else if Sender = CrossRectSpeedBtn then Style := bsCross
    else if Sender = DiagCrossRectSpeedBtn then Style := bsDiagCross;
  end;
end;

procedure TMainForm.SetColor(Sender: TObject);
begin
  with Image1.Canvas do
  begin
    if Sender = ColorLineSpeedBtn then
    begin
      ColorDialog1.Color := Pen.Color;
      if ColorDialog1.Execute then
        Pen.Color := ColorDialog1.Color;
    end
    else if Sender = ColorRectSpeedBtn then
    begin
      ColorDialog1.Color := Brush.Color;
      if ColorDialog1.Execute then
        Brush.Color := ColorDialog1.Color;
    end;
  end;
end;

procedure TMainForm.PenSpeedBtnClick(Sender: TObject);
begin
  PenPanel.Visible := not PenPanel.Visible;
end;

procedure TMainForm.BrushSpeedBtnClick(Sender: TObject);
begin
  BrushPanel.Visible := not BrushPanel.Visible;
end;

procedure TMainForm.LineSpinEditChanged(Sender: TObject;
  NewValue: Integer);
begin
  Image1.Canvas.Pen.Width := NewValue;
end;

procedure TMainForm.Image1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MDrawing := True;
  Image1.Canvas.MoveTo(X,Y);
  Origin := Point(X,Y);
  MovePt := Origin;
  StatusBar1.Panels[0].Text := Format('Origin: (%d, %d)', [X,Y]);
end;

procedure TMainForm.Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if MDrawing then
  begin
    DrawShape(Origin,MovePt,pmNotXor);
    MovePt := Point(X,Y);
    DrawShape(Origin,MovePt,pmNotXor);
  end;
  StatusBar1.Panels[1].Text := Format('Current: (%d, %d)', [X,Y]);
end;

procedure TMainForm.Image1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if MDrawing then
  begin
    DrawShape(Origin,Point(X,Y),pmCopy);
    MDrawing := False;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  FBitmap: TBitmap;
begin
  FBitmap := nil;
  try
    FBitmap := TBitmap.Create;
    FBitmap.Width := 200;
    FBitmap.Height := 200;
    FBitmap.Canvas.DrawPoint(-1,-1);
    Image1.Picture.Graphic := FBitmap;
  finally
    FBitmap.Free;
  end;
  BrushColor := Image1.Canvas.Brush.Color;
  PenColor := Image1.Canvas.Pen.Color;
end;

procedure TMainForm.SaveAs1Click(Sender: TObject);
begin
  if SaveDialog1.Execute then
  begin
    CurrentFile := SaveDialog1.FileName;
    Save1Click(Sender);
  end;
end;

procedure TMainForm.Save1Click(Sender: TObject);
begin
  if CurrentFile <> EmptyStr then
    Image1.Picture.SaveToFile(CurrentFile)
  else
    SaveAs1Click(Sender);
  StatusBar1.Panels[2].Text := EmptyStr;
end;

procedure TMainForm.Exit1Click(Sender: TObject);
var
  FButton: TMessageButton;
begin
  if StatusBar1.Panels[2].Text <> EmptyStr then
  begin
    FButton := Application.MessageBox('Do you want to save before exiting?', 'Unsaved Data',
                [smbCancel,smbNo,smbYes],smsInformation,smbNo,smbCancel);
    if FButton = smbCancel then Exit
    else if FButton = smbYes then SaveAs1Click(Sender);
  end;
  Close;
end;

procedure TMainForm.Open1Click(Sender: TObject);
var
  FButton: TMessageButton;
begin
  if OpenDialog1.Execute then
  begin
    CurrentFile := OpenDialog1.FileName;
    if StatusBar1.Panels[2].Text <> EmptyStr then
    begin
      FButton := Application.MessageBox('Modified: Do you want to save?', 'Unsaved Data',
                [smbCancel,smbNo,smbYes],smsInformation,smbNo,smbCancel);
      if FButton = smbCancel then Exit
      else if FButton = smbYes then SaveAs1Click(Sender);
    end;
    Image1.Picture.LoadFromFile(CurrentFile);
    SetOriginal;
  end;
end;

procedure TMainForm.New1Click(Sender: TObject);
var
  FBitmap: TBitmap;
  FButton: TMessageButton;
begin
  with BmpDimForm do
  begin
    if StatusBar1.Panels[2].Text <> EmptyStr then
    begin
      FButton := Application.MessageBox('Modified: Do you want to save?', 'Unsaved Data',
                [smbCancel,smbNo,smbYes],smsInformation,smbNo,smbCancel);
      if FButton = smbCancel then Exit
      else if FButton = smbYes then SaveAs1Click(Sender);
    end;
    ActiveControl := WidthEdit;
    WidthEdit.Text := IntToStr(Image1.Picture.Graphic.Width);
    HeightEdit.Text := IntToStr(Image1.Picture.Graphic.Height);
    if ShowModal = 1 then
    begin
      FBitmap := nil;
      try
        FBitmap := TBitmap.Create;
        FBitmap.Width := StrToInt(WidthEdit.Text);
        FBitmap.Height := StrToInt(HeightEdit.Text);
        FBitmap.Canvas.DrawPoint(-1,-1);
        Image1.Picture.Graphic := FBitmap;
        CurrentFile := EmptyStr;
        SetOriginal;
      finally
        FBitmap.Free;
      end;
    end;
  end;
end;

procedure TMainForm.Paste1Click(Sender: TObject);
var
  FBitmap: TBitmap;
begin
  FBitmap := TBitmap.Create;
  try
    FBitmap.Assign(Clipboard);
    Image1.Canvas.Draw(0, 0, FBitmap);
  finally
    FBitmap.Free;
  end;
  StatusBar1.Panels[2].Text := 'Modified';
end;

procedure TMainForm.Copy1Click(Sender: TObject);
begin
  Clipboard.Assign(Image1.Picture);
end;

procedure TMainForm.Cut1Click(Sender: TObject);
var
  ARect: TRect;
begin
  Copy1Click(Sender);
  with Image1.Canvas do
  begin
    CopyMode := cmWhiteness;
    ARect := Rect(0, 0, Image1.Width, Image1.Height);
    CopyRect(ARect, Image1.Canvas, ARect);
    CopyMode := cmSrcCopy;
    StatusBar1.Panels[2].Text := 'Modified';
  end;
end;

end.
