unit ImageFileDialog;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QComCtrls, QButtons, QExtCtrls, QFileCtrls, QMenus, QTypes;

type
  TPopupEvent = (peDirectoryTreeView,peFileListView);
  TImageViewForm = class(TForm)
    FileEdit1: TFileEdit;
    FileListView1: TFileListView;
    FilterComboBox1: TFilterComboBox;
    FileHistoryComboBox1: TFileHistoryComboBox;
    Bevel1: TBevel;
    Label1: TLabel;
    Panel1: TPanel;
    Image1: TImage;
    FullViewBtn: TButton;
    StretchCBox: TCheckBox;
    Bevel2: TBevel;
    ViewAsGlyphCBox: TCheckBox;
    UpDownGBox: TGroupBox;
    BitBtn1: TBitBtn;
    DisabledGBox: TGroupBox;
    BitBtn2: TBitBtn;
    Label2: TLabel;
    UpDownSEdit: TSpinEdit;
    DirectoryTreeView1: TDirectoryTreeView;
    SpeedButton1: TSpeedButton;
    SpeedButton2: TSpeedButton;
    PopupMenu1: TPopupMenu;
    CreateLink1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FileEdit1DirectoryChange(Sender: TObject;
      const NewDir: WideString);
    procedure FilterComboBox1Change(Sender: TObject);
    procedure DirectoryTreeView1Change(Sender: TObject; Node: TTreeNode);
    procedure FormLoaded(Sender: TObject);
    procedure FileHistoryComboBox1Change(Sender: TObject);
    procedure ViewAsGlyphCBoxClick(Sender: TObject);
    procedure FullViewBtnClick(Sender: TObject);
    procedure UpDownSEditChanged(Sender: TObject; NewValue: Integer);
    procedure StretchCBoxClick(Sender: TObject);
    procedure FileListView1SelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FileEdit1ReturnPressed(Sender: TObject);
    procedure SetCompEvent(Sender: TObject);
    procedure CreateLink1Click(Sender: TObject);
    procedure FileListView1ContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
    procedure DirectoryTreeView1ContextPopup(Sender: TObject;
      MousePos: TPoint; var Handled: Boolean);
  private
    FirstRan: Boolean;
    FormCaption: String;
    FCompEvent: TObject;
    PopupEvent: TPopupEvent;
    procedure ViewAsGlyph(const FileExt: string);
    { Private declarations }
  public
    DestLink: String;
    SourLink: String;
    DestLoc: String;
    { Public declarations }
  end;

var
  ImageViewForm: TImageViewForm;

implementation

{$R *.xfm}

uses
  ViewImage, CreateLinkDialog, LibC;

procedure TImageViewForm.ViewAsGlyph(const FileExt: string);
begin
  if ViewAsGlyphCBox.Checked and (FileExt <> '.ico') then
  begin
    SpeedButton1.Glyph := Image1.Picture.Bitmap;
    SpeedButton2.Glyph := Image1.Picture.Bitmap;
    UpDownSEdit.Value := SpeedButton1.NumGlyphs;
    BitBtn1.Glyph := Image1.Picture.Bitmap;
    BitBtn2.Glyph := Image1.Picture.Bitmap;
  end
  else begin
    SpeedButton1.Glyph := nil;
    SpeedButton2.Glyph := nil;
    BitBtn1.Glyph := nil;
    BitBtn2.Glyph := nil;
  end;
  UpDownSEdit.Enabled := ViewAsGlyphCBox.Checked;
  Label2.Enabled := ViewAsGlyphCBox.Checked;
end;

procedure TImageViewForm.SetCompEvent(Sender: TObject);
begin
  FCompEvent := Sender;
end;

procedure TImageViewForm.FormCreate(Sender: TObject);
begin
  FormCaption := Caption + ' - ';
  UpDownSEdit.Min := 1;
  UpDownSEdit.Max := 4;
  Label2.Enabled := ViewAsGlyphCBox.Checked;
  UpDownSEdit.Enabled := ViewAsGlyphCBox.Checked;
  DirectoryTreeView1.Directory := GetEnv('HOME');
  // This will call the FileHistoryComboBox OnChange event
  FileHistoryComboBox1.Add(DirectoryTreeView1.Directory);
  FileListView1.Directory.Location := DirectoryTreeView1.Directory;
  FileListView1.Directory.FileMask := FilterComboBox1.Mask;
  // This will call the FileEdit OnDirectoryChange event
  FileEdit1.Directory.Location := DirectoryTreeView1.Directory;
  FileEdit1.Directory.FileMask := FilterComboBox1.Mask;
  FirstRan := False;
end;

procedure TImageViewForm.FormLoaded(Sender: TObject);
begin
  FirstRan := True;
end;

procedure TImageViewForm.FileEdit1DirectoryChange(Sender: TObject;
  const NewDir: WideString);
var
  I: Integer;
  cont: Boolean;
begin
  if not FirstRan then
  begin
    if not(FCompEvent is TFileHistoryComboBox) and not(FCompEvent is TDirectoryTreeView) then
    begin
      Cont := False;
      for I:=0 to FileHistoryComboBox1.Items.Count-1 do
        if (FileHistoryComboBox1.Files[I].SR.PathOnly+FileHistoryComboBox1.Files[I].SR.Name <> NewDir) then
          Cont := True
        else begin
          Cont := False;
          break;
        end;
      FCompEvent := FileEdit1;
      if Cont then
        // This will call the FileHistoryComboBox OnChange event
        FileHistoryComboBox1.Add(NewDir)
      else
        // This will call the FileHistoryComboBox OnChange event
        FileHistoryComboBox1.ItemIndex := FileHistoryComboBox1.Items.IndexOf(NewDir);
      // This will explicitly call the FileHistoryComboBox OnChange event
      FileHistoryComboBox1Change(Sender);
    end;
  end;
end;

procedure TImageViewForm.FileEdit1ReturnPressed(Sender: TObject);
var
  I,SlashStr: Integer;
  NewDir,FName: String;
begin
  NewDir := FileEdit1.Directory.Location;
  SetLength(NewDir, Length(NewDir)-1);
  // This will explicitly call the FileEdit OnDirectoryChange event
  FileEdit1DirectoryChange(Sender,NewDir);
  for I:=Length(FileEdit1.Text) downto 0 do
    if FileEdit1.Text[I] = '/' then
    begin
      SlashStr := I;
      Break;
    end else
      SlashStr := 0;
  FName := Copy(FileEdit1.Text,SlashStr+1,Length(FileEdit1.Text)-SlashStr);
  for I:=0 to FileListView1.Items.Count-1 do
    with FileListView1.Items do
    begin
      if Item[I].Caption = FName then
      begin
        Item[I].Selected := True;
        Item[I].MakeVisible;
        // This will explicitly call the FileListView OnSelectItem event
        FileListView1SelectItem(Sender,Item[I],True);
        Break;
      end
    end;
end;

procedure TImageViewForm.DirectoryTreeView1Change(Sender: TObject;
  Node: TTreeNode);
var
  Cont: Boolean;
  I: Integer;
begin
  Cont := False;
  for I:=0 to FileHistoryComboBox1.Items.Count-1 do
  begin
    if (FileHistoryComboBox1.Files[I].SR.PathOnly+FileHistoryComboBox1.Files[I].SR.Name <> PFileInfo(Node.Data).SR.PathOnly+PFileInfo(Node.Data).SR.Name) then
      Cont := True
    else begin
      Cont := False;
      break;
    end;
  end;
  FCompEvent := DirectoryTreeView1;
  if Cont then
    // This will call FileHistoryComboBox OnChange event
    FileHistoryComboBox1.Add(PFileInfo(Node.Data).SR.PathOnly+PFileInfo(Node.Data).SR.Name)
  else
    // This will call FileHistoryComboBox OnChange event
    FileHistoryComboBox1.ItemIndex := FileHistoryComboBox1.Items.IndexOf(PFileInfo(Node.Data).SR.PathOnly+PFileInfo(Node.Data).SR.Name);
  // This will call DirectoryTreeView OnChange event
  FileListView1.Directory.Location := PFileInfo(Node.Data).SR.PathOnly+PFileInfo(Node.Data).SR.Name;
  // This will call FileEdit OnDirectoryChange event
  FileEdit1.Directory.Location := PFileInfo(Node.Data).SR.PathOnly+PFileInfo(Node.Data).SR.Name;
end;

procedure TImageViewForm.FileHistoryComboBox1Change(Sender: TObject);
begin
  if not FirstRan then
  begin
    if (not(FCompEvent is TDirectoryTreeView)) and (not(FCompEvent is TFileEdit)) then
    begin
      DirectoryTreeView1.Directory := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
      // This will call FileEdit OnDirectoryChange event
      FCompEvent := FileHistoryComboBox1;
      FileEdit1.Directory.Location := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
      FileListView1.Directory.Location := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
    end else if (Sender is TFileEdit) then
    begin
      DirectoryTreeView1.Directory := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
      // This will call FileEdit OnDirectoryChange event
      FCompEvent := FileHistoryComboBox1;
      FileEdit1.Directory.Location := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
      FileListView1.Directory.Location := FileHistoryComboBox1.Items.Strings[FileHistoryComboBox1.ItemIndex];
      DirectoryTreeView1.Selected.Expand(False);
    end;
  end
end;

procedure TImageViewForm.FilterComboBox1Change(Sender: TObject);
begin
  FileListView1.Directory.FileMask := FilterComboBox1.Mask;
  FileEdit1.Directory.FileMask := FilterComboBox1.Mask;
end;

procedure TImageViewForm.FileListView1SelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var
  FileExt: string[4];
begin
  if FileListView1.Focused or (Sender is TFileEdit) then
  begin
    if (PFileInfo(Item.Data).SR.Attr and faDirectory = 0) then begin
      FileExt := AnsiLowerCase(ExtractFileExt(PFileInfo(Item.Data).SR.Name));
      if (FileExt = '.bmp') or (FileExt = '.ico') or (FileExt = '.jpg') or
         (FileExt = '.png') or (FileExt = '.xpm') then begin
        Image1.Picture.LoadFromFile(PFileInfo(Item.Data).SR.PathOnly+PFileInfo(Item.Data).SR.Name);
        Caption := FormCaption + PFileInfo(Item.Data).SR.Name;
        FileEdit1.Text := PFileInfo(Item.Data).SR.Name;
        if (FileExt <> '.ico') then
        begin
          Caption := Caption +
            Format(' (%d x %d)', [Image1.Picture.Width, Image1.Picture.Height]);
          ImageForm.Image1.Picture := Image1.Picture;
          ImageForm.Caption := Caption;
          if ViewAsGlyphCBox.Checked then ViewAsGlyph(FileExt);
        end else
          ViewAsGlyphCBox.Checked := False;
        if FileExt = '.ico' then
        begin
          Icon := Image1.Picture.Icon;
          ImageForm.Image1.Picture.Icon := Icon;
        end;
      end;
    end;
  end;
end;

procedure TImageViewForm.ViewAsGlyphCBoxClick(Sender: TObject);
begin
  if FileListView1.Selected <> nil then
    ViewAsGlyph(AnsiLowerCase(ExtractFileExt(PFileInfo(FileListView1.Selected.Data).SR.Name)))
  else
    VIewAsGlyph('');
end;

procedure TImageViewForm.FullViewBtnClick(Sender: TObject);
begin
  ImageForm.HorzScrollBar.Range := Image1.Picture.Width;
  ImageForm.VertScrollBar.Range := Image1.Picture.Height;
  ImageForm.Caption := Caption;
  ImageForm.Show;
  ImageForm.WindowState := wsNormal;
end;

procedure TImageViewForm.UpDownSEditChanged(Sender: TObject;
  NewValue: Integer);
begin
  SpeedButton1.NumGlyphs := NewValue;
  SpeedButton2.NumGlyphs := NewValue;
end;

procedure TImageViewForm.StretchCBoxClick(Sender: TObject);
begin
  Image1.Stretch := StretchCBox.Checked;
end;

procedure TImageViewForm.CreateLink1Click(Sender: TObject);
var
  OldLocal,NewLocal: array[0..255] of Char;
begin
  with LinkForm do
  begin
    if PopupEvent = peDirectoryTreeView then
    begin
      Label1.Caption := ' Source DirName:';
      Label2.Caption := ' Link DirName';
      SourceFileName.Text := DirectoryTreeView1.Selected.Text;
      LinkFileName.Text := DirectoryTreeView1.Selected.Text;
    end else if PopupEvent = peFileListView then
    begin
      Label1.Caption := ' Source FileName:';
      Label2.Caption := ' Link FileName:';
      SourceFileName.Text := FileListView1.Selected.Caption;
      LinkFileName.Text := FileListView1.Selected.Caption;
    end;
    LinkLocation.Directory.Location := DirectoryTreeView1.Directory;
    LinkLocation.Text := LinkLocation.Directory.Location;
    LinkDirectoryTV.Directory := DirectoryTreeView1.Directory;
    if ShowModal = 1 then
    begin
      if PopupEvent = peDirectoryTreeView then
        StrPCopy(OldLocal,FileListView1.Directory.Location)
      else if PopupEvent = peFileListView then
        StrPCopy(OldLocal,FileListView1.Directory.Location+SourLink);
      StrPCopy(NewLocal,DestLoc+DestLink);
      if FileExists(NewLocal) then
        Application.MessageBox(Char(39)+DestLoc+DestLink+Char(39)+' FileName Exist!', 'Error!')
      else begin
        symlink(OldLocal,NewLocal);
        Application.MessageBox('SymLink: '+NewLocal+' -> '+OldLocal, 'Created SymLink');
      end;
    end;
  end;
end;

procedure TImageViewForm.FileListView1ContextPopup(Sender: TObject;
  MousePos: TPoint; var Handled: Boolean);
begin
  PopupEvent := peFileListView;
  CreateLink1.Enabled := (FileListView1.Selected <> nil);
end;

procedure TImageViewForm.DirectoryTreeView1ContextPopup(Sender: TObject;
  MousePos: TPoint; var Handled: Boolean);
begin
  PopupEvent := peDirectoryTreeView;
  CreateLink1.Enabled := (DirectoryTreeView1.Selected <> nil);
end;

end.
