program Inject;

{ Utility for dropping a header at the top of a unit. This particular utility
  replaces any existing header while preserving a copyright date contained
  within the existing header. }

{$APPTYPE CONSOLE}

uses
  SysUtils, Classes;

const
  AppName = 'Borland File Injection Example 1.0';
  CopyrightString = 'Copyright 2002';
  UsageString = 'Usage: Inject -i[injection file] -f[file list] files';
  NoFileWarning = 'Warning: Unable to open file %s';
  InjectDelim = 'unit';

procedure WriteHeader;
begin
  Writeln(AppName);
  Writeln(CopyrightString);
end;

procedure ReportError;
begin
  Writeln(UsageString);
end;

function SetupFileList(const FileListName: String): TStringList;
var
  Param: String;
  I, IsFileName: Integer;
begin
  Result := TStringList.Create;
  if FileListName <> '' then Result.LoadFromFile(FileListName);
  for I := 1 to ParamCount do
  begin
    Param := ParamStr(I);
    IsFileName := Pos('-', Param);
    if IsFileName = 0 then Result.Add(Param);
  end;
end;

procedure DoInjections(const Files: TStringList; const InjectionFileName: String);
var
  InjList, FilList: TStringList;
  FileName, InjBuf, FilBuf, TempFilBuf, OutFilBuf: String;
  PreDateBuf, PostDateBuf, StoredDate: String;
  I, DatePos, DelimPos : Integer;

  function GetDate(const Buf: String; const DelimPos: Integer): String;
  var
    SearchBuf, DateBuf: String;
    DatePos: Integer;
  begin
    { In the passed-in buffer, find the first occurrance of a year, and
      save it out. If there is no year, then this is the first time that
      file had a header, and it should get 2000 appended. Only pay attention
      to text that exists *prior* to the first occurance of the delimeter}
    SearchBuf := Copy(Buf, 0, DelimPos);
    DatePos := Pos('19', SearchBuf);
    if DatePos = 0 then
      DatePos := Pos('2000', SearchBuf);
    if DatePos = 0 then
      DatePos := Pos('2001', SearchBuf);
    if DatePos <> 0 then
    begin
      DateBuf := Copy(Buf, DatePos, 4);
      Result := DateBuf;
    end else Result := '2002';
  end;

begin
  InjList := TStringList.Create;
  FilList := TStringList.Create;

  try
     InjList.LoadFromFile(InjectionFileName);
  except
     Halt(3);
  end;

  { Get buffers containing the string prior to the first instance of
    a year in the header, and a string starting with the first instance
    of the year. This allows the injection of the existing year into
    the header. }

  InjBuf := InjList.Text;
  DatePos := Pos('20', InjBuf);
  PreDateBuf := Copy(InjBuf, 0, DatePos-1);
  PostDateBuf := Copy(InjBuf, DatePos, MAXINT);

  for I := 0 to Files.Count -1 do
  begin
    FileName := Files[I];
    try
      FilList.LoadFromFile(FileName);
    except
      Writeln(Format(NoFileWarning, [FileName]));
      Continue;
    end;
    FilBuf := FilList.Text;

    { Find the first instance of the delimeter in the header. }
    DelimPos := Pos(InjectDelim, FilBuf);
    { Find out what the existing copyright on the unit is. }
    StoredDate := GetDate(FilBuf, DelimPos);

    { Build the header for injection by adding the date to the parts
      ripped out of the header earlier. }
    InjBuf := PreDateBuf+StoredDate+', '+PostDateBuf;

    { Discard anything already in the file before the delimeter, and
      then inject the header. }
    TempFilBuf := Copy(FilBuf, DelimPos, Maxint);
    OutFilBuf := InjBuf+#13#10#13#10+TempFilBuf;
    FilList.SetText(PChar(OutFilBuf));
    FilList.SaveToFile(FileName);
  end;
  FilList.Free;
  InjList.Free;
end;

var
  I, SubPos : Integer;
  Files: TStringList;
  Param, InjectionFileName, FileListName: string;
begin
  WriteHeader;
  if ParamCount > 3 then
  begin
    ReportError;
    Exit;
  end;

  for I := 1 to ParamCount do
  begin
    Param := ParamStr(I);
    SubPos := Pos('-i', LowerCase(Param));
    if SubPos = 1 then InjectionFileName := Copy(Param, 3, MAXINT);
    SubPos := Pos('-f', LowerCase(Param));
    if SubPos = 1 then FileListName := Copy(Param, 3, MAXINT);
  end;

  if InjectionFileName <> '' then
  begin
    Files := SetupFileList(FileListName);
    DoInjections(Files, InjectionFileName);
    Files.Free;
  end else
  begin
    ReportError;
    Exit;
  end;

end.
