program Previewers;
{ Borland Kylix demo - miscellaneous previewers using a regular TOpenDialog }
uses
  Types,
  SysUtils,
  Classes,
  QControls,
  QForms,
  QGraphics,
  QDialogs,
  Qt,
  Libc;


const
  TxtExtensions: array[0..12] of string = ('.txt', '.conf', '.sh', '.c', '.cpp',
    '.h', '.hpp', '.bpr', '.bpk', '.xfm', '.pas', '.dpr', '.dpk');
  TarExtensions: array[0..1] of string = ('.tar', '.gz');
  RPMExtensions: array[0..0] of string = ('.rpm');

type
  TTextPreviewThread = class(TPreviewThread)
  private
    FText: string;
  protected
    procedure ReadText; virtual;
    procedure Draw; override;
  public
    procedure Execute; override;
    property Text: string read FText write FText;
  end;

  TTextPreviewer = class(TThreadedPreviewer)
  protected
    function ThreadClass: TPreviewThreadClass; override;
  public
    function CanPreview(const Filename: string): Boolean; override;
    procedure GetPreviewSize(Filename: string; Canvas: TCanvas;
      var Size: TSize); override;
  end;

  TCmdPreviewThread = class(TTextPreviewThread)
  protected
    function GetCmd(Filename: string): string; virtual; abstract;
    procedure ReadText; override;
  end;

  TCmdPreviewer = class(TTextPreviewer)
  protected
    function CanPreview(const Filename: string): Boolean; override;
  end;

  TTarPreviewerThread = class(TCmdPreviewThread)
  protected
    function GetCmd(Filename: string): string; override;
  end;

  TTarPreviewer = class(TCmdPreviewer)
  protected
    function ThreadClass: TPreviewThreadClass; override;
  public
    function CanPreview(const Filename: string): Boolean; override;
  end;

  TRPMPreviewerThread = class(TCmdPreviewThread)
  protected
    function GetCmd(Filename: string): string; override;
  end;

  TRPMPreviewer = class(TCmdPreviewer)
  protected
    function ThreadClass: TPreviewThreadClass; override;
  public
    function CanPreview(const Filename: string): Boolean; override;
  end;

function CanPreviewFile(Filename: WideString; Exts: array of string): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to High(Exts) do
  begin
    Result := AnsiLowerCase(ExtractFileExt(Filename)) = Exts[I];
    if Result then Break;
  end;
  if Result then
    Result := Result and (Libc.access(PChar(Filename), R_OK) = 0);
end;

{ TTextPreviewThread }

procedure TTextPreviewThread.Execute;
var
  Continue: Boolean;
begin
  ReadText;
  Continue := True;
  UpdateProgress(Self, psEnding, 100, True, Rect, '', Continue);
end;

procedure TTextPreviewThread.ReadText;
var
  Stream: TFileStream;
begin
  SetLength(FText, 512);
  Stream := TFileStream.Create(Filename, fmOpenRead);
  try
    SetLength(FText, Stream.Read(FText[1], Length(FText)-1));
  finally
    Stream.Free;
  end;
end;

procedure TTextPreviewThread.Draw;
var
  FontRecall: TFontRecall;
  PenRecall: TPenRecall;
  BrushRecall: TBrushRecall;
  R: TRect;
begin
  // be sure to leave Canvas in its original state.
  FontRecall := TFontRecall.Create(Canvas.Font);
  PenRecall := TPenRecall.Create(Canvas.Pen);
  BrushRecall := TBrushRecall.Create(Canvas.Brush);
  try
    R := Rect;
    InflateRect(R, -5, -5);
    Canvas.Brush.Color := clMidLight;
    Canvas.FillRect(R);
    Canvas.Pen.Color := clLight;
    Canvas.PenPos := Point(R.Left, Rect.Bottom);
    Canvas.LineTo(R.Left, R.Top);
    Canvas.LineTo(R.Right, R.Top);
    Canvas.Pen.Color := clShadow;
    Canvas.LineTo(R.Right, R.Bottom);
    Canvas.LineTo(R.Left, R.Bottom);
    Canvas.Font.Name := 'Courier';
    Canvas.Font.Size := 9;
    Canvas.Font.Color := clDark;
    Canvas.TextRect(R, 10, 10, FText);
  finally
    FontRecall.Free;
    BrushRecall.Free;
    PenRecall.Free;
  end;
end;

{ TTextPreviewer }

function TTextPreviewer.CanPreview(const Filename: string): Boolean;
begin
  Result := CanPreviewFile(Filename, TxtExtensions);
end;

procedure TTextPreviewer.GetPreviewSize(Filename: string; Canvas: TCanvas;
  var Size: TSize);
begin
  Size.cx := Canvas.TextWidth('X') * 80;
end;

function TTextPreviewer.ThreadClass: TPreviewThreadClass;
begin
  Result := TTextPreviewThread;
end;

{ TCmdPreviewThread }

procedure TCmdPreviewThread.ReadText;
var
  Output: PIOFile;
  Tmp: PChar;
  BytesRead: Integer;
begin
  Output := popen(PChar(GetCmd(Filename) + ' 2>/dev/null'), 'r');
  if Output <> nil then
  try
    Tmp := StrAlloc(512);
    try
      BytesRead := fread(Tmp, 1, 512, Output);
      if BytesRead > 0 then
      begin
        Tmp[BytesRead] := #0;
        Text := Trim(Tmp);
        SetLength(FText, Length(Tmp));
      end;
    finally
      StrDispose(Tmp);
    end;
  finally
    pclose(Output);
  end;
end;

{ TCmdPreviewer }

function TCmdPreviewer.CanPreview(const Filename: string): Boolean;
begin
  Result := False;
end;

{ TTarPreviewerThread }

function TTarPreviewerThread.GetCmd(Filename: string): string;
var
  Swts: string;
begin
  if AnsiLowerCase(ExtractFileExt(Filename)) = '.gz' then
    Swts := 'z';
  Result := Format('tar -t%sf "%s"', [Swts, Filename]);
end;

{ TTarPreviewer }

function TTarPreviewer.CanPreview(const Filename: string): Boolean;
begin
  Result := CanPreviewFile(Filename, TarExtensions);
end;

function TTarPreviewer.ThreadClass: TPreviewThreadClass;
begin
  Result := TTarPreviewerThread;
end;

{ TRPMPreviewerThread }

function TRPMPreviewerThread.GetCmd(Filename: string): string;
begin
  Result := Format('rpm -qpi %s', [Filename]);
end;

{ TRPMPreviewer }

function TRPMPreviewer.CanPreview(const Filename: string): Boolean;
begin
  Result := CanPreviewFile(Filename, RPMExtensions);
end;

function TRPMPreviewer.ThreadClass: TPreviewThreadClass;
begin
  Result := TRPMPreviewerThread;
end;


var
  OpenDlg: TOpenDialog;
begin
  Application.Initialize;
  QDialogs.RegisterFilePreviewer(TTextPreviewer.Create);
  QDialogs.RegisterFilePreviewer(TTarPreviewer.Create);
  QDialogs.RegisterFilePreviewer(TRPMPreviewer.Create);
  OpenDlg := TOpenDialog.Create(nil);
  OpenDlg.Options := OpenDlg.Options + [ofPreview, ofAutoPreview];
  OpenDlg.Execute;
end.
