#!/bin/bash

# BEGIN STRING TABLE

SHARED_WARNING1="This application installed files which may be shared by other"
SHARED_WARNING2="programs. Removing the shared files may cause these other"
SHARED_WARNING3="applications to stop working."
REMOVE_SHARED="Do you want to remove the shared files? [N/y] "
DIFF_USER1="This application was installed by a different user. You may not"
DIFF_USER2="have the correct permissions to correctly uninstall it."
DIFF_USER3="Do you want to attempt to uninstall anyhow? [N/y] "
ABORTED="Uninstall aborted. Your system has been left unchanged."
UNINSTALL_VERIFY="Are you sure you want to uninstall this application? [N/y] "
# END STRING TABLE

auto_remove=N
uid=`id -u`

if [ -z "$SETUP_USERID" ]; then
	# Setup canceled
	SETUP_USERID=$UID
	auto_remove=Y
fi

if [ $uid -ne $SETUP_USERID ]; then
    echo $DIFF_USER1
    echo $DIFF_USER2
    echo -n $DIFF_USER3
    read response
    case $response in
	y|Y|yes|Yes|YES )
	    # do nothing....
	;;
	* )
	    echo $ABORTED
	    exit 0
	;;
    esac
else
    if [ $auto_remove != "Y" ]; then
    	echo -n "$UNINSTALL_VERIFY"
    	read response
    	case $response in
		y|Y|yes|Yes|YES )
	    		# do nothing....
		;;
		* )
	    		echo $ABORTED
	    		exit 0
		;;
    	esac
    fi
fi

function RemoveLink() {
   if [ -L "$1" ]; then
      rm -f "$1"
   fi
}

function RemoveFile() {
   if [ -e "$1" ]; then
      rm -f "$1"
   fi
}

function RemoveLinkIfNonRPM() {
   if [ -L "$1" ]; then
      if [ "$SETUP_USED_RPM" != "1" ]; then
         rm -f "$1"
      fi
   fi
}


# Borland pre-uninstall cleanup

RemoveLink $SETUP_INSTALLPATH/bin/bplhtmlide.so.6.9
RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libdcc.so.6.9


RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libborcrtl.so.1
RemoveLink $SETUP_INSTALLPATH/bin/libborstl.so
RemoveLink $SETUP_INSTALLPATH/bin/dclmlwiz.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libxmlide.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bpldesigndgm.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplbcbide.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplclxdesigner.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bpldelphiclxide.so.6.9

RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libilinkintf.so.6.9
RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libbortoken.so.6.9

RemoveLink $SETUP_INSTALLPATH/bin/bplbcbclxide.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/dcluser.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libqt.so.2
RemoveLink $SETUP_INSTALLPATH/bin/libqtintf-6.9-qt2.3.so
RemoveLink $SETUP_INSTALLPATH/bin/libborqt-6.9-qt2.3.so
RemoveLink $SETUP_INSTALLPATH/bin/libcomctl32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libcomdlg32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libgdi32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libimm32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/liblz32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libmpr.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libole32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/liboleaut32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libolecli32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/liboledlg.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libolepro32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libolesvr32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libuser32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libversion.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwine.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwineoss.drv.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwinmm.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwinspool.drv.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libx11drv.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libshell32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwininet.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libshlwapi.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libkernel32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwineps.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/bplrtl.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplvisualclx.so.6.9


RemoveLink $SETUP_INSTALLPATH/bin/bplvcl.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplvclex.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplcoreide.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bpldelphide.so.6.9
RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/bpldesignide.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplvclide.so.6.9


RemoveLink $SETUP_INSTALLPATH/bin/dclstd.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/librpcrt4.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/dclindy.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libadvapi32.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/libwine_unicode.borland.so
RemoveLink $SETUP_INSTALLPATH/bin/dclnet.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplnetclx.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplnetdataclx.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libclxthemes.so.2
RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libborunwind.so
RemoveLinkIfNonRPM $SETUP_INSTALLPATH/bin/libborunwind.so.6
RemoveLink $SETUP_INSTALLPATH/bin/bpldesignhooks.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libboredit.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libborkbd.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/bplHelpViewers.so.6.9
RemoveLink $SETUP_INSTALLPATH/bin/libdbk.so.1


# remove hyperhelp script file
RemoveFile $SETUP_INSTALLPATH/help/hyperhelp.sh

# remove hyperhelp symlinks
RemoveLink $SETUP_INSTALLPATH/help/app-defaults/ja_JP.eucjp
RemoveLink $SETUP_INSTALLPATH/help/lib/locale/ja_JP.eucjp

# remove start IDE script files
RemoveFile $SETUP_INSTALLPATH/bin/startdelphi
RemoveFile $SETUP_INSTALLPATH/bin/startkylix
RemoveFile $SETUP_INSTALLPATH/bin/startbcb
RemoveFile $SETUP_INSTALLPATH/bin/registerkylix


# remove kylixpath
RemoveFile $SETUP_INSTALLPATH/bin/kylixpath

# remove libdl.so.2 if we installed it
RemoveFile $SETUP_INSTALLPATH/bin/libdl.so.2

# remove Debian 2.2 hack if we installed it
RemoveFile $SETUP_INSTALLPATH/bin/libstdc++-libc6.1-1.so.2

# remove .borlandrc file
RemoveFile ~/.borland/.borlandrc

# remove bin/libborcrtl.so
RemoveFile $SETUP_INSTALLPATH/bin/libborcrtl.so

# remove reg. info
RemoveFile ~/.borland/registry.bin

# remove .borland/.delphi69rc file
RemoveFile ~/.borland/delphi69rc
RemoveFile ~/.borland/bcb69rc
RemoveFile ~/.borland/debug69rc

if [ -e /usr/local/etc/delphi69rc.conf -a $uid -eq 0 ]; then
	rm -f /usr/local/etc/delphi69rc.conf
	rm -f /usr/local/etc/bcb69rc.conf
	rm -f /usr/local/etc/debug69rc.conf
fi	

# remove delphi69dro and bcb69dro files
if [ -e /usr/local/etc/delphi69dro.conf -a $uid -eq 0 ]; then
	rm -f /usr/local/etc/delphi69dro.conf
fi
if [ -e /usr/local/etc/bcb69dro.conf -a $uid -eq 0 ]; then
	rm -f /usr/local/etc/bcb69dro.conf
fi

RemoveFile ~/.borland/delphi69dro
RemoveFile ~/.borland/bcb69dro


# check to see if shared files were installed
if [ -e $SETUP_INSTALLPATH/uninstall ]; then
	search=`grep "connections" $SETUP_INSTALLPATH/uninstall`
	rpmsearch=`grep "dbexpress" $SETUP_INSTALLPATH/uninstall`
fi

if [ -n "$search" -o -n "$rpmsearch" -o $auto_remove = "Y" ]; then
if [ $auto_remove != "Y" ]; then
    cat <<EOF

$SHARED_WARNING1
$SHARED_WARNING2
$SHARED_WARNING3

EOF
fi
    if [ $auto_remove != "Y" ]; then
    	echo -n "$REMOVE_SHARED"
    	read response
    else
    	response="Y"
    fi

    case $response in
	y|Y|yes|Yes|YES )
	    # remove shared files
	    if [ -e ~/.borland/dbxconnections ]; then
		mv ~/.borland/dbxconnections $SETUP_INSTALLPATH/bin/
	    fi
	    if [ -e ~/.borland/dbxdrivers ]; then
		mv ~/.borland/dbxdrivers $SETUP_INSTALLPATH/bin/
	    fi
	    RemoveFile ~/.borland/cachedmetrics.:0.0

	    mach_name=`echo $HOSTNAME |awk -F. '{printf $1}'`
	    if [ -e ~/.borland/wineserver-$mach_name$DISPLAY ]; then
		rm -r ~/.borland/wineserver-$mach_name$DISPLAY 
	    fi
	    if [ -e ~/.borland/delphi69dci ]; then
		mv ~/.borland/delphi69dci $SETUP_INSTALLPATH/bin/delphi.dci
	    fi
	    if [ -e ~/.borland/delphi69upg ]; then
	    	mv ~/.borland/delphi69upg $SETUP_INSTALLPATH/bin/delphi69upg
	    fi
	    if [ -e ~/.borland/bcb69dci ]; then
		mv ~/.borland/bcb69dci $SETUP_INSTALLPATH/bin/bcb.dci
	    fi
	    if [ -e ~/.borland/delphi69dmt ]; then
		mv ~/.borland/delphi69dmt $SETUP_INSTALLPATH/bin/
	    fi
	    if [ -e ~/.borland/bcb69dmt ]; then
		mv ~/.borland/bcb69dmt $SETUP_INSTALLPATH/bin/
	    fi
	    if [ -e ~/.borland/incfiles.dat ]; then
	    	mv ~/.borland/incfiles.dat $SETUP_INSTALLPATH/bin/
	    fi
	    if [ -e /usr/local/etc/dbxconnections.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/dbxconnections.conf
	    fi
	    if [ -e /usr/local/etc/dbxdrivers.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/dbxdrivers.conf
	    fi
	    if [ -e /usr/local/etc/delphi69dci.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/delphi69dci.conf
	    fi
	    if [ -e /usr/local/etc/bcb69dci.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/bcb69dci.conf
	    fi
	    if [ -e /usr/local/etc/delphi69upg.conf -a $uid -eq 0 ]; then
	    	rm -f /usr/local/etc/delphi69upg.conf
	    fi
	    if [ -e /usr/local/etc/delphi69dmt.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/delphi69dmt.conf
	    fi
	    if [ -e /usr/local/etc/bcb69dmt.conf -a $uid -eq 0 ]; then
		rm -f /usr/local/etc/bcb69dmt.conf
	    fi
	    if [ -e /usr/local/etc/incfilesdat.conf -a $uid -eq 0 ]; then
	    	rm -f /usr/local/etc/incfilesdat.conf
	    fi

	    if [ -e /usr/local/etc/borlandrc.conf -a $uid -eq 0 ]; then
	    	rm -f /usr/local/etc/borlandrc.conf
	    fi
	
	    # remove dcc.cfg file
	    if [ "$uid" = "0" ]; then
	    	RemoveFile /usr/local/etc/dcc.conf
	    else
	    	RemoveFile ~/.dcc
	    fi

	    # remove bcc.cfg file
	    if [ "$uid" = "0" ]; then
	    	RemoveFile /usr/local/etc/bccrc
	    	RemoveFile /usr/local/etc/ilinkrc
	    else
	    	RemoveFile ~/.borland/bccrc
	    	RemoveFile ~/.borland/ilinkrc
	    fi
		
	    ;;
	* )
	    # put dummy files in the original installed location
	    # to eliminate error messages during uninstall
	    if [ -n "$search" -o -n "$rpmsearch" ]; then
		touch $SETUP_INSTALLPATH/bin/dbxconnections
		if [ -e ~/.borland/dbxconnect_ora ]; then
		    touch $SETUP_INSTALLPATH/bin/dbxconnect_ora
		fi
		if [ -e ~/.borland/dbxconnect_db2 ]; then
		    touch $SETUP_INSTALLPATH/bin/dbxconnect_db2
		fi
	    fi
	    search=`grep "drivers" $SETUP_INSTALLPATH/uninstall`
	    if [ -n "$search" -o -n "$rpmsearch" ]; then
		touch $SETUP_INSTALLPATH/bin/dbxdrivers
		if [ -e ~/.borland/dbxdrivers_ora ]; then
		    touch $SETUP_INSTALLPATH/bin/dbxdrivers_ora
		fi
		if [ -e ~/.borland/dbxdrivers_db2 ]; then
		    touch $SETUP_INSTALLPATH/bin/dbxdrivers_db2
		fi
	    fi
	    touch $SETUP_INSTALLPATH/bin/delphi.dci
	    touch $SETUP_INSTALLPATH/bin/delphi69upg
	    touch $SETUP_INSTALLPATH/bin/delphi69dmt
	    touch $SETUP_INSTALLPATH/bin/bcb69dmt
	    touch $SETUP_INSTALLPATH/bin/incfiles.dat
	    ;;
    esac
fi
