{ *************************************************************************** }
{                                                                             }
{ Delphi and Kylix Cross-Platform Visual Component Library                    }
{                                                                             }
{ Copyright (c) 2001, 2002 Borland Software Corporation                       }
{                                                                             }
{ IMPORTANT: The Kylix Open Edition(tm) compiler does not support             }
{ modification of this file in any manner.                                    }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Software Foundation and appearing at http://www.borland.com/kylix/gpl.html. }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }

unit QOpenBanner;

{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J-,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}
{$APPTYPE GUI}

interface

uses
  SysUtils, Types, Classes, Variants, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QTypes, QExtCtrls;

type
  TOpenBanner = class(TForm)
    MainPanel: TPanel;
    Timer: TTimer;
    TitlePanel: TPanel;
    TitleImage: TImage;
    IconPanel: TPanel;
    IconImage: TImage;
    LicensePanel: TPanel;
    LicenseLabel: TLabel;
    TopTimer: TTimer;
    procedure TimerTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure IconImageClick(Sender: TObject);
    procedure TopTimerTimer(Sender: TObject);
  private
    function PanelSize(Panel: TPanel): TSize;
    procedure SetLicenseText(const Value: WideString);
  protected
    function WidgetFlags: Integer; override;
    procedure Init;
  public
    constructor Create(AOwner: TComponent); override;
  end;

procedure DisplayLicense;

implementation

uses Math, Qt, Xlib, Libc;

procedure DisplayLicense;
var
  I: Integer;
begin
  for I := 1 to ParamCount do
    if CompareText(ParamStr(I), '-ns') = 0 then
      Exit;
  TOpenBanner.Create(Application).ShowModal;
end;

type
  TLicenseLocale = (llDefault, llDe, llFr, llJp);

const
  LicenseMessages: array [TLicenseLocale] of string = (
    { Default }
      'This application was built with Borland Kylix Open Edition(tm).'
    + 'The application must be distributed under the terms of the GNU General '
    + 'Public License (GPL), version 2. A copy of this license can be found at:'#10
    + 'http://www.borland.com/kylix/gpl.html.',

    { De }
      'Diese Anwendung wurde mit Borland Kylix '+ 
      'Open Edition(tm) entwickelt. Die Anwendung '+ 
      'mu unter den Bedingungen der GNU General '+ 
      'Public License (GPL) Version 2 vertrieben '+ 
      'werden. Diese Lizenzbedingungen finden Sie '+
      'unter:'#10 + 
      'http://www.borland.com/kylix/gpl.html.',

    { Fr }
      'Cette application a t conue avec Borland Kylix Open Edition(tm). '
    + 'Elle doit tre distribue sous les termes de la licence GPL (General '
    + 'Public License) GNU, version 2. Une copie de cette licence est '
    + 'disponible  l''adresse :'#10
    + 'http://www.borland.com/kylix/gpl.html.',

    { Ja }
      'ΥץꥱϡBorland Kylix Open Edition(tm) Ǻޤ '
    + 'ΥץꥱϡGNU General Public LicenseGPL Version '
    + ' 2 ξΤȤۤޤ'#10
    + 'Υ饤󥹤 http://www.borland.co.jp/kylix/gpl.html Ǥ '
    + 'ޤ'
  );

function LicenseLocale: TLicenseLocale;
var
  P: PChar;
begin
  P := getenv('LC_ALL');
  if (P = nil) or (P[0] = #0) then
    P := getenv('LANG');
  if (P = nil) or (P[0] = #0) then
    Result := llDefault
  else
  begin
    if Pos('de', P) = 1 then
      Result := llDe
    else if Pos('fr', P) = 1 then
      Result := llFr
    else if Pos('ja', P) = 1 then
      Result := llJp
    else
      Result := llDefault;
  end;
end;

procedure TOpenBanner.TimerTimer(Sender: TObject);
begin
  Close;
end;

procedure TOpenBanner.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TOpenBanner.SetLicenseText(const Value: WideString);
var
  FM: QFontMetricsH;
  Str: WideString;
  R: TRect;
  NewHeight,
  Extra: Integer;
begin
  FM := QFontMetrics_create(LicenseLabel.Font.Handle);
  try
    Str := Value;
    LicenseLabel.Caption := Str;
    QFontMetrics_boundingRect(FM, @R, 0, 0, LicenseLabel.Width, MaxInt,
    QLabel_alignment(LicenseLabel.Handle) or Integer(AlignmentFlags_WordBreak),
      @Str, -1, 0, nil, nil);
    NewHeight := R.Bottom - R.Top + (LicensePanel.BorderWidth) * 2;
    Extra := QFrame_frameWidth(LicenseLabel.Handle);
    if Extra > 0 then
    begin
      Inc(NewHeight, Extra * 2);
      QFrame_contentsRect(LicenseLabel.Handle, @R);
      Inc(NewHeight, R.Top * 2);
    end;
    Height := Max(NewHeight + PanelSize(TitlePanel).cy,
      PanelSize(IconPanel).cy + PanelSize(TitlePanel).cy)
      + MainPanel.BevelWidth * 2;
  finally
    QFontMetrics_destroy(FM);
  end;
end;

function TOpenBanner.WidgetFlags: Integer;
begin
  Result := inherited WidgetFlags or Integer(WidgetFlags_WStyle_NoBorderEx);
end;

function TOpenBanner.PanelSize(Panel: TPanel): TSize;
var
  Graphic: TGraphic;
begin
  Result.cx := 0;
  Result.cy := 0;
  if Panel = IconPanel then
    Graphic := IconImage.Picture.Graphic
  else if Panel = TitlePanel then
    Graphic := TitleImage.Picture.Graphic
  else
    Graphic := nil;
  if Graphic = nil then Exit;
  Result.cx := Graphic.Width + Panel.BorderWidth * 2;
  Result.cy := Graphic.Height + Panel.BorderWidth * 2;
end;

procedure TOpenBanner.IconImageClick(Sender: TObject);
begin
  TimerTimer(nil);
end;

procedure TOpenBanner.TopTimerTimer(Sender: TObject);
var
  Display: Qt.PDisplay;
  Drawable: XID;
  Changes: XWindowChanges;
begin
  Display := QPaintDevice_x11AppDisplay;
  Drawable := QWidget_winID(Handle);
  Changes.stack_mode := TopIf;
  XConfigureWindow(Display, Drawable, 1 shl 6, @Changes);
end;

constructor TOpenBanner.Create(AOwner: TComponent);
begin
  CreateNew(AOwner);
  Init;
end;

procedure TOpenBanner.Init;

  procedure LoadImageFromResource(ResName: string; Picture: TPicture);
  begin
    if FindResource(HInstance, PChar(ResName), RT_BITMAP) <> 0 then
      Picture.Bitmap.LoadFromResourceName(HInstance, ResName);
  end;

var
  DeskWidth,
  DeskHeight: Integer;
begin
  Width := 383;
  Height := 159;
  BorderStyle := fbsNone;
  Color := clBackground;
  FormStyle := fsStayOnTop;
  Position := poMainFormCenter;
  OnClose := FormClose;
  PixelsPerInch := 75;

  MainPanel := TPanel.Create(Self);
  with MainPanel do
  begin
    Parent := Self;
    Left := 0;
    Top := 0;
    Width := 383;
    Height := 159;
    Align := alClient;
    BevelWidth := 2;
  end;

  TitlePanel := TPanel.Create(Self);
  with TitlePanel do
  begin
    Parent := MainPanel;
    Left := 2;
    Top := 2;
    Width := 379;
    Height := 41;
    Align := alTop;
    BevelOuter := bvNone;
    BorderWidth := 5;
    OnClick := IconImageClick;
  end;

  TitleImage := TImage.Create(Self);
  with TitleImage do
  begin
    Parent := TitlePanel;
    Left := 5;
    Top := 5;
    Width := 369;
    Height := 31;
    Align := alClient;
    Center := True;
    Transparent := True;
    OnClick := IconImageClick;
  end;

  IconPanel := TPanel.Create(Self);
  with IconPanel do
  begin
    Parent := MainPanel;
    Left := 2;
    Top := 43;
    Width := 103;
    Height := 114;
    Align := alLeft;
    BevelOuter := bvNone;
    BorderWidth := 5;
    OnClick := IconImageClick;
  end;

  IconImage := TImage.Create(Self);
  with IconImage do
  begin
    Parent := IconPanel;
    Left := 5;
    Top := 5;
    Width := 93;
    Height := 104;
    Align := alClient;
    Transparent := True;
    OnClick := IconImageClick;
  end;

  LicensePanel := TPanel.Create(Self);
  with LicensePanel do
  begin
    Parent := MainPanel;
    Left := 105;
    Top := 43;
    Width := 276;
    Height := 114;
    Align := alClient;
    BevelOuter := bvNone;
    BorderWidth := 5;
    OnClick := IconImageClick;
  end;

  LicenseLabel := TLabel.Create(Self);
  with LicenseLabel do
  begin
    Parent := LicensePanel;
    Left := 5;
    Top := 5;
    Width := 266;
    Height := 104;
    Align := alClient;
    BorderStyle := bsSunkenPanel;
    Caption := 'LicenseLabel';
    Color := clLight;
    ParentColor := False;
    WordWrap := True;
    OnClick := IconImageClick;
  end;

  Timer := TTimer.Create(Self);
  with Timer do
  begin
    Interval := 3000;
    OnTimer := TimerTimer;
  end;

  TopTimer := TTimer.Create(Self);
  with TopTimer do
  begin
    Interval := 100;
    OnTimer := TopTimerTimer;
  end;

  Height := MainPanel.BevelWidth * 2;
  Width := LicensePanel.Width + MainPanel.BevelWidth * 2;

  LoadImageFromResource('OSB_TITLEIMAGE', TitleImage.Picture);
  LoadImageFromResource('OSB_ICONIMAGE', IconImage.Picture);

  TitlePanel.Visible := TitleImage.Picture.Graphic <> nil;
  if TitlePanel.Visible then
  begin
    TitlePanel.Height := TitleImage.Picture.Graphic.Height + TitlePanel.BorderWidth * 2;
    Height := TitlePanel.Height + MainPanel.BevelWidth * 2;
    Width := Max(Width, PanelSize(TitlePanel).cx);
  end;
  IconPanel.Visible := IconImage.Picture.Graphic <> nil;
  if IconPanel.Visible then
  begin
    IconPanel.Width := IconImage.Picture.Graphic.Width + IconPanel.BorderWidth * 2;
    Height := Max(Height, PanelSize(TitlePanel).cy
      + IconImage.Picture.Graphic.Height + IconPanel.BorderWidth * 4
      + MainPanel.BevelWidth * 2);
    if PanelSize(TitlePanel).cx = 0 then
      Width := Width + IconPanel.Width
    else
      Width := Max(Width, IconPanel.Width + LicensePanel.Width);
  end;
  SetLicenseText(LicenseMessages[LicenseLocale]);
  DeskWidth := QWidget_width(QApplication_desktop);
  DeskHeight := QWidget_height(QApplication_desktop);
  Left := (DeskWidth div 2) - (Width div 2);
  Top := (DeskHeight div 2) - (Height div 2);

  Visible := True;
end;


initialization

  DisplayLicense;

finalization

end.
