{ *********************************************************************** }
{                                                                         }
{  Borland Kylix Runtime Library                                          }
{  Linux API Interface Unit (ELF)                                         }
{                                                                         }
{  This translation is based on glibc 2.2.1,                              }
{  compiled from sources on a Linux 2.2.16 kernel.                        }
{                                                                         }
{  Copyright (C) 2001-2002 Borland Software Corporation                   }
{                                                                         }
{  Translator: Borland Software Corporation                               }
{                                                                         }
{  This file may be distributed and/or modified under the terms of the    }
{  GNU General Public License version 2 as published by the Free Software }
{  Foundation and appearing at http://www.borland.com/kylix/gpl.html.     }
{                                                                         }
{  Licensees holding a valid Borland No-Nonsense License for this         }
{  Software may use this file in accordance with such license, which      }
{  appears in the file license.txt that came with this software.          }
{                                                                         }
{ *********************************************************************** }
unit LibcElf;

{$WEAKPACKAGEUNIT}

{$ALIGN 4}
{$MINENUMSIZE 4}
{$ASSERTIONS OFF}

interface

uses
  Libc;

// Translated from elf.h

{$HPPEMIT '#include <elf.h>'}

{ Type for a 16-bit quantity.  }
type
  Elf32_Half = uint16_t;
  {$EXTERNALSYM Elf32_Half}
  Elf64_Half = uint16_t;
  {$EXTERNALSYM Elf64_Half}

{ Types for signed and unsigned 32-bit quantities.  }
type
  Elf32_Word = uint32_t;
  {$EXTERNALSYM Elf32_Word}
  Elf32_Sword = int32_t;
  {$EXTERNALSYM Elf32_Sword}
  Elf64_Word = uint32_t;
  {$EXTERNALSYM Elf64_Word}
  Elf64_Sword = int32_t;
  {$EXTERNALSYM Elf64_Sword}

{ Types for signed and unsigned 64-bit quantities.  }
type
  Elf32_Xword = uint64_t;
  {$EXTERNALSYM Elf32_Xword}
  Elf32_Sxword = int64_t;
  {$EXTERNALSYM Elf32_Sxword}
  Elf64_Xword = uint64_t;
  {$EXTERNALSYM Elf64_Xword}
  Elf64_Sxword = int64_t;
  {$EXTERNALSYM Elf64_Sxword}

{ Type of addresses.  }
type
  Elf32_Addr = uint32_t;
  {$EXTERNALSYM Elf32_Addr}
  Elf64_Addr = uint64_t;
  {$EXTERNALSYM Elf64_Addr}

{ Type of file offsets.  }
type
  Elf32_Off = uint32_t;
  {$EXTERNALSYM Elf32_Off}
  Elf64_Off = uint64_t;
  {$EXTERNALSYM Elf64_Off}

{ Type for section indices, which are 16-bit quantities.  }
type
  Elf32_Section = uint16_t;
  {$EXTERNALSYM Elf32_Section}
  Elf64_Section = uint16_t;
  {$EXTERNALSYM Elf64_Section}

{ Type for version symbol information.  }
type
  Elf32_Versym = Elf32_Half;
  {$EXTERNALSYM Elf32_Versym}
  Elf64_Versym = Elf64_Half;
  {$EXTERNALSYM Elf64_Versym}


{ The ELF file header.  This appears at the start of every ELF file.  }
const
  EI_NIDENT = 16;
  {$EXTERNALSYM EI_NIDENT}

type
  Elf32_Ehdr = {packed} record
    e_ident: packed array[0..EI_NIDENT-1] of Byte;  { Magic number and other info }
    e_type: Elf32_Half;         { Object file type }
    e_machine: Elf32_Half;      { Architecture }
    e_version: Elf32_Word;      { Object file version }
    e_entry: Elf32_Addr;        { Entry point virtual address }
    e_phoff: Elf32_Off;         { Program header table file offset }
    e_shoff: Elf32_Off;         { Section header table file offset }
    e_flags: Elf32_Word;        { Processor-specific flags }
    e_ehsize: Elf32_Half;       { ELF header size in bytes }
    e_phentsize: Elf32_Half;    { Program header table entry size }
    e_phnum: Elf32_Half;        { Program header table entry count }
    e_shentsize: Elf32_Half;    { Section header table entry size }
    e_shnum: Elf32_Half;        { Section header table entry count }
    e_shstrndx: Elf32_Half;     { Section header string table index }
  end;
  {$EXTERNALSYM Elf32_Ehdr}
  PElf32_Ehdr = ^Elf32_Ehdr;

type
  Elf64_Ehdr = {packed} record
    e_ident: packed array[0..EI_NIDENT-1] of Byte;  { Magic number and other info }
    e_type: Elf64_Half;         { Object file type }
    e_machine: Elf64_Half;      { Architecture }
    e_version: Elf64_Word;      { Object file version }
    e_entry: Elf64_Addr;        { Entry point virtual address }
    e_phoff: Elf64_Off;         { Program header table file offset }
    e_shoff: Elf64_Off;         { Section header table file offset }
    e_flags: Elf64_Word;        { Processor-specific flags }
    e_ehsize: Elf64_Half;       { ELF header size in bytes }
    e_phentsize: Elf64_Half;    { Program header table entry size }
    e_phnum: Elf64_Half;        { Program header table entry count }
    e_shentsize: Elf64_Half;    { Section header table entry size }
    e_shnum: Elf64_Half;        { Section header table entry count }
    e_shstrndx: Elf64_Half;     { Section header string table index }
  end;
  {$EXTERNALSYM Elf64_Ehdr}
  PElf64_Ehdr = ^Elf64_Ehdr;

{ Fields in the e_ident array.  The EI_* macros are indices into the
   array.  The macros under each EI_* macro are the values the byte
   may have.  }
const
  EI_MAG0        = 0;      { File identification byte 0 index }
  {$EXTERNALSYM EI_MAG0}
  ELFMAG0        = $7f;    { Magic number byte 0 }
  {$EXTERNALSYM ELFMAG0}

  EI_MAG1        = 1;      { File identification byte 1 index }
  {$EXTERNALSYM EI_MAG1}
  ELFMAG1        = 'E';    { Magic number byte 1 }
  {$EXTERNALSYM ELFMAG1}

  EI_MAG2        = 2;      { File identification byte 2 index }
  {$EXTERNALSYM EI_MAG2}
  ELFMAG2        = 'L';    { Magic number byte 2 }
  {$EXTERNALSYM ELFMAG2}

  EI_MAG3        = 3;      { File identification byte 3 index }
  {$EXTERNALSYM EI_MAG3}
  ELFMAG3        = 'F';    { Magic number byte 3 }
  {$EXTERNALSYM ELFMAG3}

{ Conglomeration of the identification bytes, for easy testing as a word.  }
  ELFMAG = $464C457F; // "\177ELF"
  {$EXTERNALSYM ELFMAG}
  SELFMAG = 4;
  {$EXTERNALSYM SELFMAG}

  EI_CLASS          = 4;    { File class byte index }
  {$EXTERNALSYM EI_CLASS}
  ELFCLASSNONE      = 0;    { Invalid class }
  {$EXTERNALSYM ELFCLASSNONE}
  ELFCLASS32        = 1;    { 32-bit objects }
  {$EXTERNALSYM ELFCLASS32}
  ELFCLASS64        = 2;    { 64-bit objects }
  {$EXTERNALSYM ELFCLASS64}
  ELFCLASSNUM       = 3;
  {$EXTERNALSYM ELFCLASSNUM}

  EI_DATA           = 5;    { Data encoding byte index }
  {$EXTERNALSYM EI_DATA}
  ELFDATANONE       = 0;    { Invalid data encoding }
  {$EXTERNALSYM ELFDATANONE}
  ELFDATA2LSB       = 1;    { 2's complement, little endian }
  {$EXTERNALSYM ELFDATA2LSB}
  ELFDATA2MSB       = 2;    { 2's complement, big endian }
  {$EXTERNALSYM ELFDATA2MSB}
  ELFDATANUM        = 3;
  {$EXTERNALSYM ELFDATANUM}

  EI_VERSION        = 6;    { File version byte index }
                            { Value must be EV_CURRENT }
  {$EXTERNALSYM EI_VERSION}

  EI_OSABI          = 7;      { OS ABI identification }
  {$EXTERNALSYM EI_OSABI}
  ELFOSABI_SYSV     = 0;      { UNIX System V ABI }
  {$EXTERNALSYM ELFOSABI_SYSV}
  ELFOSABI_HPUX     = 1;      { HP-UX }
  {$EXTERNALSYM ELFOSABI_HPUX}
  ELFOSABI_ARM      = 97;     { ARM }
  {$EXTERNALSYM ELFOSABI_ARM}
  ELFOSABI_STANDALONE = 255;  { Standalone (embedded) application }
  {$EXTERNALSYM ELFOSABI_STANDALONE}

  EI_ABIVERSION     = 8;      { ABI version }
  {$EXTERNALSYM EI_ABIVERSION}

  EI_PAD            = 9;      { Byte index of padding bytes }
  {$EXTERNALSYM EI_PAD}

{ Legal values for e_type (object file type).  }

  ET_NONE      = 0;      { No file type }
  {$EXTERNALSYM ET_NONE}
  ET_REL       = 1;      { Relocatable file }
  {$EXTERNALSYM ET_REL}
  ET_EXEC      = 2;      { Executable file }
  {$EXTERNALSYM ET_EXEC}
  ET_DYN       = 3;      { Shared object file }
  {$EXTERNALSYM ET_DYN}
  ET_CORE      = 4;      { Core file }
  {$EXTERNALSYM ET_CORE}
  ET_NUM       = 5;      { Number of defined types }
  {$EXTERNALSYM ET_NUM}
  ET_LOOS      = $fe00;  { OS-specific range start }
  {$EXTERNALSYM ET_LOOS}
  ET_HIOS      = $feff;  { OS-specific range end }
  {$EXTERNALSYM ET_HIOS}
  ET_LOPROC    = $ff00;  { Processor-specific range start }
  {$EXTERNALSYM ET_LOPROC}
  ET_HIPROC    = $ffff;  { Processor-specific range end }
  {$EXTERNALSYM ET_HIPROC}

{ Legal values for e_machine (architecture).  }

  EM_NONE         = 0;       { No machine }
  {$EXTERNALSYM EM_NONE}
  EM_M32          = 1;       { AT&T WE 32100 }
  {$EXTERNALSYM EM_M32}
  EM_SPARC        = 2;       { SUN SPARC }
  {$EXTERNALSYM EM_SPARC}
  EM_386          = 3;       { Intel 80386 }
  {$EXTERNALSYM EM_386}
  EM_68K          = 4;       { Motorola m68k family }
  {$EXTERNALSYM EM_68K}
  EM_88K          = 5;       { Motorola m88k family }
  {$EXTERNALSYM EM_88K}
  EM_860          = 7;       { Intel 80860 }
  {$EXTERNALSYM EM_860}
  EM_MIPS         = 8;       { MIPS R3000 big-endian }
  {$EXTERNALSYM EM_MIPS}
  EM_MIPS_RS3_LE  = 10;      { MIPS R3000 little-endian }
  {$EXTERNALSYM EM_MIPS_RS3_LE}

  EM_PARISC       = 15;      { HPPA }
  {$EXTERNALSYM EM_PARISC}
  EM_VPP500       = 17;      { Fujitsu VPP500 }
  {$EXTERNALSYM EM_VPP500}
  EM_SPARC32PLUS  = 18;      { Sun's "v8plus" }
  {$EXTERNALSYM EM_SPARC32PLUS}
  EM_960          = 19;      { Intel 80960 }
  {$EXTERNALSYM EM_960}
  EM_PPC          = 20;      { PowerPC }
  {$EXTERNALSYM EM_PPC}
  EM_PPC64        = 21;      { PowerPC 64-bit }
  {$EXTERNALSYM EM_PPC64}

  EM_V800         = 36;      { NEC V800 series }
  {$EXTERNALSYM EM_V800}
  EM_FR20         = 37;      { Fujitsu FR20 }
  {$EXTERNALSYM EM_FR20}
  EM_RH32         = 38;      { TRW RH-32 }
  {$EXTERNALSYM EM_RH32}
  EM_RCE          = 39;      { Motorola RCE }
  {$EXTERNALSYM EM_RCE}
  EM_ARM          = 40;      { ARM }
  {$EXTERNALSYM EM_ARM}
  EM_FAKE_ALPHA   = 41;      { Digital Alpha }
  {$EXTERNALSYM EM_FAKE_ALPHA}
  EM_SH           = 42;      { Hitachi SH }
  {$EXTERNALSYM EM_SH}
  EM_SPARCV9      = 43;      { SPARC v9 64-bit }
  {$EXTERNALSYM EM_SPARCV9}
  EM_TRICORE      = 44;      { Siemens Tricore }
  {$EXTERNALSYM EM_TRICORE}
  EM_ARC          = 45;      { Argonaut RISC Core }
  {$EXTERNALSYM EM_ARC}
  EM_H8_300       = 46;      { Hitachi H8/300 }
  {$EXTERNALSYM EM_H8_300}
  EM_H8_300H      = 47;      { Hitachi H8/300H }
  {$EXTERNALSYM EM_H8_300H}
  EM_H8S          = 48;      { Hitachi H8S }
  {$EXTERNALSYM EM_H8S}
  EM_H8_500       = 49;      { Hitachi H8/500 }
  {$EXTERNALSYM EM_H8_500}
  EM_IA_64        = 50;      { Intel Merced }
  {$EXTERNALSYM EM_IA_64}
  EM_MIPS_X       = 51;      { Stanford MIPS-X }
  {$EXTERNALSYM EM_MIPS_X}
  EM_COLDFIRE     = 52;      { Motorola Coldfire }
  {$EXTERNALSYM EM_COLDFIRE}
  EM_68HC12       = 53;      { Motorola M68HC12 }
  {$EXTERNALSYM EM_68HC12}
  EM_MMA          = 54;      { Fujitsu MMA Multimedia Accelerator}
  {$EXTERNALSYM EM_MMA}
  EM_PCP          = 55;      { Siemens PCP }
  {$EXTERNALSYM EM_PCP}
  EM_NCPU         = 56;      { Sony nCPU embeeded RISC }
  {$EXTERNALSYM EM_NCPU}
  EM_NDR1         = 57;      { Denso NDR1 microprocessor }
  {$EXTERNALSYM EM_NDR1}
  EM_STARCORE     = 58;      { Motorola Start*Core processor }
  {$EXTERNALSYM EM_STARCORE}
  EM_ME16         = 59;      { Toyota ME16 processor }
  {$EXTERNALSYM EM_ME16}
  EM_ST100        = 60;      { STMicroelectronic ST100 processor }
  {$EXTERNALSYM EM_ST100}
  EM_TINYJ        = 61;      { Advanced Logic Corp. Tinyj emb.fam}
  {$EXTERNALSYM EM_TINYJ}
  EM_X8664        = 62;      { AMD x86-64 }
  {$EXTERNALSYM EM_X8664}

  EM_FX66         = 66;      { Siemens FX66 microcontroller }
  {$EXTERNALSYM EM_FX66}
  EM_ST9PLUS      = 67;      { STMicroelectronics ST9+ 8/16 mc }
  {$EXTERNALSYM EM_ST9PLUS}
  EM_ST7          = 68;      { STmicroelectronics ST7 8 bit mc }
  {$EXTERNALSYM EM_ST7}
  EM_68HC16       = 69;      { Motorola MC68HC16 microcontroller }
  {$EXTERNALSYM EM_68HC16}
  EM_68HC11       = 70;      { Motorola MC68HC11 microcontroller }
  {$EXTERNALSYM EM_68HC11}
  EM_68HC08       = 71;      { Motorola MC68HC08 microcontroller }
  {$EXTERNALSYM EM_68HC08}
  EM_68HC05       = 72;      { Motorola MC68HC05 microcontroller }
  {$EXTERNALSYM EM_68HC05}
  EM_SVX          = 73;      { Silicon Graphics SVx }
  {$EXTERNALSYM EM_SVX}
  EM_AT19         = 74;      { STMicroelectronics ST19 8 bit mc }
  {$EXTERNALSYM EM_AT19}
  EM_VAX          = 75;      { Digital VAX }
  {$EXTERNALSYM EM_VAX}
  EM_NUM          = 76;
  {$EXTERNALSYM EM_NUM}

{ If it is necessary to assign new unofficial EM_* values, please
   pick large random numbers (0x8523, 0xa7f2, etc.) to minimize the
   chances of collision with official or non-GNU unofficial values.  }

  EM_ALPHA        = $9026;
  {$EXTERNALSYM EM_ALPHA}
  EM_S390         = $A390;   { IBM S390 }
  {$EXTERNALSYM EM_S390}

{ Legal values for e_version (version).  }

  EV_NONE        = 0;      { Invalid ELF version }
  {$EXTERNALSYM EV_NONE}
  EV_CURRENT     = 1;      { Current version }
  {$EXTERNALSYM EV_CURRENT}
  EV_NUM         = 2;
  {$EXTERNALSYM EV_NUM}

{ Section header.  }

type
  Elf32_Shdr = {packed} record
    sh_name: Elf32_Word;        { Section name (string tbl index) }
    sh_type: Elf32_Word;        { Section type }
    sh_flags: Elf32_Word;       { Section flags }
    sh_addr: Elf32_Addr;        { Section virtual addr at execution }
    sh_offset: Elf32_Off;       { Section file offset }
    sh_size: Elf32_Word;        { Section size in bytes }
    sh_link: Elf32_Word;        { Link to another section }
    sh_info: Elf32_Word;        { Additional section information }
    sh_addralign: Elf32_Word;   { Section alignment }
    sh_entsize: Elf32_Word;     { Entry size if section holds table }
  end;
  {$EXTERNALSYM Elf32_Shdr}
  PElf32_Shdr = ^Elf32_Shdr;

type
  Elf64_Shdr = {packed} record
    sh_name: Elf64_Word;        { Section name (string tbl index) }
    sh_type: Elf64_Word;        { Section type }
    sh_flags: Elf64_Xword;      { Section flags }
    sh_addr: Elf64_Addr;        { Section virtual addr at execution }
    sh_offset: Elf64_Off;       { Section file offset }
    sh_size: Elf64_Xword;       { Section size in bytes }
    sh_link: Elf64_Word;        { Link to another section }
    sh_info: Elf64_Word;        { Additional section information }
    sh_addralign: Elf64_Xword;  { Section alignment }
    sh_entsize: Elf64_Xword;    { Entry size if section holds table }
  end;
  {$EXTERNALSYM Elf64_Shdr}
  PElf64_Shdr = ^Elf64_Shdr;

{ Special section indices.  }
const
  SHN_UNDEF       = 0;          { Undefined section }
  {$EXTERNALSYM SHN_UNDEF}
  SHN_LORESERVE   = $ff00;      { Start of reserved indices }
  {$EXTERNALSYM SHN_LORESERVE}
  SHN_LOPROC      = $ff00;      { Start of processor-specific }
  {$EXTERNALSYM SHN_LOPROC}
  SHN_HIPROC      = $ff1f;      { End of processor-specific }
  {$EXTERNALSYM SHN_HIPROC}
  SHN_LOOS        = $ff20;      { Start of OS-specific }
  {$EXTERNALSYM SHN_LOOS}
  SHN_HIOS        = $ff3f;      { End of OS-specific }
  {$EXTERNALSYM SHN_HIOS}
  SHN_ABS         = $fff1;      { Associated symbol is absolute }
  {$EXTERNALSYM SHN_ABS}
  SHN_COMMON      = $fff2;      { Associated symbol is common }
  {$EXTERNALSYM SHN_COMMON}
  SHN_XINDEX      = $ffff;      { Index is in extra table.  }
  {$EXTERNALSYM SHN_XINDEX}
  SHN_HIRESERVE   = $ffff;      { End of reserved indices }
  {$EXTERNALSYM SHN_HIRESERVE}

{ Legal values for sh_type (section type).  }
const
  SHT_NULL          = 0;         { Section header table entry unused }
  {$EXTERNALSYM SHT_NULL}
  SHT_PROGBITS      = 1;         { Program data }
  {$EXTERNALSYM SHT_PROGBITS}
  SHT_SYMTAB        = 2;         { Symbol table }
  {$EXTERNALSYM SHT_SYMTAB}
  SHT_STRTAB        = 3;         { String table }
  {$EXTERNALSYM SHT_STRTAB}
  SHT_RELA          = 4;         { Relocation entries with addends }
  {$EXTERNALSYM SHT_RELA}
  SHT_HASH          = 5;         { Symbol hash table }
  {$EXTERNALSYM SHT_HASH}
  SHT_DYNAMIC       = 6;         { Dynamic linking information }
  {$EXTERNALSYM SHT_DYNAMIC}
  SHT_NOTE          = 7;         { Notes }
  {$EXTERNALSYM SHT_NOTE}
  SHT_NOBITS        = 8;         { Program space with no data (bss) }
  {$EXTERNALSYM SHT_NOBITS}
  SHT_REL           = 9;         { Relocation entries, no addends }
  {$EXTERNALSYM SHT_REL}
  SHT_SHLIB         = 10;        { Reserved }
  {$EXTERNALSYM SHT_SHLIB}
  SHT_DYNSYM        = 11;        { Dynamic linker symbol table }
  {$EXTERNALSYM SHT_DYNSYM}
  SHT_INIT_ARRAY    = 14;        { Array of constructors }
  {$EXTERNALSYM SHT_INIT_ARRAY}
  SHT_FINI_ARRAY    = 15;        { Array of destructors }
  {$EXTERNALSYM SHT_FINI_ARRAY}
  SHT_PREINIT_ARRAY = 16;        { Array of pre-constructors }
  {$EXTERNALSYM SHT_PREINIT_ARRAY}
  SHT_GROUP         = 17;        { Section group }
  {$EXTERNALSYM SHT_GROUP}
  SHT_SYMTAB_SHNDX  = 18;        { Extended section indeces }
  {$EXTERNALSYM SHT_SYMTAB_SHNDX}
  SHT_NUM           = 19;        { Number of defined types.  }
  {$EXTERNALSYM SHT_NUM}
  SHT_LOOS          = $60000000; { Start OS-specific }
  {$EXTERNALSYM SHT_LOOS}
  SHT_LOSUNW        = $6ffffffa; { Sun-specific low bound.  }
  {$EXTERNALSYM SHT_LOSUNW}
  SHT_SUNW_move     = $6ffffffa;
  {$EXTERNALSYM SHT_SUNW_move}
  SHT_SUNW_COMDAT   = $6ffffffb;
  {$EXTERNALSYM SHT_SUNW_COMDAT}
  SHT_SUNW_syminfo  = $6ffffffc;
  {$EXTERNALSYM SHT_SUNW_syminfo}
  SHT_GNU_verdef    = $6ffffffd; { Version definition section.  }
  {$EXTERNALSYM SHT_GNU_verdef}
  SHT_GNU_verneed   = $6ffffffe; { Version needs section.  }
  {$EXTERNALSYM SHT_GNU_verneed}
  SHT_GNU_versym    = $6fffffff; { Version symbol table.  }
  {$EXTERNALSYM SHT_GNU_versym}
  SHT_HISUNW        = $6fffffff; { Sun-specific high bound.  }
  {$EXTERNALSYM SHT_HISUNW}
  SHT_HIOS          = $6fffffff; { End OS-specific type }
  {$EXTERNALSYM SHT_HIOS}
  SHT_LOPROC        = $70000000; { Start of processor-specific }
  {$EXTERNALSYM SHT_LOPROC}
  SHT_HIPROC        = $7fffffff; { End of processor-specific }
  {$EXTERNALSYM SHT_HIPROC}
  SHT_LOUSER        = $80000000; { Start of application-specific }
  {$EXTERNALSYM SHT_LOUSER}
  SHT_HIUSER        = $8fffffff; { End of application-specific }
  {$EXTERNALSYM SHT_HIUSER}

{ Legal values for sh_flags (section flags).  }
const
  SHF_WRITE             = (1 shl 0);    { Writable }
  {$EXTERNALSYM SHF_WRITE}
  SHF_ALLOC             = (1 shl 1);    { Occupies memory during execution }
  {$EXTERNALSYM SHF_ALLOC}
  SHF_EXECINSTR         = (1 shl 2);    { Executable }
  {$EXTERNALSYM SHF_EXECINSTR}
  SHF_MERGE             = (1 shl 4);    { Might be merged }
  {$EXTERNALSYM SHF_MERGE}
  SHF_STRINGS           = (1 shl 5);    { Contains nul-terminated strings }
  {$EXTERNALSYM SHF_STRINGS}
  SHF_INFO_LINK	        = (1 shl 6);    { `sh_info' contains SHT index }
  {$EXTERNALSYM SHF_INFO_LINK}
  SHF_LINK_ORDER        = (1 shl 7);    { Preserve order after combining }
  {$EXTERNALSYM SHF_LINK_ORDER}
  SHF_OS_NONCONFORMING  = (1 shl 8);    { Non-standard OS specific handling required }
  {$EXTERNALSYM SHF_OS_NONCONFORMING}
  SHF_MASKOS            = $0ff00000;    { OS-specific.  }
  {$EXTERNALSYM SHF_MASKOS}
  SHF_MASKPROC          = $f0000000;    { Processor-specific }
  {$EXTERNALSYM SHF_MASKPROC}

{ Section group handling.  }
  GRP_COMDAT            = $1;           { Mark group as COMDAT.  }
  {$EXTERNALSYM GRP_COMDAT}

{ Symbol table entry.  }

type
  Elf32_Sym = {packed} record
    st_name: Elf32_Word;      { Symbol name (string tbl index) }
    st_value: Elf32_Addr;     { Symbol value }
    st_size: Elf32_Word;      { Symbol size }
    st_info: Byte;            { Symbol type and binding }
    st_other: Byte;           { Symbol visibility }
    st_shndx: Elf32_Section;  { Section index }
  end;
  {$EXTERNALSYM Elf32_Sym}
  PElf32_Sym = ^Elf32_Sym;

type
  Elf64_Sym = {packed} record
    st_name: Elf64_Word;      { Symbol name (string tbl index) }
    st_info: Byte;            { Symbol type and binding }
    st_other: Byte;           { Symbol visibility }
    st_shndx: Elf64_Section;  { Section index }
    st_value: Elf64_Addr;     { Symbol value }
    st_size: Elf64_Xword;     { Symbol size }
  end;
  {$EXTERNALSYM Elf64_Sym}
  PElf64_Sym = ^Elf64_Sym;

{ The syminfo section if available contains additional information about
   every dynamic symbol.  }

type
  Elf32_Syminfo = {packed} record
    si_boundto: Elf32_Half;    { Direct bindings, symbol bound to }
    si_flags: Elf32_Half;      { Per symbol flags }
  end;
  {$EXTERNALSYM Elf32_Syminfo}
  PElf32_Syminfo = ^Elf32_Syminfo;

type
  Elf64_Syminfo = {packed} record
    si_boundto: Elf64_Half;    { Direct bindings, symbol bound to }
    si_flags: Elf64_Half;      { Per symbol flags }
  end;
  {$EXTERNALSYM Elf64_Syminfo}
  PElf64_Syminfo = ^Elf64_Syminfo;

{ Possible values for si_boundto.  }
const
  SYMINFO_BT_SELF        = $ffff;       { Symbol bound to self }
  {$EXTERNALSYM SYMINFO_BT_SELF}
  SYMINFO_BT_PARENT      = $fffe;       { Symbol bound to parent }
  {$EXTERNALSYM SYMINFO_BT_PARENT}
  SYMINFO_BT_LOWRESERVE  = $ff00;       { Beginning of reserved entries }
  {$EXTERNALSYM SYMINFO_BT_LOWRESERVE}

{ Possible bitmasks for si_flags.  }
  SYMINFO_FLG_DIRECT     = $0001;       { Direct bound symbol }
  {$EXTERNALSYM SYMINFO_FLG_DIRECT}
  SYMINFO_FLG_PASSTHRU   = $0002;       { Pass-thru symbol for translator }
  {$EXTERNALSYM SYMINFO_FLG_PASSTHRU}
  SYMINFO_FLG_COPY       = $0004;       { Symbol is a copy-reloc }
  {$EXTERNALSYM SYMINFO_FLG_COPY}
  SYMINFO_FLG_LAZYLOAD   = $0008;       { Symbol bound to object to be lazy loaded }
  {$EXTERNALSYM SYMINFO_FLG_LAZYLOAD}

{ Syminfo version values.  }
  SYMINFO_NONE           = 0;
  {$EXTERNALSYM SYMINFO_NONE}
  SYMINFO_CURRENT        = 1;
  {$EXTERNALSYM SYMINFO_CURRENT}
  SYMINFO_NUM            = 2;
  {$EXTERNALSYM SYMINFO_NUM}


{ Special section index.  }

(* SHN_UNDEF declared twice...
  SHN_UNDEF              = 0;       { No section, undefined symbol.  }
*)

{ How to extract and insert information held in the st_info field.  }

function ELF32_ST_BIND(const val: Byte): Byte;
{$EXTERNALSYM ELF32_ST_BIND}
function ELF32_ST_TYPE(const val: Byte): Byte;
{$EXTERNALSYM ELF32_ST_TYPE}
function ELF32_ST_INFO(const bind, __type: Byte): Byte;
{$EXTERNALSYM ELF32_ST_INFO}

{ Both Elf32_Sym and Elf64_Sym use the same one-byte st_info field.  }
function ELF64_ST_BIND(const val: Byte): Byte;
{$EXTERNALSYM ELF64_ST_BIND}
function ELF64_ST_TYPE(const val: Byte): Byte;
{$EXTERNALSYM ELF64_ST_TYPE}
function ELF64_ST_INFO(const bind, __type: Byte): Byte;
{$EXTERNALSYM ELF64_ST_INFO}

{ Legal values for ST_BIND subfield of st_info (symbol binding).  }
const
  STB_LOCAL     = 0;        { Local symbol }
  {$EXTERNALSYM STB_LOCAL}
  STB_GLOBAL    = 1;        { Global symbol }
  {$EXTERNALSYM STB_GLOBAL}
  STB_WEAK      = 2;        { Weak symbol }
  {$EXTERNALSYM STB_WEAK}
  STB_NUM       = 3;        { Number of defined types.  }
  {$EXTERNALSYM STB_NUM}
  STB_LOOS      = 10;       { Start of OS-specific }
  {$EXTERNALSYM STB_LOOS}
  STB_HIOS      = 12;       { End of OS-specific }
  {$EXTERNALSYM STB_HIOS}
  STB_LOPROC    = 13;       { Start of processor-specific }
  {$EXTERNALSYM STB_LOPROC}
  STB_HIPROC    = 15;       { End of processor-specific }
  {$EXTERNALSYM STB_HIPROC}

{ Legal values for ST_TYPE subfield of st_info (symbol type).  }

  STT_NOTYPE    = 0;        { Symbol type is unspecified }
  {$EXTERNALSYM STT_NOTYPE}
  STT_OBJECT    = 1;        { Symbol is a data object }
  {$EXTERNALSYM STT_OBJECT}
  STT_FUNC      = 2;        { Symbol is a code object }
  {$EXTERNALSYM STT_FUNC}
  STT_SECTION   = 3;        { Symbol associated with a section }
  {$EXTERNALSYM STT_SECTION}
  STT_FILE      = 4;        { Symbol's name is file name }
  {$EXTERNALSYM STT_FILE}
  STT_COMMON    = 5;        { Symbol is a common data object }
  {$EXTERNALSYM STT_COMMON}
  STT_NUM       = 6;        { Number of defined types.  }
  {$EXTERNALSYM STT_NUM}
  STT_LOOS      = 10;       { Start of OS-specific }
  {$EXTERNALSYM STT_LOOS}
  STT_HIOS      = 12;       { End of OS-specific }
  {$EXTERNALSYM STT_HIOS}
  STT_LOPROC    = 13;       { Start of processor-specific }
  {$EXTERNALSYM STT_LOPROC}
  STT_HIPROC    = 15;       { End of processor-specific }
  {$EXTERNALSYM STT_HIPROC}


{ Symbol table indices are found in the hash buckets and chain table
   of a symbol hash table section.  This special index value indicates
   the end of a chain, meaning no further symbols are found in that bucket.  }

  STN_UNDEF     = 0;        { End of a chain.  }
  {$EXTERNALSYM STN_UNDEF}


{ How to extract and insert information held in the st_other field.  }

function ELF32_ST_VISIBILITY(const o: Byte): Byte;
{$EXTERNALSYM ELF32_ST_VISIBILITY}

{ For ELF64 the definitions are the same.  }

function ELF64_ST_VISIBILITY(const o: Byte): Byte;
{$EXTERNALSYM ELF64_ST_VISIBILITY}

{ Symbol visibility specification encoded in the st_other field.  }
const
  STV_DEFAULT      = 0;         { Default symbol visibility rules }
  {$EXTERNALSYM STV_DEFAULT}
  STV_INTERNAL     = 1;         { Processor specific hidden class }
  {$EXTERNALSYM STV_INTERNAL}
  STV_HIDDEN       = 2;         { Sym unavailable in other modules }
  {$EXTERNALSYM STV_HIDDEN}
  STV_PROTECTED    = 3;         { Not preemptible, not exported }
  {$EXTERNALSYM STV_PROTECTED}


{ Relocation table entry without addend (in section of type SHT_REL).  }

type
  Elf32_Rel = {packed} record
    r_offset: Elf32_Addr;       { Address }
    r_info: Elf32_Word;         { Relocation type and symbol index }
  end;
  {$EXTERNALSYM Elf32_Rel}
  PElf32_Rel = ^Elf32_Rel;

{ I have seen two different definitions of the Elf64_Rel and
   Elf64_Rela structures, so we'll leave them out until Novell (or
   whoever) gets their act together.  }
{ The following, at least, is used on Sparc v9, MIPS, and Alpha.  }

type
  Elf64_Rel = {packed} record
    r_offset: Elf64_Addr;       { Address }
    r_info: Elf64_Xword;        { Relocation type and symbol index }
  end;
  {$EXTERNALSYM Elf64_Rel}
  PElf64_Rel = ^Elf64_Rel;

{ Relocation table entry with addend (in section of type SHT_RELA).  }

type
  Elf32_Rela = {packed} record
    r_offset: Elf32_Addr;       { Address }
    r_info: Elf32_Word;         { Relocation type and symbol index }
    r_addend: Elf32_Sword;      { Addend }
  end;
  {$EXTERNALSYM Elf32_Rela}
  PElf32_Rela = ^Elf32_Rela;

type
  Elf64_Rela = {packed} record
    r_offset: Elf64_Addr;       { Address }
    r_info: Elf64_Xword;        { Relocation type and symbol index }
    r_addend: Elf64_Sxword;     { Addend }
  end;
  {$EXTERNALSYM Elf64_Rela}
  PElf64_Rela = ^Elf64_Rela;

{ How to extract and insert information held in the r_info field.  }

function ELF32_R_SYM(const val: Elf32_Word): Elf32_Word;
{$EXTERNALSYM ELF32_R_SYM}
function ELF32_R_TYPE(const val: Elf32_Word): Elf32_Word;
{$EXTERNALSYM ELF32_R_TYPE}
function ELF32_R_INFO(const sym, __type: Elf32_Word): Elf32_Word;
{$EXTERNALSYM ELF32_R_INFO}

function ELF64_R_SYM(const val: Elf64_Xword): Elf64_Xword;
{$EXTERNALSYM ELF64_R_SYM}
function ELF64_R_TYPE(const val: Elf64_Xword): Elf64_Xword;
{$EXTERNALSYM ELF64_R_TYPE}
function ELF64_R_INFO(const sym, __type: Elf64_Xword): Elf64_Xword;
{$EXTERNALSYM ELF64_R_INFO}

{ Program segment header.  }

type
  Elf32_Phdr = {packed} record
    p_type: Elf32_Word;      { Segment type }
    p_offset: Elf32_Off;     { Segment file offset }
    p_vaddr: Elf32_Addr;     { Segment virtual address }
    p_paddr: Elf32_Addr;     { Segment physical address }
    p_filesz: Elf32_Word;    { Segment size in file }
    p_memsz: Elf32_Word;     { Segment size in memory }
    p_flags: Elf32_Word;     { Segment flags }
    p_align: Elf32_Word;     { Segment alignment }
  end;
  {$EXTERNALSYM Elf32_Phdr}
  PElf32_Phdr = ^Elf32_Phdr;

type
  Elf64_Phdr = {packed} record
    p_type: Elf64_Word;	     { Segment type }
    p_flags: Elf64_Word;     { Segment flags }
    p_offset: Elf64_Off;     { Segment file offset }
    p_vaddr: Elf64_Addr;     { Segment virtual address }
    p_paddr: Elf64_Addr;     { Segment physical address }
    p_filesz: Elf64_Xword;   { Segment size in file }
    p_memsz: Elf64_Xword;    { Segment size in memory }
    p_align: Elf64_Xword;    { Segment alignment }
  end;
  {$EXTERNALSYM Elf64_Phdr}
  PElf64_Phdr = ^Elf64_Phdr;

{ Legal values for p_type (segment type).  }
const
  PT_NULL       = 0;            { Program header table entry unused }
  {$EXTERNALSYM PT_NULL}
  PT_LOAD       = 1;            { Loadable program segment }
  {$EXTERNALSYM PT_LOAD}
  PT_DYNAMIC    = 2;            { Dynamic linking information }
  {$EXTERNALSYM PT_DYNAMIC}
  PT_INTERP     = 3;            { Program interpreter }
  {$EXTERNALSYM PT_INTERP}
  PT_NOTE       = 4;            { Auxiliary information }
  {$EXTERNALSYM PT_NOTE}
  PT_SHLIB      = 5;            { Reserved }
  {$EXTERNALSYM PT_SHLIB}
  PT_PHDR       = 6;            { Entry for header table itself }
  {$EXTERNALSYM PT_PHDR}
  PT_NUM        = 7;            { Number of defined types }
  {$EXTERNALSYM PT_NUM}
  PT_LOOS       = $60000000;    { Start of OS-specific }
  {$EXTERNALSYM PT_LOOS}
  PT_HIOS       = $6fffffff;    { End of OS-specific }
  {$EXTERNALSYM PT_HIOS}
  PT_LOPROC     = $70000000;    { Start of processor-specific }
  {$EXTERNALSYM PT_LOPROC}
  PT_HIPROC     = $7fffffff;    { End of processor-specific }
  {$EXTERNALSYM PT_HIPROC}

{ Legal values for p_flags (segment flags).  }

  PF_X          = (1 shl 0);    { Segment is executable }
  {$EXTERNALSYM PF_X}
  PF_W          = (1 shl 1);    { Segment is writable }
  {$EXTERNALSYM PF_W}
  PF_R          = (1 shl 2);    { Segment is readable }
  {$EXTERNALSYM PF_R}
  PF_MASKOS     = $0ff00000;    { OS-specific }
  {$EXTERNALSYM PF_MASKOS}
  PF_MASKPROC   = $f0000000;    { Processor-specific }
  {$EXTERNALSYM PF_MASKPROC}

{ Legal values for note segment descriptor types for core files. }

  NT_PRSTATUS   = 1;            { Contains copy of prstatus struct }
  {$EXTERNALSYM NT_PRSTATUS}
  NT_FPREGSET   = 2;            { Contains copy of fpregset struct }
  {$EXTERNALSYM NT_FPREGSET}
  NT_PRPSINFO   = 3;            { Contains copy of prpsinfo struct }
  {$EXTERNALSYM NT_PRPSINFO}
  NT_PRXREG     = 4;            { Contains copy of prxregset struct }
  {$EXTERNALSYM NT_PRXREG}
  NT_PLATFORM   = 5;            { String from sysinfo(SI_PLATFORM) }
  {$EXTERNALSYM NT_PLATFORM}
  NT_AUXV       = 6;            { Contains copy of auxv array }
  {$EXTERNALSYM NT_AUXV}
  NT_GWINDOWS   = 7;            { Contains copy of gwindows struct }
  {$EXTERNALSYM NT_GWINDOWS}
  NT_PSTATUS    = 10;           { Contains copy of pstatus struct }
  {$EXTERNALSYM NT_PSTATUS}
  NT_PSINFO     = 13;           { Contains copy of psinfo struct }
  {$EXTERNALSYM NT_PSINFO}
  NT_PRCRED     = 14;           { Contains copy of prcred struct }
  {$EXTERNALSYM NT_PRCRED}
  NT_UTSNAME    = 15;           { Contains copy of utsname struct }
  {$EXTERNALSYM NT_UTSNAME}
  NT_LWPSTATUS  = 16;           { Contains copy of lwpstatus struct }
  {$EXTERNALSYM NT_LWPSTATUS}
  NT_LWPSINFO   = 17;           { Contains copy of lwpinfo struct }
  {$EXTERNALSYM NT_LWPSINFO}
  NT_PRFPXREG   = 20;           { Contains copy of fprxregset struct}
  {$EXTERNALSYM NT_PRFPXREG}

{ Legal values for the note segment descriptor types for object files.  }

  NT_VERSION    = 1;            { Contains a version string.  }
  {$EXTERNALSYM NT_VERSION}


{ Dynamic section entry.  }

type
  Elf32_Dyn = {packed} record
    d_tag: Elf32_Sword;           { Dynamic entry type }
    case Integer of
      0: (d_val: Elf32_Word);     { Integer value }
      1: (d_ptr: Elf32_Addr);     { Address value }
    end;
  {$EXTERNALSYM Elf32_Dyn}
  PElf32_Dyn = ^Elf32_Dyn;

type
  Elf64_Dyn = {packed} record
    d_tag: Elf64_Sxword;          { Dynamic entry type }
    case Integer of
      0: (d_val: Elf64_Xword);    { Integer value }
      1: (d_ptr: Elf64_Addr);     { Address value }
    end;
  {$EXTERNALSYM Elf64_Dyn}
  PElf64_Dyn = ^Elf64_Dyn;

{ Legal values for d_tag (dynamic entry type).  }
const
  DT_NULL               = 0;         { Marks end of dynamic section }
  {$EXTERNALSYM DT_NULL}
  DT_NEEDED             = 1;         { Name of needed library }
  {$EXTERNALSYM DT_NEEDED}
  DT_PLTRELSZ           = 2;         { Size in bytes of PLT relocs }
  {$EXTERNALSYM DT_PLTRELSZ}
  DT_PLTGOT             = 3;         { Processor defined value }
  {$EXTERNALSYM DT_PLTGOT}
  DT_HASH               = 4;         { Address of symbol hash table }
  {$EXTERNALSYM DT_HASH}
  DT_STRTAB             = 5;         { Address of string table }
  {$EXTERNALSYM DT_STRTAB}
  DT_SYMTAB             = 6;         { Address of symbol table }
  {$EXTERNALSYM DT_SYMTAB}
  DT_RELA               = 7;         { Address of Rela relocs }
  {$EXTERNALSYM DT_RELA}
  DT_RELASZ             = 8;         { Total size of Rela relocs }
  {$EXTERNALSYM DT_RELASZ}
  DT_RELAENT            = 9;         { Size of one Rela reloc }
  {$EXTERNALSYM DT_RELAENT}
  DT_STRSZ              = 10;        { Size of string table }
  {$EXTERNALSYM DT_STRSZ}
  DT_SYMENT             = 11;        { Size of one symbol table entry }
  {$EXTERNALSYM DT_SYMENT}
  DT_INIT               = 12;        { Address of init function }
  {$EXTERNALSYM DT_INIT}
  DT_FINI               = 13;        { Address of termination function }
  {$EXTERNALSYM DT_FINI}
  DT_SONAME             = 14;        { Name of shared object }
  {$EXTERNALSYM DT_SONAME}
  DT_RPATH              = 15;        { Library search path (deprecated) }
  {$EXTERNALSYM DT_RPATH}
  DT_SYMBOLIC           = 16;        { Start symbol search here }
  {$EXTERNALSYM DT_SYMBOLIC}
  DT_REL                = 17;        { Address of Rel relocs }
  {$EXTERNALSYM DT_REL}
  DT_RELSZ              = 18;        { Total size of Rel relocs }
  {$EXTERNALSYM DT_RELSZ}
  DT_RELENT             = 19;        { Size of one Rel reloc }
  {$EXTERNALSYM DT_RELENT}
  DT_PLTREL             = 20;        { Type of reloc in PLT }
  {$EXTERNALSYM DT_PLTREL}
  DT_DEBUG              = 21;        { For debugging; unspecified }
  {$EXTERNALSYM DT_DEBUG}
  DT_TEXTREL            = 22;        { Reloc might modify .text }
  {$EXTERNALSYM DT_TEXTREL}
  DT_JMPREL             = 23;        { Address of PLT relocs }
  {$EXTERNALSYM DT_JMPREL}
  DT_BIND_NOW           = 24;        { Process relocations of object }
  {$EXTERNALSYM DT_BIND_NOW}
  DT_INIT_ARRAY         = 25;        { Array with addresses of init fct }
  {$EXTERNALSYM DT_INIT_ARRAY}
  DT_FINI_ARRAY         = 26;        { Array with addresses of fini fct }
  {$EXTERNALSYM DT_FINI_ARRAY}
  DT_INIT_ARRAYSZ       = 27;        { Size in bytes of DT_INIT_ARRAY }
  {$EXTERNALSYM DT_INIT_ARRAYSZ}
  DT_FINI_ARRAYSZ       = 28;        { Size in bytes of DT_FINI_ARRAY }
  {$EXTERNALSYM DT_FINI_ARRAYSZ}
  DT_RUNPATH            = 29;        { Library search path }
  {$EXTERNALSYM DT_RUNPATH}
  DT_FLAGS              = 30;        { Flags for the object being loaded }
  {$EXTERNALSYM DT_FLAGS}
  DT_ENCODING           = 32;        { Start of encoded range }
  {$EXTERNALSYM DT_ENCODING}
  DT_PREINIT_ARRAY      = 32;        { Array with addresses of preinit fct}
  {$EXTERNALSYM DT_PREINIT_ARRAY}
  DT_PREINIT_ARRAYSZ    = 33;        { size in bytes of DT_PREINIT_ARRAY }
  {$EXTERNALSYM DT_PREINIT_ARRAYSZ}
  DT_NUM                = 34;        { Number used }
  {$EXTERNALSYM DT_NUM}
  DT_LOOS               = $60000000; { Start of OS-specific }
  {$EXTERNALSYM DT_LOOS}
  DT_HIOS               = $6fffffff; { End of OS-specific }
  {$EXTERNALSYM DT_HIOS}
  DT_LOPROC             = $70000000; { Start of processor-specific }
  {$EXTERNALSYM DT_LOPROC}
  DT_HIPROC             = $7fffffff; { End of processor-specific }
  {$EXTERNALSYM DT_HIPROC}

  (* Moved DT_MIPS_NUM up to resolve dependency *)
  DT_MIPS_NUM           = $32;
  {$EXTERNALSYM DT_MIPS_NUM}

  DT_PROCNUM            = DT_MIPS_NUM; { Most used by any processor }
  {$EXTERNALSYM DT_PROCNUM}

{ DT_* entries which fall between DT_VALRNGHI & DT_VALRNGLO use the
   Dyn.d_un.d_val field of the Elf*_Dyn structure.  This follows Sun's
   approach.  }
  DT_VALRNGLO           = $6ffffd00;
  {$EXTERNALSYM DT_VALRNGLO}
  DT_PLTPADSZ           = $6ffffdf9;
  {$EXTERNALSYM DT_PLTPADSZ}
  DT_MOVEENT            = $6ffffdfa;
  {$EXTERNALSYM DT_MOVEENT}
  DT_MOVESZ             = $6ffffdfb;
  {$EXTERNALSYM DT_MOVESZ}
  DT_FEATURE_1          = $6ffffdfc; { Feature selection (DTF_*).  }
  {$EXTERNALSYM DT_FEATURE_1}
  DT_POSFLAG_1          = $6ffffdfd; { Flags for DT_* entries, effecting
                                       the following DT_* entry.  }
  {$EXTERNALSYM DT_POSFLAG_1}
  DT_SYMINSZ            = $6ffffdfe; { Size of syminfo table (in bytes) }
  {$EXTERNALSYM DT_SYMINSZ}
  DT_SYMINENT           = $6ffffdff; { Entry size of syminfo }
  {$EXTERNALSYM DT_SYMINENT}
  DT_VALRNGHI           = $6ffffdff;
  {$EXTERNALSYM DT_VALRNGHI}

{ DT_* entries which fall between DT_ADDRRNGHI & DT_ADDRRNGLO use the
   Dyn.d_un.d_ptr field of the Elf*_Dyn structure.

   If any adjustment is made to the ELF object after it has been
   built these entries will need to be adjusted.  }
  DT_ADDRRNGLO          = $6ffffe00;
  {$EXTERNALSYM DT_ADDRRNGLO}
  DT_SYMINFO            = $6ffffeff; { syminfo table }
  {$EXTERNALSYM DT_SYMINFO}
  DT_ADDRRNGHI          = $6ffffeff;
  {$EXTERNALSYM DT_ADDRRNGHI}

{ The versioning entry types.  The next are defined as part of the
   GNU extension.  }
  DT_VERSYM             = $6ffffff0;
  {$EXTERNALSYM DT_VERSYM}

  DT_RELACOUNT          = $6ffffff9;
  {$EXTERNALSYM DT_RELACOUNT}
  DT_RELCOUNT           = $6ffffffa;
  {$EXTERNALSYM DT_RELCOUNT}

{ These were chosen by Sun.  }
  DT_FLAGS_1            = $6ffffffb; { State flags, see DF_1_* below.  }
  {$EXTERNALSYM DT_FLAGS_1}
  DT_VERDEF             = $6ffffffc; { Address of version definition table }
  {$EXTERNALSYM DT_VERDEF}
  DT_VERDEFNUM          = $6ffffffd; { Number of version definitions }
  {$EXTERNALSYM DT_VERDEFNUM}
  DT_VERNEED            = $6ffffffe; { Address of table with needed versions }
  {$EXTERNALSYM DT_VERNEED}
  DT_VERNEEDNUM         = $6fffffff; { Number of needed versions }
  {$EXTERNALSYM DT_VERNEEDNUM}

function DT_VERSIONTAGIDX(const tag: LongWord): LongWord;
{$EXTERNALSYM DT_VERSIONTAGIDX}

const
  DT_VERSIONTAGNUM      = $16;
  {$EXTERNALSYM DT_VERSIONTAGNUM}

{ Sun added these machine-independent extensions in the "processor-specific"
   range.  Be compatible.  }
  DT_AUXILIARY          = $7ffffffd; { Shared object to load before self }
  {$EXTERNALSYM DT_AUXILIARY}
  DT_FILTER             = $7fffffff; { Shared object to get values from }
  {$EXTERNALSYM DT_FILTER}

function DT_EXTRATAGIDX(const tag: Elf32_Word): Elf32_Word;
{$EXTERNALSYM DT_EXTRATAGIDX}

const
  DT_EXTRANUM           = 3;
  {$EXTERNALSYM DT_EXTRANUM}

{ Values of `d_un.d_val' in the DT_FLAGS entry.  }
  DF_ORIGIN             = $00000001; { Object may use DF_ORIGIN }
  {$EXTERNALSYM DF_ORIGIN}
  DF_SYMBOLIC           = $00000002; { Symbol resolutions starts here }
  {$EXTERNALSYM DF_SYMBOLIC}
  DF_TEXTREL            = $00000004; { Object contains text relocations }
  {$EXTERNALSYM DF_TEXTREL}
  DF_BIND_NOW           = $00000008; { No lazy binding for this object }
  {$EXTERNALSYM DF_BIND_NOW}

{ State flags selectable in the `d_un.d_val' element of the DT_FLAGS_1
   entry in the dynamic section.  }
  DF_1_NOW              = $00000001; { Set RTLD_NOW for this object.  }
  {$EXTERNALSYM DF_1_NOW}
  DF_1_GLOBAL           = $00000002; { Set RTLD_GLOBAL for this object.  }
  {$EXTERNALSYM DF_1_GLOBAL}
  DF_1_GROUP            = $00000004; { Set RTLD_GROUP for this object.  }
  {$EXTERNALSYM DF_1_GROUP}
  DF_1_NODELETE         = $00000008; { Set RTLD_NODELETE for this object.}
  {$EXTERNALSYM DF_1_NODELETE}
  DF_1_LOADFLTR         = $00000010; { Trigger filtee loading at runtime.}
  {$EXTERNALSYM DF_1_LOADFLTR}
  DF_1_INITFIRST        = $00000020; { Set RTLD_INITFIRST for this object}
  {$EXTERNALSYM DF_1_INITFIRST}
  DF_1_NOOPEN           = $00000040; { Set RTLD_NOOPEN for this object.  }
  {$EXTERNALSYM DF_1_NOOPEN}
  DF_1_ORIGIN           = $00000080; { $ORIGIN must be handled.  }
  {$EXTERNALSYM DF_1_ORIGIN}
  DF_1_DIRECT           = $00000100; { Direct binding enabled.  }
  {$EXTERNALSYM DF_1_DIRECT}
  DF_1_TRANS            = $00000200;
  {$EXTERNALSYM DF_1_TRANS}
  DF_1_INTERPOSE        = $00000400; { Object is used to interpose.  }
  {$EXTERNALSYM DF_1_INTERPOSE}
  DF_1_NODEFLIB         = $00000800; { Ignore default lib search path.  }
  {$EXTERNALSYM DF_1_NODEFLIB}
  DF_1_NODUMP           = $00001000;
  {$EXTERNALSYM DF_1_NODUMP}
  DF_1_CONFALT          = $00002000;
  {$EXTERNALSYM DF_1_CONFALT}
  DF_1_ENDFILTEE        = $00004000;
  {$EXTERNALSYM DF_1_ENDFILTEE}

{ Flags for the feature selection in DT_FEATURE_1.  }
  DTF_1_PARINIT         = $00000001;
  {$EXTERNALSYM DTF_1_PARINIT}
  DTF_1_CONFEXP         = $00000002;
  {$EXTERNALSYM DTF_1_CONFEXP}

{ Flags in the DT_POSFLAG_1 entry effecting only the next DT_* entry.  }
  DF_P1_LAZYLOAD        = $00000001; { Lazyload following object.  }
  {$EXTERNALSYM DF_P1_LAZYLOAD}
  DF_P1_GROUPPERM       = $00000002; { Symbols from next object are not
                                       generally available.  }
  {$EXTERNALSYM DF_P1_GROUPPERM}

{ Version definition sections.  }

type
  Elf32_Verdef = {packed} record
    vd_version: Elf32_Half;     { Version revision }
    vd_flags: Elf32_Half;       { Version information }
    vd_ndx: Elf32_Half;         { Version Index }
    vd_cnt: Elf32_Half;         { Number of associated aux entries }
    vd_hash: Elf32_Word;        { Version name hash value }
    vd_aux: Elf32_Word;         { Offset in bytes to verdaux array }
    vd_next: Elf32_Word;        { Offset in bytes to next verdef entry }
  end;
  {$EXTERNALSYM Elf32_Verdef}
  PElf32_Verdef = ^Elf32_Verdef;

type
  Elf64_Verdef = {packed} record
    vd_version: Elf64_Half;     { Version revision }
    vd_flags: Elf64_Half;       { Version information }
    vd_ndx: Elf64_Half;         { Version Index }
    vd_cnt: Elf64_Half;         { Number of associated aux entries }
    vd_hash: Elf64_Word;        { Version name hash value }
    vd_aux: Elf64_Word;         { Offset in bytes to verdaux array }
    vd_next: Elf64_Word;        { Offset in bytes to next verdef entry }
  end;
  {$EXTERNALSYM Elf64_Verdef}
  PElf64_Verdef = ^Elf64_Verdef;


{ Legal values for vd_version (version revision).  }
const
  VER_DEF_NONE       = 0;         { No version }
  {$EXTERNALSYM VER_DEF_NONE}
  VER_DEF_CURRENT    = 1;         { Current version }
  {$EXTERNALSYM VER_DEF_CURRENT}
  VER_DEF_NUM        = 2;         { Given version number }
  {$EXTERNALSYM VER_DEF_NUM}

{ Legal values for vd_flags (version information flags).  }
  VER_FLG_BASE       = $1;        { Version definition of file itself }
  {$EXTERNALSYM VER_FLG_BASE}
  VER_FLG_WEAK       = $2;        { Weak version identifier }
  {$EXTERNALSYM VER_FLG_WEAK}

{ Auxiliary version information.  }

type
  Elf32_Verdaux = {packed} record
    vda_name: Elf32_Word;      { Version or dependency names }
    vda_next: Elf32_Word;      { Offset in bytes to next verdaux entry }
  end;
  {$EXTERNALSYM Elf32_Verdaux}
  PElf32_Verdaux = ^Elf32_Verdaux;

type
  Elf64_Verdaux = {packed} record
    vda_name: Elf64_Word;      { Version or dependency names }
    vda_next: Elf64_Word;      { Offset in bytes to next verdaux entry }
  end;
  {$EXTERNALSYM Elf64_Verdaux}
  PElf64_Verdaux = ^Elf64_Verdaux;


{ Version dependency section.  }

type
  Elf32_Verneed = {packed} record
    vn_version: Elf32_Half;    { Version of structure }
    vn_cnt: Elf32_Half;        { Number of associated aux entries }
    vn_file: Elf32_Word;       { Offset of filename for this dependency }
    vn_aux: Elf32_Word;        { Offset in bytes to vernaux array }
    vn_next: Elf32_Word;       { Offset in bytes to next verneed entry }
  end;
  {$EXTERNALSYM Elf32_Verneed}
  PElf32_Verneed = ^Elf32_Verneed;

type
  Elf64_Verneed = {packed} record
    vn_version: Elf64_Half;    { Version of structure }
    vn_cnt: Elf64_Half;        { Number of associated aux entries }
    vn_file: Elf64_Word;       { Offset of filename for this dependency }
    vn_aux: Elf64_Word;        { Offset in bytes to vernaux array }
    vn_next: Elf64_Word;       { Offset in bytes to next verneed entry }
  end;
  {$EXTERNALSYM Elf64_Verneed}
  PElf64_Verneed = ^Elf64_Verneed;


{ Legal values for vn_version (version revision).  }
const
  VER_NEED_NONE      = 0;   { No version }
  {$EXTERNALSYM VER_NEED_NONE}
  VER_NEED_CURRENT   = 1;   { Current version }
  {$EXTERNALSYM VER_NEED_CURRENT}
  VER_NEED_NUM       = 2;   { Given version number }
  {$EXTERNALSYM VER_NEED_NUM}

{ Auxiliary needed version information.  }

type
  Elf32_Vernaux = {packed} record
    vna_hash: Elf32_Word;   { Hash value of dependency name }
    vna_flags: Elf32_Half;  { Dependency specific information }
    vna_other: Elf32_Half;  { Unused }
    vna_name: Elf32_Word;   { Dependency name string offset }
    vna_next: Elf32_Word;   { Offset in bytes to next vernaux entry }
  end;
  {$EXTERNALSYM Elf32_Vernaux}
  PElf32_Vernaux = ^Elf32_Vernaux;

type
  Elf64_Vernaux = {packed} record
    vna_hash: Elf64_Word;   { Hash value of dependency name }
    vna_flags: Elf64_Half;  { Dependency specific information }
    vna_other: Elf64_Half;  { Unused }
    vna_name: Elf64_Word;   { Dependency name string offset }
    vna_next: Elf64_Word;   { Offset in bytes to next vernaux entry }
  end;
  {$EXTERNALSYM Elf64_Vernaux}
  PElf64_Vernaux = ^Elf64_Vernaux;


{ Legal values for vna_flags.  }

(* Already declared and defined above
const
  VER_FLG_WEAK    = $2;     { Weak version identifier }
*)

{ Auxiliary vector.  }

{ This vector is normally only used by the program interpreter.  The
   usual definition in an ABI supplement uses the name auxv_t.  The
   vector is not usually defined in a standard <elf.h> file, but it
   can't hurt.  We rename it to avoid conflicts.  The sizes of these
   types are an arrangement between the exec server and the program
   interpreter, so we don't fully specify them here.  }

type
  Elf32_auxv_t = {packed} record
    case a_type: Integer of        { Entry type }
      0: (a_val: Longint);         { Integer value }
      1: (a_ptr: Pointer);         { Pointer value }
      2: (a_fcn: procedure);       { Function pointer value }
    end;
  {$EXTERNALSYM Elf32_auxv_t}
  PElf32_auxv_t = ^Elf32_auxv_t;

  Elf64_auxv_t = {packed} record
    case a_type: Longint of        { Entry type }
      0: (a_val: Longint);         { Integer value }
      1: (a_ptr: Pointer);         { Pointer value }
      2: (a_fcn: procedure);       { Function pointer value }
    end;
  {$EXTERNALSYM Elf64_auxv_t}
  PElf64_auxv_t = ^Elf64_auxv_t;

{ Legal values for a_type (entry type).  }
const
  AT_NULL       = 0;            { End of vector }
  {$EXTERNALSYM AT_NULL}
  AT_IGNORE     = 1;            { Entry should be ignored }
  {$EXTERNALSYM AT_IGNORE}
  AT_EXECFD     = 2;            { File descriptor of program }
  {$EXTERNALSYM AT_EXECFD}
  AT_PHDR       = 3;            { Program headers for program }
  {$EXTERNALSYM AT_PHDR}
  AT_PHENT      = 4;            { Size of program header entry }
  {$EXTERNALSYM AT_PHENT}
  AT_PHNUM      = 5;            { Number of program headers }
  {$EXTERNALSYM AT_PHNUM}
  AT_PAGESZ     = 6;            { System page size }
  {$EXTERNALSYM AT_PAGESZ}
  AT_BASE       = 7;            { Base address of interpreter }
  {$EXTERNALSYM AT_BASE}
  AT_FLAGS      = 8;            { Flags }
  {$EXTERNALSYM AT_FLAGS}
  AT_ENTRY      = 9;            { Entry point of program }
  {$EXTERNALSYM AT_ENTRY}
  AT_NOTELF     = 10;           { Program is not ELF }
  {$EXTERNALSYM AT_NOTELF}
  AT_UID        = 11;           { Real uid }
  {$EXTERNALSYM AT_UID}
  AT_EUID       = 12;           { Effective uid }
  {$EXTERNALSYM AT_EUID}
  AT_GID        = 13;           { Real gid }
  {$EXTERNALSYM AT_GID}
  AT_EGID       = 14;           { Effective gid }
  {$EXTERNALSYM AT_EGID}
  AT_CLKTCK     = 17;           { Frequency of times() }
  {$EXTERNALSYM AT_CLKTCK}

{ Some more special a_type values describing the hardware.  }
  AT_PLATFORM   = 15;           { String identifying platform.  }
  {$EXTERNALSYM AT_PLATFORM}
  AT_HWCAP      = 16;           { Machine dependent hints about
                                  processor capabilities.  }
  {$EXTERNALSYM AT_HWCAP}

{ This entry gives some information about the FPU initialization
   performed by the kernel.  }
  AT_FPUCW      = 18;           { Used FPU control word.  }
  {$EXTERNALSYM AT_FPUCW}


{ Note section contents.  Each entry in the note section begins with
   a header of a fixed form.  }

type
  Elf32_Nhdr = {packed} record
    n_namesz: Elf32_Word;       { Length of the note's name.  }
    n_descsz: Elf32_Word;       { Length of the note's descriptor.  }
    n_type: Elf32_Word;         { Type of the note.  }
  end;
  {$EXTERNALSYM Elf32_Nhdr}
  PElf32_Nhdr = ^Elf32_Nhdr;

type
  Elf64_Nhdr = {packed} record
    n_namesz: Elf64_Word;       { Length of the note's name.  }
    n_descsz: Elf64_Word;       { Length of the note's descriptor.  }
    n_type: Elf64_Word;         { Type of the note.  }
  end;
  {$EXTERNALSYM Elf64_Nhdr}
  PElf64_Nhdr = ^Elf64_Nhdr;

{ Known names of notes.  }

{ Solaris entries in the note section have this name.  }
const
  ELF_NOTE_SOLARIS    = 'SUNW Solaris';
  {$EXTERNALSYM ELF_NOTE_SOLARIS}

{ Note entries for GNU systems have this name.  }
  ELF_NOTE_GNU        = 'GNU';
  {$EXTERNALSYM ELF_NOTE_GNU}


{ Defined types of notes for Solaris.  }

{ Value of descriptor (one word) is desired pagesize for the binary.  }
  ELF_NOTE_PAGESIZE_HINT = 1;
  {$EXTERNALSYM ELF_NOTE_PAGESIZE_HINT}


{ Defined note types for GNU systems.  }

{ ABI information.  The descriptor consists of words:
   word 0: OS descriptor
   word 1: major version of the ABI
   word 2: minor version of the ABI
   word 3: subminor version of the ABI
}
  ELF_NOTE_ABI = 1;
  {$EXTERNALSYM ELF_NOTE_ABI}

{ Known OSes.  These value can appear in word 0 of an ELF_NOTE_ABI
   note section entry.  }
  ELF_NOTE_OS_LINUX     = 0;
  {$EXTERNALSYM ELF_NOTE_OS_LINUX}
  ELF_NOTE_OS_GNU       = 1;
  {$EXTERNALSYM ELF_NOTE_OS_GNU}
  ELF_NOTE_OS_SOLARIS2  = 2;
  {$EXTERNALSYM ELF_NOTE_OS_SOLARIS2}


{ Move records.  }
type
  Elf32_Move = {packed} record
    m_value: Elf32_Xword;     { Symbol value.  }
    m_info: Elf32_Word;       { Size and index.  }
    m_poffset: Elf32_Word;    { Symbol offset.  }
    m_repeat: Elf32_Half;     { Repeat count.  }
    m_stride: Elf32_Half;     { Stride info.  }
  end;
  {$EXTERNALSYM Elf32_Move}
  PElf32_Move = ^Elf32_Move;

type
  Elf64_Move = {packed} record
    m_value: Elf64_Xword;     { Symbol value.  }
    m_info: Elf64_Xword;      { Size and index.  }
    m_poffset: Elf64_Xword;   { Symbol offset.  }
    m_repeat: Elf64_Half;     { Repeat count.  }
    m_stride: Elf64_Half;     { Stride info.  }
  end;
  {$EXTERNALSYM Elf64_Move}
  PElf64_Move = ^Elf64_Move;

{ Macro to construct move records.  }
function ELF32_M_SYM(const info: Elf32_Word): Elf32_Word;
{$EXTERNALSYM ELF32_M_SYM}
function ELF32_M_SIZE(const info: Elf32_Word): Byte;
{$EXTERNALSYM ELF32_M_SIZE}
function ELF32_M_INFO(const sym: Elf32_Word; const size: Byte): Elf32_Word;
{$EXTERNALSYM ELF32_M_INFO}

function ELF64_M_SYM(const info: Elf64_Xword): Elf64_Xword;
{$EXTERNALSYM ELF64_M_SYM}
function ELF64_M_SIZE(const info: Elf64_Xword): Byte;
{$EXTERNALSYM ELF64_M_SIZE}
function ELF64_M_INFO(const sym: Elf64_Xword; const size: Byte): Elf64_Xword;
{$EXTERNALSYM ELF64_M_INFO}

{ Motorola 68k specific definitions.  }

{ Values for Elf32_Ehdr.e_flags.  }
const
  EF_CPU32 = $00810000;
  {$EXTERNALSYM EF_CPU32}

{ m68k relocs.  }

  R_68K_NONE        = 0;        { No reloc }
  {$EXTERNALSYM R_68K_NONE}
  R_68K_32          = 1;        { Direct 32 bit  }
  {$EXTERNALSYM R_68K_32}
  R_68K_16          = 2;        { Direct 16 bit  }
  {$EXTERNALSYM R_68K_16}
  R_68K_8           = 3;        { Direct 8 bit  }
  {$EXTERNALSYM R_68K_8}
  R_68K_PC32        = 4;        { PC relative 32 bit }
  {$EXTERNALSYM R_68K_PC32}
  R_68K_PC16        = 5;        { PC relative 16 bit }
  {$EXTERNALSYM R_68K_PC16}
  R_68K_PC8         = 6;        { PC relative 8 bit }
  {$EXTERNALSYM R_68K_PC8}
  R_68K_GOT32       = 7;        { 32 bit PC relative GOT entry }
  {$EXTERNALSYM R_68K_GOT32}
  R_68K_GOT16       = 8;        { 16 bit PC relative GOT entry }
  {$EXTERNALSYM R_68K_GOT16}
  R_68K_GOT8        = 9;        { 8 bit PC relative GOT entry }
  {$EXTERNALSYM R_68K_GOT8}
  R_68K_GOT32O      = 10;       { 32 bit GOT offset }
  {$EXTERNALSYM R_68K_GOT32O}
  R_68K_GOT16O      = 11;       { 16 bit GOT offset }
  {$EXTERNALSYM R_68K_GOT16O}
  R_68K_GOT8O       = 12;       { 8 bit GOT offset }
  {$EXTERNALSYM R_68K_GOT8O}
  R_68K_PLT32       = 13;       { 32 bit PC relative PLT address }
  {$EXTERNALSYM R_68K_PLT32}
  R_68K_PLT16       = 14;       { 16 bit PC relative PLT address }
  {$EXTERNALSYM R_68K_PLT16}
  R_68K_PLT8        = 15;       { 8 bit PC relative PLT address }
  {$EXTERNALSYM R_68K_PLT8}
  R_68K_PLT32O      = 16;       { 32 bit PLT offset }
  {$EXTERNALSYM R_68K_PLT32O}
  R_68K_PLT16O      = 17;       { 16 bit PLT offset }
  {$EXTERNALSYM R_68K_PLT16O}
  R_68K_PLT8O       = 18;       { 8 bit PLT offset }
  {$EXTERNALSYM R_68K_PLT8O}
  R_68K_COPY        = 19;       { Copy symbol at runtime }
  {$EXTERNALSYM R_68K_COPY}
  R_68K_GLOB_DAT    = 20;       { Create GOT entry }
  {$EXTERNALSYM R_68K_GLOB_DAT}
  R_68K_JMP_SLOT    = 21;       { Create PLT entry }
  {$EXTERNALSYM R_68K_JMP_SLOT}
  R_68K_RELATIVE    = 22;       { Adjust by program base }
  {$EXTERNALSYM R_68K_RELATIVE}
{ Keep this the last entry.  }
  R_68K_NUM = 23;
  {$EXTERNALSYM R_68K_NUM}

{ Intel 80386 specific definitions.  }

{ i386 relocs.  }

  R_386_NONE        = 0;        { No reloc }
  {$EXTERNALSYM R_386_NONE}
  R_386_32          = 1;        { Direct 32 bit  }
  {$EXTERNALSYM R_386_32}
  R_386_PC32        = 2;        { PC relative 32 bit }
  {$EXTERNALSYM R_386_PC32}
  R_386_GOT32       = 3;        { 32 bit GOT entry }
  {$EXTERNALSYM R_386_GOT32}
  R_386_PLT32       = 4;        { 32 bit PLT address }
  {$EXTERNALSYM R_386_PLT32}
  R_386_COPY        = 5;        { Copy symbol at runtime }
  {$EXTERNALSYM R_386_COPY}
  R_386_GLOB_DAT    = 6;        { Create GOT entry }
  {$EXTERNALSYM R_386_GLOB_DAT}
  R_386_JMP_SLOT    = 7;        { Create PLT entry }
  {$EXTERNALSYM R_386_JMP_SLOT}
  R_386_RELATIVE    = 8;        { Adjust by program base }
  {$EXTERNALSYM R_386_RELATIVE}
  R_386_GOTOFF      = 9;        { 32 bit offset to GOT }
  {$EXTERNALSYM R_386_GOTOFF}
  R_386_GOTPC       = 10;       { 32 bit PC relative offset to GOT }
  {$EXTERNALSYM R_386_GOTPC}
{ Keep this the last entry.  }
  R_386_NUM         = 11;
  {$EXTERNALSYM R_386_NUM}

{ SUN SPARC specific definitions.  }

{ Legal values for ST_TYPE subfield of st_info (symbol type).  }

  STT_REGISTER      = 13;       { Global register reserved to app. }
  {$EXTERNALSYM STT_REGISTER}

{ Values for Elf64_Ehdr.e_flags.  }

  EF_SPARCV9_MM	    = 3;
  {$EXTERNALSYM EF_SPARCV9_MM}
  EF_SPARCV9_TSO    = 0;
  {$EXTERNALSYM EF_SPARCV9_TSO}
  EF_SPARCV9_PSO    = 1;
  {$EXTERNALSYM EF_SPARCV9_PSO}
  EF_SPARCV9_RMO    = 2;
  {$EXTERNALSYM EF_SPARCV9_RMO}
  EF_SPARC_LEDATA   = $800000; { little endian data }
  {$EXTERNALSYM EF_SPARC_LEDATA}
  EF_SPARC_EXT_MASK = $FFFF00;
  {$EXTERNALSYM EF_SPARC_EXT_MASK}
  EF_SPARC_32PLUS   = $000100; { generic V8+ features }
  {$EXTERNALSYM EF_SPARC_32PLUS}
  EF_SPARC_SUN_US1  = $000200; { Sun UltraSPARC1 extensions }
  {$EXTERNALSYM EF_SPARC_SUN_US1}
  EF_SPARC_HAL_R1   = $000400; { HAL R1 extensions }
  {$EXTERNALSYM EF_SPARC_HAL_R1}
  EF_SPARC_SUN_US3  = $000800; { Sun UltraSPARCIII extensions }
  {$EXTERNALSYM EF_SPARC_SUN_US3}

{ SPARC relocs.  }

  R_SPARC_NONE      = 0;        { No reloc }
  {$EXTERNALSYM R_SPARC_NONE}
  R_SPARC_8         = 1;        { Direct 8 bit }
  {$EXTERNALSYM R_SPARC_8}
  R_SPARC_16        = 2;        { Direct 16 bit }
  {$EXTERNALSYM R_SPARC_16}
  R_SPARC_32        = 3;        { Direct 32 bit }
  {$EXTERNALSYM R_SPARC_32}
  R_SPARC_DISP8     = 4;        { PC relative 8 bit }
  {$EXTERNALSYM R_SPARC_DISP8}
  R_SPARC_DISP16    = 5;        { PC relative 16 bit }
  {$EXTERNALSYM R_SPARC_DISP16}
  R_SPARC_DISP32    = 6;        { PC relative 32 bit }
  {$EXTERNALSYM R_SPARC_DISP32}
  R_SPARC_WDISP30   = 7;        { PC relative 30 bit shifted }
  {$EXTERNALSYM R_SPARC_WDISP30}
  R_SPARC_WDISP22   = 8;        { PC relative 22 bit shifted }
  {$EXTERNALSYM R_SPARC_WDISP22}
  R_SPARC_HI22      = 9;        { High 22 bit }
  {$EXTERNALSYM R_SPARC_HI22}
  R_SPARC_22        = 10;       { Direct 22 bit }
  {$EXTERNALSYM R_SPARC_22}
  R_SPARC_13        = 11;       { Direct 13 bit }
  {$EXTERNALSYM R_SPARC_13}
  R_SPARC_LO10      = 12;       { Truncated 10 bit }
  {$EXTERNALSYM R_SPARC_LO10}
  R_SPARC_GOT10     = 13;       { Truncated 10 bit GOT entry }
  {$EXTERNALSYM R_SPARC_GOT10}
  R_SPARC_GOT13     = 14;       { 13 bit GOT entry }
  {$EXTERNALSYM R_SPARC_GOT13}
  R_SPARC_GOT22     = 15;       { 22 bit GOT entry shifted }
  {$EXTERNALSYM R_SPARC_GOT22}
  R_SPARC_PC10      = 16;       { PC relative 10 bit truncated }
  {$EXTERNALSYM R_SPARC_PC10}
  R_SPARC_PC22      = 17;       { PC relative 22 bit shifted }
  {$EXTERNALSYM R_SPARC_PC22}
  R_SPARC_WPLT30    = 18;       { 30 bit PC relative PLT address }
  {$EXTERNALSYM R_SPARC_WPLT30}
  R_SPARC_COPY      = 19;       { Copy symbol at runtime }
  {$EXTERNALSYM R_SPARC_COPY}
  R_SPARC_GLOB_DAT  = 20;       { Create GOT entry }
  {$EXTERNALSYM R_SPARC_GLOB_DAT}
  R_SPARC_JMP_SLOT  = 21;       { Create PLT entry }
  {$EXTERNALSYM R_SPARC_JMP_SLOT}
  R_SPARC_RELATIVE  = 22;       { Adjust by program base }
  {$EXTERNALSYM R_SPARC_RELATIVE}
  R_SPARC_UA32      = 23;       { Direct 32 bit unaligned }
  {$EXTERNALSYM R_SPARC_UA32}

{ Additional Sparc64 relocs.  }

  R_SPARC_PLT32     = 24;       { Direct 32 bit ref to PLT entry }
  {$EXTERNALSYM R_SPARC_PLT32}
  R_SPARC_HIPLT22   = 25;       { High 22 bit PLT entry }
  {$EXTERNALSYM R_SPARC_HIPLT22}
  R_SPARC_LOPLT10   = 26;       { Truncated 10 bit PLT entry }
  {$EXTERNALSYM R_SPARC_LOPLT10}
  R_SPARC_PCPLT32   = 27;       { PC rel 32 bit ref to PLT entry }
  {$EXTERNALSYM R_SPARC_PCPLT32}
  R_SPARC_PCPLT22   = 28;       { PC rel high 22 bit PLT entry }
  {$EXTERNALSYM R_SPARC_PCPLT22}
  R_SPARC_PCPLT10   = 29;       { PC rel trunc 10 bit PLT entry }
  {$EXTERNALSYM R_SPARC_PCPLT10}
  R_SPARC_10        = 30;       { Direct 10 bit }
  {$EXTERNALSYM R_SPARC_10}
  R_SPARC_11        = 31;       { Direct 11 bit }
  {$EXTERNALSYM R_SPARC_11}
  R_SPARC_64        = 32;       { Direct 64 bit }
  {$EXTERNALSYM R_SPARC_64}
  R_SPARC_OLO10     = 33;       { 10bit with secondary 13bit addend }
  {$EXTERNALSYM R_SPARC_OLO10}
  R_SPARC_HH22      = 34;       { Top 22 bits of direct 64 bit }
  {$EXTERNALSYM R_SPARC_HH22}
  R_SPARC_HM10      = 35;       { High middle 10 bits of ... }
  {$EXTERNALSYM R_SPARC_HM10}
  R_SPARC_LM22      = 36;       { Low middle 22 bits of ... }
  {$EXTERNALSYM R_SPARC_LM22}
  R_SPARC_PC_HH22   = 37;       { Top 22 bits of pc rel 64 bit }
  {$EXTERNALSYM R_SPARC_PC_HH22}
  R_SPARC_PC_HM10   = 38;       { High middle 10 bit of ... }
  {$EXTERNALSYM R_SPARC_PC_HM10}
  R_SPARC_PC_LM22   = 39;       { Low miggle 22 bits of ... }
  {$EXTERNALSYM R_SPARC_PC_LM22}
  R_SPARC_WDISP16   = 40;       { PC relative 16 bit shifted }
  {$EXTERNALSYM R_SPARC_WDISP16}
  R_SPARC_WDISP19   = 41;       { PC relative 19 bit shifted }
  {$EXTERNALSYM R_SPARC_WDISP19}
  R_SPARC_7         = 43;       { Direct 7 bit }
  {$EXTERNALSYM R_SPARC_7}
  R_SPARC_5         = 44;       { Direct 5 bit }
  {$EXTERNALSYM R_SPARC_5}
  R_SPARC_6         = 45;       { Direct 6 bit }
  {$EXTERNALSYM R_SPARC_6}
  R_SPARC_DISP64    = 46;       { PC relative 64 bit }
  {$EXTERNALSYM R_SPARC_DISP64}
  R_SPARC_PLT64     = 47;       { Direct 64 bit ref to PLT entry }
  {$EXTERNALSYM R_SPARC_PLT64}
  R_SPARC_HIX22     = 48;       { High 22 bit complemented }
  {$EXTERNALSYM R_SPARC_HIX22}
  R_SPARC_LOX10     = 49;       { Truncated 11 bit complemented }
  {$EXTERNALSYM R_SPARC_LOX10}
  R_SPARC_H44       = 50;       { Direct high 12 of 44 bit }
  {$EXTERNALSYM R_SPARC_H44}
  R_SPARC_M44       = 51;       { Direct mid 22 of 44 bit }
  {$EXTERNALSYM R_SPARC_M44}
  R_SPARC_L44       = 52;       { Direct low 10 of 44 bit }
  {$EXTERNALSYM R_SPARC_L44}
  R_SPARC_REGISTER  = 53;       { Global register usage }
  {$EXTERNALSYM R_SPARC_REGISTER}
  R_SPARC_UA64      = 54;       { Direct 64 bit unaligned }
  {$EXTERNALSYM R_SPARC_UA64}
  R_SPARC_UA16      = 55;       { Direct 16 bit unaligned }
  {$EXTERNALSYM R_SPARC_UA16}
{ Keep this the last entry.  }
  R_SPARC_NUM       = 56;
  {$EXTERNALSYM R_SPARC_NUM}

{ For Sparc64, legal values for d_tag of Elf64_Dyn.  }

  DT_SPARC_REGISTER = $70000001;
  {$EXTERNALSYM DT_SPARC_REGISTER}
  DT_SPARC_NUM      = 2;
  {$EXTERNALSYM DT_SPARC_NUM}

{ Bits present in AT_HWCAP, primarily for Sparc32.  }

  HWCAP_SPARC_FLUSH  = 1;       { The cpu supports flush insn.  }
  {$EXTERNALSYM HWCAP_SPARC_FLUSH}
  HWCAP_SPARC_STBAR  = 2;
  {$EXTERNALSYM HWCAP_SPARC_STBAR}
  HWCAP_SPARC_SWAP   = 4;
  {$EXTERNALSYM HWCAP_SPARC_SWAP}
  HWCAP_SPARC_MULDIV = 8;
  {$EXTERNALSYM HWCAP_SPARC_MULDIV}
  HWCAP_SPARC_V9     = 16;      { The cpu is v9, so v8plus is ok.  }
  {$EXTERNALSYM HWCAP_SPARC_V9}

{ MIPS R3000 specific definitions.  }

{ Legal values for e_flags field of Elf32_Ehdr.  }

  EF_MIPS_NOREORDER    = 1;         { A .noreorder directive was used }
  {$EXTERNALSYM EF_MIPS_NOREORDER}
  EF_MIPS_PIC          = 2;         { Contains PIC code }
  {$EXTERNALSYM EF_MIPS_PIC}
  EF_MIPS_CPIC         = 4;         { Uses PIC calling sequence }
  {$EXTERNALSYM EF_MIPS_CPIC}
  EF_MIPS_XGOT         = 8;
  {$EXTERNALSYM EF_MIPS_XGOT}
  EF_MIPS_64BIT_WHIRL  = 16;
  {$EXTERNALSYM EF_MIPS_64BIT_WHIRL}
  EF_MIPS_ABI2         = 32;
  {$EXTERNALSYM EF_MIPS_ABI2}
  EF_MIPS_ABI_ON32     = 64;
  {$EXTERNALSYM EF_MIPS_ABI_ON32}
  EF_MIPS_ARCH         = $f0000000; { MIPS architecture level }
  {$EXTERNALSYM EF_MIPS_ARCH}

{ Legal values for MIPS architecture level.  }

  EF_MIPS_ARCH_1 = $00000000;   { -mips1 code.  }
  {$EXTERNALSYM EF_MIPS_ARCH_1}
  EF_MIPS_ARCH_2 = $10000000;   { -mips2 code.  }
  {$EXTERNALSYM EF_MIPS_ARCH_2}
  EF_MIPS_ARCH_3 = $20000000;   { -mips3 code.  }
  {$EXTERNALSYM EF_MIPS_ARCH_3}
  EF_MIPS_ARCH_4 = $30000000;   { -mips4 code.  }
  {$EXTERNALSYM EF_MIPS_ARCH_4}
  EF_MIPS_ARCH_5 = $40000000;   { -mips5 code.  }
  {$EXTERNALSYM EF_MIPS_ARCH_5}

{ The following are non-official names and should not be used.  }

  E_MIPS_ARCH_1 = $00000000;    { -mips1 code.  }
  {$EXTERNALSYM E_MIPS_ARCH_1}
  E_MIPS_ARCH_2 = $10000000;    { -mips2 code.  }
  {$EXTERNALSYM E_MIPS_ARCH_2}
  E_MIPS_ARCH_3 = $20000000;    { -mips3 code.  }
  {$EXTERNALSYM E_MIPS_ARCH_3}
  E_MIPS_ARCH_4 = $30000000;    { -mips4 code.  }
  {$EXTERNALSYM E_MIPS_ARCH_4}
  E_MIPS_ARCH_5 = $40000000;    { -mips5 code.  }
  {$EXTERNALSYM E_MIPS_ARCH_5}

{ Special section indices.  }

  SHN_MIPS_ACOMMON    = $ff00;    { Allocated common symbols }
  {$EXTERNALSYM SHN_MIPS_ACOMMON}
  SHN_MIPS_TEXT       = $ff01;    { Allocated test symbols.  }
  {$EXTERNALSYM SHN_MIPS_TEXT}
  SHN_MIPS_DATA       = $ff02;    { Allocated data symbols.  }
  {$EXTERNALSYM SHN_MIPS_DATA}
  SHN_MIPS_SCOMMON    = $ff03;    { Small common symbols }
  {$EXTERNALSYM SHN_MIPS_SCOMMON}
  SHN_MIPS_SUNDEFINED = $ff04;    { Small undefined symbols }
  {$EXTERNALSYM SHN_MIPS_SUNDEFINED}

{ Legal values for sh_type field of Elf32_Shdr.  }

  SHT_MIPS_LIBLIST          = $70000000; { Shared objects used in link }
  {$EXTERNALSYM SHT_MIPS_LIBLIST}
  SHT_MIPS_MSYM             = $70000001;
  {$EXTERNALSYM SHT_MIPS_MSYM}
  SHT_MIPS_CONFLICT         = $70000002; { Conflicting symbols }
  {$EXTERNALSYM SHT_MIPS_CONFLICT}
  SHT_MIPS_GPTAB            = $70000003; { Global data area sizes }
  {$EXTERNALSYM SHT_MIPS_GPTAB}
  SHT_MIPS_UCODE            = $70000004; { Reserved for SGI/MIPS compilers }
  {$EXTERNALSYM SHT_MIPS_UCODE}
  SHT_MIPS_DEBUG            = $70000005; { MIPS ECOFF debugging information}
  {$EXTERNALSYM SHT_MIPS_DEBUG}
  SHT_MIPS_REGINFO          = $70000006; { Register usage information }
  {$EXTERNALSYM SHT_MIPS_REGINFO}
  SHT_MIPS_PACKAGE          = $70000007;
  {$EXTERNALSYM SHT_MIPS_PACKAGE}
  SHT_MIPS_PACKSYM          = $70000008;
  {$EXTERNALSYM SHT_MIPS_PACKSYM}
  SHT_MIPS_RELD             = $70000009;
  {$EXTERNALSYM SHT_MIPS_RELD}
  SHT_MIPS_IFACE            = $7000000b;
  {$EXTERNALSYM SHT_MIPS_IFACE}
  SHT_MIPS_CONTENT          = $7000000c;
  {$EXTERNALSYM SHT_MIPS_CONTENT}
  SHT_MIPS_OPTIONS          = $7000000d; { Miscellaneous options.  }
  {$EXTERNALSYM SHT_MIPS_OPTIONS}
  SHT_MIPS_SHDR             = $70000010;
  {$EXTERNALSYM SHT_MIPS_SHDR}
  SHT_MIPS_FDESC            = $70000011;
  {$EXTERNALSYM SHT_MIPS_FDESC}
  SHT_MIPS_EXTSYM           = $70000012;
  {$EXTERNALSYM SHT_MIPS_EXTSYM}
  SHT_MIPS_DENSE            = $70000013;
  {$EXTERNALSYM SHT_MIPS_DENSE}
  SHT_MIPS_PDESC            = $70000014;
  {$EXTERNALSYM SHT_MIPS_PDESC}
  SHT_MIPS_LOCSYM           = $70000015;
  {$EXTERNALSYM SHT_MIPS_LOCSYM}
  SHT_MIPS_AUXSYM           = $70000016;
  {$EXTERNALSYM SHT_MIPS_AUXSYM}
  SHT_MIPS_OPTSYM           = $70000017;
  {$EXTERNALSYM SHT_MIPS_OPTSYM}
  SHT_MIPS_LOCSTR           = $70000018;
  {$EXTERNALSYM SHT_MIPS_LOCSTR}
  SHT_MIPS_LINE             = $70000019;
  {$EXTERNALSYM SHT_MIPS_LINE}
  SHT_MIPS_RFDESC           = $7000001a;
  {$EXTERNALSYM SHT_MIPS_RFDESC}
  SHT_MIPS_DELTASYM         = $7000001b;
  {$EXTERNALSYM SHT_MIPS_DELTASYM}
  SHT_MIPS_DELTAINST        = $7000001c;
  {$EXTERNALSYM SHT_MIPS_DELTAINST}
  SHT_MIPS_DELTACLASS       = $7000001d;
  {$EXTERNALSYM SHT_MIPS_DELTACLASS}
  SHT_MIPS_DWARF            = $7000001e; { DWARF debugging information.  }
  {$EXTERNALSYM SHT_MIPS_DWARF}
  SHT_MIPS_DELTADECL        = $7000001f;
  {$EXTERNALSYM SHT_MIPS_DELTADECL}
  SHT_MIPS_SYMBOL_LIB       = $70000020;
  {$EXTERNALSYM SHT_MIPS_SYMBOL_LIB}
  SHT_MIPS_EVENTS           = $70000021; { Event section.  }
  {$EXTERNALSYM SHT_MIPS_EVENTS}
  SHT_MIPS_TRANSLATE        = $70000022;
  {$EXTERNALSYM SHT_MIPS_TRANSLATE}
  SHT_MIPS_PIXIE            = $70000023;
  {$EXTERNALSYM SHT_MIPS_PIXIE}
  SHT_MIPS_XLATE            = $70000024;
  {$EXTERNALSYM SHT_MIPS_XLATE}
  SHT_MIPS_XLATE_DEBUG      = $70000025;
  {$EXTERNALSYM SHT_MIPS_XLATE_DEBUG}
  SHT_MIPS_WHIRL            = $70000026;
  {$EXTERNALSYM SHT_MIPS_WHIRL}
  SHT_MIPS_EH_REGION        = $70000027;
  {$EXTERNALSYM SHT_MIPS_EH_REGION}
  SHT_MIPS_XLATE_OLD        = $70000028;
  {$EXTERNALSYM SHT_MIPS_XLATE_OLD}
  SHT_MIPS_PDR_EXCEPTION    = $70000029;
  {$EXTERNALSYM SHT_MIPS_PDR_EXCEPTION}

{ Legal values for sh_flags field of Elf32_Shdr.  }

  SHF_MIPS_GPREL   = $10000000; { Must be part of global data area }
  {$EXTERNALSYM SHF_MIPS_GPREL}
  SHF_MIPS_MERGE   = $20000000;
  {$EXTERNALSYM SHF_MIPS_MERGE}
  SHF_MIPS_ADDR    = $40000000;
  {$EXTERNALSYM SHF_MIPS_ADDR}
  SHF_MIPS_STRINGS = $80000000;
  {$EXTERNALSYM SHF_MIPS_STRINGS}
  SHF_MIPS_NOSTRIP = $08000000;
  {$EXTERNALSYM SHF_MIPS_NOSTRIP}
  SHF_MIPS_LOCAL   = $04000000;
  {$EXTERNALSYM SHF_MIPS_LOCAL}
  SHF_MIPS_NAMES   = $02000000;
  {$EXTERNALSYM SHF_MIPS_NAMES}
  SHF_MIPS_NODUPE  = $01000000;
  {$EXTERNALSYM SHF_MIPS_NODUPE}


{ Symbol tables.  }

{ MIPS specific values for `st_other'.  }
  STO_MIPS_DEFAULT         = $0;
  {$EXTERNALSYM STO_MIPS_DEFAULT}
  STO_MIPS_INTERNAL        = $1;
  {$EXTERNALSYM STO_MIPS_INTERNAL}
  STO_MIPS_HIDDEN          = $2;
  {$EXTERNALSYM STO_MIPS_HIDDEN}
  STO_MIPS_PROTECTED       = $3;
  {$EXTERNALSYM STO_MIPS_PROTECTED}
  STO_MIPS_SC_ALIGN_UNUSED = $ff;
  {$EXTERNALSYM STO_MIPS_SC_ALIGN_UNUSED}

{ MIPS specific values for `st_info'.  }
  STB_MIPS_SPLIT_COMMON = 13;
  {$EXTERNALSYM STB_MIPS_SPLIT_COMMON}

{ Entries found in sections of type SHT_MIPS_GPTAB.  }

type
  gt_header_record = {packed} record
    gt_current_g_value: Elf32_Word; { -G value used for compilation }
    gt_unused: Elf32_Word;          { Not used }
  end;
  // Used anonymously in header file.

  gt_entry_record = {packed} record
    gt_g_value: Elf32_Word;         { If this value were used for -G }
    gt_bytes: Elf32_Word;           { This many bytes would be used }
  end;
  // Used anonymously in header file.

  Elf32_gptab = {packed} record
    case Integer of
      0: (gt_header: gt_header_record); { First entry in section }
      1: (gt_entry: gt_entry_record); { Subsequent entries in section }
    end;
  {$EXTERNALSYM Elf32_gptab}
  PElf32_gptab = ^Elf32_gptab;

{ Entry found in sections of type SHT_MIPS_REGINFO.  }

type
  Elf32_RegInfo = {packed} record
    ri_gprmask: Elf32_Word;      { General registers used }
    ri_cprmask: packed array[0..4-1] of Elf32_Word; { Coprocessor registers used }
    ri_gp_value: Elf32_Sword;    { $gp register value }
  end;
  {$EXTERNALSYM Elf32_RegInfo}
  PElf32_RegInfo = ^Elf32_RegInfo;

{ Entries found in sections of type SHT_MIPS_OPTIONS.  }

type
  Elf_Options = {packed} record
    kind: Byte;              { Determines interpretation of the
                               variable part of descriptor.  }
    size: Byte;              { Size of descriptor, including header.  }
    section: Elf32_Section;  { Section header index of section affected,
                               0 for global options.  }
    info: Elf32_Word;        { Kind-specific information.  }
  end;
  {$EXTERNALSYM Elf_Options}
  PElf_Options = ^Elf_Options;

{ Values for `kind' field in Elf_Options.  }
const
  ODK_NULL       = 0;   { Undefined.  }
  {$EXTERNALSYM ODK_NULL}
  ODK_REGINFO    = 1;   { Register usage information.  }
  {$EXTERNALSYM ODK_REGINFO}
  ODK_EXCEPTIONS = 2;   { Exception processing options.  }
  {$EXTERNALSYM ODK_EXCEPTIONS}
  ODK_PAD        = 3;   { Section padding options.  }
  {$EXTERNALSYM ODK_PAD}
  ODK_HWPATCH    = 4;   { Hardware workarounds performed }
  {$EXTERNALSYM ODK_HWPATCH}
  ODK_FILL       = 5;   { record the fill value used by the linker. }
  {$EXTERNALSYM ODK_FILL}
  ODK_TAGS       = 6;   { reserve space for desktop tools to write. }
  {$EXTERNALSYM ODK_TAGS}
  ODK_HWAND      = 7;   { HW workarounds.  'AND' bits when merging. }
  {$EXTERNALSYM ODK_HWAND}
  ODK_HWOR       = 8;   { HW workarounds.  'OR' bits when merging.  }
  {$EXTERNALSYM ODK_HWOR}

{ Values for `info' in Elf_Options for ODK_EXCEPTIONS entries.  }

  OEX_FPU_MIN    = $1f;         { FPE's which MUST be enabled.  }
  {$EXTERNALSYM OEX_FPU_MIN}
  OEX_FPU_MAX    = $1f00;       { FPE's which MAY be enabled.  }
  {$EXTERNALSYM OEX_FPU_MAX}
  OEX_PAGE0      = $10000;      { page zero must be mapped.  }
  {$EXTERNALSYM OEX_PAGE0}
  OEX_SMM        = $20000;      { Force sequential memory mode?  }
  {$EXTERNALSYM OEX_SMM}
  OEX_FPDBUG     = $40000;      { Force floating point debug mode?  }
  {$EXTERNALSYM OEX_FPDBUG}
  OEX_PRECISEFP  = OEX_FPDBUG;
  {$EXTERNALSYM OEX_PRECISEFP}
  OEX_DISMISS    = $80000;      { Dismiss invalid address faults?  }
  {$EXTERNALSYM OEX_DISMISS}

  OEX_FPU_INVAL = $10;
  {$EXTERNALSYM OEX_FPU_INVAL}
  OEX_FPU_DIV0  = $08;
  {$EXTERNALSYM OEX_FPU_DIV0}
  OEX_FPU_OFLO  = $04;
  {$EXTERNALSYM OEX_FPU_OFLO}
  OEX_FPU_UFLO  = $02;
  {$EXTERNALSYM OEX_FPU_UFLO}
  OEX_FPU_INEX  = $01;
  {$EXTERNALSYM OEX_FPU_INEX}

{ Masks for `info' in Elf_Options for an ODK_HWPATCH entry.  }

  OHW_R4KEOP    = $1;   { R4000 end-of-page patch.  }
  {$EXTERNALSYM OHW_R4KEOP}
  OHW_R8KPFETCH = $2;   { may need R8000 prefetch patch.  }
  {$EXTERNALSYM OHW_R8KPFETCH}
  OHW_R5KEOP    = $4;   { R5000 end-of-page patch.  }
  {$EXTERNALSYM OHW_R5KEOP}
  OHW_R5KCVTL   = $8;   { R5000 cvt.[ds].l bug.  clean=1.  }
  {$EXTERNALSYM OHW_R5KCVTL}

  OPAD_PREFIX   = $1;
  {$EXTERNALSYM OPAD_PREFIX}
  OPAD_POSTFIX  = $2;
  {$EXTERNALSYM OPAD_POSTFIX}
  OPAD_SYMBOL   = $4;
  {$EXTERNALSYM OPAD_SYMBOL}

{ Entry found in `.options' section.  }

type
  Elf_Options_Hw = {packed} record
    hwp_flags1: Elf32_Word;    { Extra flags.  }
    hwp_flags2: Elf32_Word;    { Extra flags.  }
  end;
  {$EXTERNALSYM Elf_Options_Hw}
  PElf_Options_Hw = ^Elf_Options_Hw;

{ Masks for `info' in ElfOptions for ODK_HWAND and ODK_HWOR entries.  }

const
  OHWA0_R4KEOP_CHECKED  = $00000001;
  {$EXTERNALSYM OHWA0_R4KEOP_CHECKED}
  OHWA1_R4KEOP_CLEAN    = $00000002;
  {$EXTERNALSYM OHWA1_R4KEOP_CLEAN}

{ MIPS relocs.  }

  R_MIPS_NONE           = 0;    { No reloc }
  {$EXTERNALSYM R_MIPS_NONE}
  R_MIPS_16             = 1;    { Direct 16 bit }
  {$EXTERNALSYM R_MIPS_16}
  R_MIPS_32             = 2;    { Direct 32 bit }
  {$EXTERNALSYM R_MIPS_32}
  R_MIPS_REL32          = 3;    { PC relative 32 bit }
  {$EXTERNALSYM R_MIPS_REL32}
  R_MIPS_26             = 4;    { Direct 26 bit shifted }
  {$EXTERNALSYM R_MIPS_26}
  R_MIPS_HI16           = 5;    { High 16 bit }
  {$EXTERNALSYM R_MIPS_HI16}
  R_MIPS_LO16           = 6;    { Low 16 bit }
  {$EXTERNALSYM R_MIPS_LO16}
  R_MIPS_GPREL16        = 7;    { GP relative 16 bit }
  {$EXTERNALSYM R_MIPS_GPREL16}
  R_MIPS_LITERAL        = 8;    { 16 bit literal entry }
  {$EXTERNALSYM R_MIPS_LITERAL}
  R_MIPS_GOT16          = 9;    { 16 bit GOT entry }
  {$EXTERNALSYM R_MIPS_GOT16}
  R_MIPS_PC16           = 10;   { PC relative 16 bit }
  {$EXTERNALSYM R_MIPS_PC16}
  R_MIPS_CALL16	        = 11;   { 16 bit GOT entry for function }
  {$EXTERNALSYM R_MIPS_CALL16}
  R_MIPS_GPREL32        = 12;   { GP relative 32 bit }
  {$EXTERNALSYM R_MIPS_GPREL32}

  R_MIPS_SHIFT5         = 16;
  {$EXTERNALSYM R_MIPS_SHIFT5}
  R_MIPS_SHIFT6         = 17;
  {$EXTERNALSYM R_MIPS_SHIFT6}
  R_MIPS_64             = 18;
  {$EXTERNALSYM R_MIPS_64}
  R_MIPS_GOT_DISP       = 19;
  {$EXTERNALSYM R_MIPS_GOT_DISP}
  R_MIPS_GOT_PAGE       = 20;
  {$EXTERNALSYM R_MIPS_GOT_PAGE}
  R_MIPS_GOT_OFST       = 21;
  {$EXTERNALSYM R_MIPS_GOT_OFST}
  R_MIPS_GOT_HI16       = 22;
  {$EXTERNALSYM R_MIPS_GOT_HI16}
  R_MIPS_GOT_LO16       = 23;
  {$EXTERNALSYM R_MIPS_GOT_LO16}
  R_MIPS_SUB            = 24;
  {$EXTERNALSYM R_MIPS_SUB}
  R_MIPS_INSERT_A       = 25;
  {$EXTERNALSYM R_MIPS_INSERT_A}
  R_MIPS_INSERT_B       = 26;
  {$EXTERNALSYM R_MIPS_INSERT_B}
  R_MIPS_DELETE         = 27;
  {$EXTERNALSYM R_MIPS_DELETE}
  R_MIPS_HIGHER         = 28;
  {$EXTERNALSYM R_MIPS_HIGHER}
  R_MIPS_HIGHEST        = 29;
  {$EXTERNALSYM R_MIPS_HIGHEST}
  R_MIPS_CALL_HI16      = 30;
  {$EXTERNALSYM R_MIPS_CALL_HI16}
  R_MIPS_CALL_LO16      = 31;
  {$EXTERNALSYM R_MIPS_CALL_LO16}
  R_MIPS_SCN_DISP       = 32;
  {$EXTERNALSYM R_MIPS_SCN_DISP}
  R_MIPS_REL16          = 33;
  {$EXTERNALSYM R_MIPS_REL16}
  R_MIPS_ADD_IMMEDIATE  = 34;
  {$EXTERNALSYM R_MIPS_ADD_IMMEDIATE}
  R_MIPS_PJUMP          = 35;
  {$EXTERNALSYM R_MIPS_PJUMP}
  R_MIPS_RELGOT         = 36;
  {$EXTERNALSYM R_MIPS_RELGOT}
  R_MIPS_JALR           = 37;
  {$EXTERNALSYM R_MIPS_JALR}
{ Keep this the last entry.  }
  R_MIPS_NUM            = 38;
  {$EXTERNALSYM R_MIPS_NUM}

{ Legal values for p_type field of Elf32_Phdr.  }

  PT_MIPS_REGINFO = $70000000;  { Register usage information }
  {$EXTERNALSYM PT_MIPS_REGINFO}
  PT_MIPS_RTPROC  = $70000001;  { Runtime procedure table. }
  {$EXTERNALSYM PT_MIPS_RTPROC}
  PT_MIPS_OPTIONS = $70000002;
  {$EXTERNALSYM PT_MIPS_OPTIONS}

{ Special program header types.  }

  PF_MIPS_LOCAL = $10000000;
  {$EXTERNALSYM PF_MIPS_LOCAL}

{ Legal values for d_tag field of Elf32_Dyn.  }

  DT_MIPS_RLD_VERSION            = $70000001;   { Runtime linker interface version }
  {$EXTERNALSYM DT_MIPS_RLD_VERSION}
  DT_MIPS_TIME_STAMP             = $70000002;   { Timestamp }
  {$EXTERNALSYM DT_MIPS_TIME_STAMP}
  DT_MIPS_ICHECKSUM              = $70000003;   { Checksum }
  {$EXTERNALSYM DT_MIPS_ICHECKSUM}
  DT_MIPS_IVERSION               = $70000004;   { Version string (string tbl index) }
  {$EXTERNALSYM DT_MIPS_IVERSION}
  DT_MIPS_FLAGS                  = $70000005;   { Flags }
  {$EXTERNALSYM DT_MIPS_FLAGS}
  DT_MIPS_BASE_ADDRESS           = $70000006;   { Base address }
  {$EXTERNALSYM DT_MIPS_BASE_ADDRESS}
  DT_MIPS_MSYM                   = $70000007;
  {$EXTERNALSYM DT_MIPS_MSYM}
  DT_MIPS_CONFLICT               = $70000008;   { Address of CONFLICT section }
  {$EXTERNALSYM DT_MIPS_CONFLICT}
  DT_MIPS_LIBLIST                = $70000009;   { Address of LIBLIST section }
  {$EXTERNALSYM DT_MIPS_LIBLIST}
  DT_MIPS_LOCAL_GOTNO            = $7000000a;   { Number of local GOT entries }
  {$EXTERNALSYM DT_MIPS_LOCAL_GOTNO}
  DT_MIPS_CONFLICTNO             = $7000000b;   { Number of CONFLICT entries }
  {$EXTERNALSYM DT_MIPS_CONFLICTNO}
  DT_MIPS_LIBLISTNO              = $70000010;   { Number of LIBLIST entries }
  {$EXTERNALSYM DT_MIPS_LIBLISTNO}
  DT_MIPS_SYMTABNO               = $70000011;   { Number of DYNSYM entries }
  {$EXTERNALSYM DT_MIPS_SYMTABNO}
  DT_MIPS_UNREFEXTNO             = $70000012;   { First external DYNSYM }
  {$EXTERNALSYM DT_MIPS_UNREFEXTNO}
  DT_MIPS_GOTSYM                 = $70000013;   { First GOT entry in DYNSYM }
  {$EXTERNALSYM DT_MIPS_GOTSYM}
  DT_MIPS_HIPAGENO               = $70000014;   { Number of GOT page table entries }
  {$EXTERNALSYM DT_MIPS_HIPAGENO}
  DT_MIPS_RLD_MAP                = $70000016;   { Address of run time loader map.  }
  {$EXTERNALSYM DT_MIPS_RLD_MAP}
  DT_MIPS_DELTA_CLASS            = $70000017;   { Delta C++ class definition.  }
  {$EXTERNALSYM DT_MIPS_DELTA_CLASS}
  DT_MIPS_DELTA_CLASS_NO         = $70000018;   { Number of entries in
                                                  DT_MIPS_DELTA_CLASS.  }
  {$EXTERNALSYM DT_MIPS_DELTA_CLASS_NO}
  DT_MIPS_DELTA_INSTANCE         = $70000019;   { Delta C++ class instances.  }
  {$EXTERNALSYM DT_MIPS_DELTA_INSTANCE}
  DT_MIPS_DELTA_INSTANCE_NO      = $7000001a;   { Number of entries in
                                                  DT_MIPS_DELTA_INSTANCE.  }
  {$EXTERNALSYM DT_MIPS_DELTA_INSTANCE_NO}
  DT_MIPS_DELTA_RELOC            = $7000001b;   { Delta relocations.  }
  {$EXTERNALSYM DT_MIPS_DELTA_RELOC}
  DT_MIPS_DELTA_RELOC_NO         = $7000001c;   { Number of entries in
                                                  DT_MIPS_DELTA_RELOC.  }
  {$EXTERNALSYM DT_MIPS_DELTA_RELOC_NO}
  DT_MIPS_DELTA_SYM              = $7000001d;   { Delta symbols that Delta
                                                  relocations refer to.  }
  {$EXTERNALSYM DT_MIPS_DELTA_SYM}
  DT_MIPS_DELTA_SYM_NO           = $7000001e;   { Number of entries in
                                                  DT_MIPS_DELTA_SYM.  }
  {$EXTERNALSYM DT_MIPS_DELTA_SYM_NO}
  DT_MIPS_DELTA_CLASSSYM         = $70000020;   { Delta symbols that hold the
                                                  class declaration.  }
  {$EXTERNALSYM DT_MIPS_DELTA_CLASSSYM}
  DT_MIPS_DELTA_CLASSSYM_NO      = $70000021;   { Number of entries in
                                                  DT_MIPS_DELTA_CLASSSYM.  }
  {$EXTERNALSYM DT_MIPS_DELTA_CLASSSYM_NO}
  DT_MIPS_CXX_FLAGS              = $70000022;   { Flags indicating for C++ flavor.  }
  {$EXTERNALSYM DT_MIPS_CXX_FLAGS}
  DT_MIPS_PIXIE_INIT             = $70000023;
  {$EXTERNALSYM DT_MIPS_PIXIE_INIT}
  DT_MIPS_SYMBOL_LIB             = $70000024;
  {$EXTERNALSYM DT_MIPS_SYMBOL_LIB}
  DT_MIPS_LOCALPAGE_GOTIDX       = $70000025;
  {$EXTERNALSYM DT_MIPS_LOCALPAGE_GOTIDX}
  DT_MIPS_LOCAL_GOTIDX           = $70000026;
  {$EXTERNALSYM DT_MIPS_LOCAL_GOTIDX}
  DT_MIPS_HIDDEN_GOTIDX          = $70000027;
  {$EXTERNALSYM DT_MIPS_HIDDEN_GOTIDX}
  DT_MIPS_PROTECTED_GOTIDX       = $70000028;
  {$EXTERNALSYM DT_MIPS_PROTECTED_GOTIDX}
  DT_MIPS_OPTIONS                = $70000029;   { Address of .options.  }
  {$EXTERNALSYM DT_MIPS_OPTIONS}
  DT_MIPS_INTERFACE              = $7000002a;   { Address of .interface.  }
  {$EXTERNALSYM DT_MIPS_INTERFACE}
  DT_MIPS_DYNSTR_ALIGN           = $7000002b;
  {$EXTERNALSYM DT_MIPS_DYNSTR_ALIGN}
  DT_MIPS_INTERFACE_SIZE         = $7000002c;   { Size of the .interface section. }
  {$EXTERNALSYM DT_MIPS_INTERFACE_SIZE}
  DT_MIPS_RLD_TEXT_RESOLVE_ADDR  = $7000002d;   { Address of rld_text_rsolve
                                                  function stored in GOT.  }
  {$EXTERNALSYM DT_MIPS_RLD_TEXT_RESOLVE_ADDR}
  DT_MIPS_PERF_SUFFIX            = $7000002e;   { Default suffix of dso to be added
                                                  by rld on dlopen() calls.  }
  {$EXTERNALSYM DT_MIPS_PERF_SUFFIX}
  DT_MIPS_COMPACT_SIZE           = $7000002f;   { (O32)Size of compact rel section. }
  {$EXTERNALSYM DT_MIPS_COMPACT_SIZE}
  DT_MIPS_GP_VALUE               = $70000030;   { GP value for aux GOTs.  }
  {$EXTERNALSYM DT_MIPS_GP_VALUE}
  DT_MIPS_AUX_DYNAMIC            = $70000031;   { Address of aux .dynamic.  }
  {$EXTERNALSYM DT_MIPS_AUX_DYNAMIC}

(* Moved up to resolve dependency
  DT_MIPS_NUM = $32;
*)

{ Legal values for DT_MIPS_FLAGS Elf32_Dyn entry.  }
const
  RHF_NONE                   = 0;           { No flags }
  {$EXTERNALSYM RHF_NONE}
  RHF_QUICKSTART             = (1 shl 0);   { Use quickstart }
  {$EXTERNALSYM RHF_QUICKSTART}
  RHF_NOTPOT                 = (1 shl 1);   { Hash size not power of 2 }
  {$EXTERNALSYM RHF_NOTPOT}
  RHF_NO_LIBRARY_REPLACEMENT = (1 shl 2);   { Ignore LD_LIBRARY_PATH }
  {$EXTERNALSYM RHF_NO_LIBRARY_REPLACEMENT}
  RHF_NO_MOVE                = (1 shl 3);
  {$EXTERNALSYM RHF_NO_MOVE}
  RHF_SGI_ONLY               = (1 shl 4);
  {$EXTERNALSYM RHF_SGI_ONLY}
  RHF_GUARANTEE_INIT         = (1 shl 5);
  {$EXTERNALSYM RHF_GUARANTEE_INIT}
  RHF_DELTA_C_PLUS_PLUS      = (1 shl 6);
  {$EXTERNALSYM RHF_DELTA_C_PLUS_PLUS}
  RHF_GUARANTEE_START_INIT   = (1 shl 7);
  {$EXTERNALSYM RHF_GUARANTEE_START_INIT}
  RHF_PIXIE                  = (1 shl 8);
  {$EXTERNALSYM RHF_PIXIE}
  RHF_DEFAULT_DELAY_LOAD     = (1 shl 9);
  {$EXTERNALSYM RHF_DEFAULT_DELAY_LOAD}
  RHF_REQUICKSTART           = (1 shl 10);
  {$EXTERNALSYM RHF_REQUICKSTART}
  RHF_REQUICKSTARTED         = (1 shl 11);
  {$EXTERNALSYM RHF_REQUICKSTARTED}
  RHF_CORD                   = (1 shl 12);
  {$EXTERNALSYM RHF_CORD}
  RHF_NO_UNRES_UNDEF         = (1 shl 13);
  {$EXTERNALSYM RHF_NO_UNRES_UNDEF}
  RHF_RLD_ORDER_SAFE         = (1 shl 14);
  {$EXTERNALSYM RHF_RLD_ORDER_SAFE}

{ Entries found in sections of type SHT_MIPS_LIBLIST.  }

type
  Elf32_Lib = {packed} record
    l_name: Elf32_Word;         { Name (string table index) }
    l_time_stamp: Elf32_Word;   { Timestamp }
    l_checksum: Elf32_Word;     { Checksum }
    l_version: Elf32_Word;      { Interface version }
    l_flags: Elf32_Word;        { Flags }
  end;
  {$EXTERNALSYM Elf32_Lib}
  PElf32_Lib = ^Elf32_Lib;

type
  Elf64_Lib = {packed} record
    l_name: Elf64_Word;	        { Name (string table index) }
    l_time_stamp: Elf64_Word;   { Timestamp }
    l_checksum: Elf64_Word;     { Checksum }
    l_version: Elf64_Word;      { Interface version }
    l_flags: Elf64_Word;        { Flags }
  end;
  {$EXTERNALSYM Elf64_Lib}
  PElf64_Lib = ^Elf64_Lib;


{ Legal values for l_flags.  }
const
  LL_NONE           = 0;
  {$EXTERNALSYM LL_NONE}
  LL_EXACT_MATCH    = (1 shl 0); { Require exact match }
  {$EXTERNALSYM LL_EXACT_MATCH}
  LL_IGNORE_INT_VER = (1 shl 1); { Ignore interface version }
  {$EXTERNALSYM LL_IGNORE_INT_VER}
  LL_REQUIRE_MINOR  = (1 shl 2);
  {$EXTERNALSYM LL_REQUIRE_MINOR}
  LL_EXPORTS        = (1 shl 3);
  {$EXTERNALSYM LL_EXPORTS}
  LL_DELAY_LOAD     = (1 shl 4);
  {$EXTERNALSYM LL_DELAY_LOAD}
  LL_DELTA          = (1 shl 5);
  {$EXTERNALSYM LL_DELTA}

{ Entries found in sections of type SHT_MIPS_CONFLICT.  }

type
  Elf32_Conflict = Elf32_Addr;
  {$EXTERNALSYM Elf32_Conflict}


{ HPPA specific definitions.  }

{ Legal values for e_flags field of Elf32_Ehdr.  }
const
  EF_PARISC_TRAPNIL  = $00010000; { Trap nil pointer dereference.  }
  {$EXTERNALSYM EF_PARISC_TRAPNIL}
  EF_PARISC_EXT      = $00020000; { Program uses arch. extensions. }
  {$EXTERNALSYM EF_PARISC_EXT}
  EF_PARISC_LSB      = $00040000; { Program expects little endian. }
  {$EXTERNALSYM EF_PARISC_LSB}
  EF_PARISC_WIDE     = $00080000; { Program expects wide mode.  }
  {$EXTERNALSYM EF_PARISC_WIDE}
  EF_PARISC_NO_KABP  = $00100000; { No kernel assisted branch prediction.  }
  {$EXTERNALSYM EF_PARISC_NO_KABP}
  EF_PARISC_LAZYSWAP = $00400000; { Allow lazy swapping.  }
  {$EXTERNALSYM EF_PARISC_LAZYSWAP}
  EF_PARISC_ARCH     = $0000ffff; { Architecture version.  }
  {$EXTERNALSYM EF_PARISC_ARCH}

{ Defined values for `e_flags & EF_PARISC_ARCH' are:  }

  EFA_PARISC_1_0 = $020b; { PA-RISC 1.0 big-endian.  }
  {$EXTERNALSYM EFA_PARISC_1_0}
  EFA_PARISC_1_1 = $0210; { PA-RISC 1.1 big-endian.  }
  {$EXTERNALSYM EFA_PARISC_1_1}
  EFA_PARISC_2_0 = $0214; { PA-RISC 2.0 big-endian.  }
  {$EXTERNALSYM EFA_PARISC_2_0}

{ Additional section indeces.  }

  SHN_PARISC_ANSI_COMMON = $ff00;  { Section for tenatively declared
                                     symbols in ANSI C.  }
  {$EXTERNALSYM SHN_PARISC_ANSI_COMMON}
  SHN_PARISC_HUGE_COMMON = $ff01;  { Common blocks in huge model.  }
  {$EXTERNALSYM SHN_PARISC_HUGE_COMMON}

{ Legal values for sh_type field of Elf32_Shdr.  }

  SHT_PARISC_EXT    = $70000000; { Contains product specific ext. }
  {$EXTERNALSYM SHT_PARISC_EXT}
  SHT_PARISC_UNWIND = $70000001; { Unwind information.  }
  {$EXTERNALSYM SHT_PARISC_UNWIND}
  SHT_PARISC_DOC    = $70000002; { Debug info for optimized code. }
  {$EXTERNALSYM SHT_PARISC_DOC}

{ Legal values for sh_flags field of Elf32_Shdr.  }

  SHF_PARISC_SHORT = $20000000; { Section with short addressing. }
  {$EXTERNALSYM SHF_PARISC_SHORT}
  SHF_PARISC_HUGE  = $40000000; { Section far from gp.  }
  {$EXTERNALSYM SHF_PARISC_HUGE}
  SHF_PARISC_SBP   = $80000000; { Static branch prediction code. }
  {$EXTERNALSYM SHF_PARISC_SBP}

{ Legal values for ST_TYPE subfield of st_info (symbol type).  }

  STT_PARISC_MILLICODE = 13;   { Millicode function entry point.  }
  {$EXTERNALSYM STT_PARISC_MILLICODE}

  STT_HP_OPAQUE = (STT_LOOS + $1);
  {$EXTERNALSYM STT_HP_OPAQUE}
  STT_HP_STUB   = (STT_LOOS + $2);
  {$EXTERNALSYM STT_HP_STUB}

{ HPPA relocs.  }

  R_PARISC_NONE            = 0;    { No reloc.  }
  {$EXTERNALSYM R_PARISC_NONE}
  R_PARISC_DIR32           = 1;    { Direct 32-bit reference.  }
  {$EXTERNALSYM R_PARISC_DIR32}
  R_PARISC_DIR21L          = 2;    { Left 21 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR21L}
  R_PARISC_DIR17R          = 3;    { Right 17 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR17R}
  R_PARISC_DIR17F          = 4;    { 17 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR17F}
  R_PARISC_DIR14R          = 5;    { Right 14 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR14R}
  R_PARISC_PCREL32         = 9;    { 32-bit rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL32}
  R_PARISC_PCREL21L        = 10;   { Left 21 bits of rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL21L}
  R_PARISC_PCREL17R        = 11;   { Right 17 bits of rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL17R}
  R_PARISC_PCREL17F        = 12;   { 17 bits of rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL17F}
  R_PARISC_PCREL14R        = 14;   { Right 14 bits of rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL14R}
  R_PARISC_GPREL21L        = 26;   { GP-relative, left 21 bits.  }
  {$EXTERNALSYM R_PARISC_GPREL21L}
  R_PARISC_GPREL14R        = 30;   { GP-relative, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_GPREL14R}
  R_PARISC_LTOFF21L        = 34;   { LT-relative, left 21 bits.  }
  {$EXTERNALSYM R_PARISC_LTOFF21L}
  R_PARISC_LTOFF14R        = 38;   { LT-relative, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_LTOFF14R}
  R_PARISC_SECREL32        = 41;   { 32 bits section rel. address.  }
  {$EXTERNALSYM R_PARISC_SECREL32}
  R_PARISC_SEGBASE         = 48;   { No relocation, set segment base.  }
  {$EXTERNALSYM R_PARISC_SEGBASE}
  R_PARISC_SEGREL32        = 49;   { 32 bits segment rel. address.  }
  {$EXTERNALSYM R_PARISC_SEGREL32}
  R_PARISC_PLTOFF21L       = 50;   { PLT rel. address, left 21 bits.  }
  {$EXTERNALSYM R_PARISC_PLTOFF21L}
  R_PARISC_PLTOFF14R       = 54;   { PLT rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_PLTOFF14R}
  R_PARISC_LTOFF_FPTR32    = 57;   { 32 bits LT-rel. function pointer. }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR32}
  R_PARISC_LTOFF_FPTR21L   = 58;   { LT-rel. fct ptr, left 21 bits. }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR21L}
  R_PARISC_LTOFF_FPTR14R   = 62;   { LT-rel. fct ptr, right 14 bits. }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR14R}
  R_PARISC_FPTR64          = 64;   { 64 bits function address.  }
  {$EXTERNALSYM R_PARISC_FPTR64}
  R_PARISC_PLABEL32        = 65;   { 32 bits function address.  }
  {$EXTERNALSYM R_PARISC_PLABEL32}
  R_PARISC_PCREL64         = 72;   { 64 bits PC-rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL64}
  R_PARISC_PCREL22F        = 74;   { 22 bits PC-rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL22F}
  R_PARISC_PCREL14WR       = 75;   { PC-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_PCREL14WR}
  R_PARISC_PCREL14DR       = 76;   { PC rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_PCREL14DR}
  R_PARISC_PCREL16F        = 77;   { 16 bits PC-rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL16F}
  R_PARISC_PCREL16WF       = 78;   { 16 bits PC-rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL16WF}
  R_PARISC_PCREL16DF       = 79;   { 16 bits PC-rel. address.  }
  {$EXTERNALSYM R_PARISC_PCREL16DF}
  R_PARISC_DIR64           = 80;   { 64 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR64}
  R_PARISC_DIR14WR         = 83;   { 14 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR14WR}
  R_PARISC_DIR14DR         = 84;   { 14 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR14DR}
  R_PARISC_DIR16F          = 85;   { 16 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR16F}
  R_PARISC_DIR16WF         = 86;   { 16 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR16WF}
  R_PARISC_DIR16DF         = 87;   { 16 bits of eff. address.  }
  {$EXTERNALSYM R_PARISC_DIR16DF}
  R_PARISC_GPREL64         = 88;   { 64 bits of GP-rel. address.  }
  {$EXTERNALSYM R_PARISC_GPREL64}
  R_PARISC_GPREL14WR       = 91;   { GP-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_GPREL14WR}
  R_PARISC_GPREL14DR       = 92;   { GP-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_GPREL14DR}
  R_PARISC_GPREL16F        = 93;   { 16 bits GP-rel. address.  }
  {$EXTERNALSYM R_PARISC_GPREL16F}
  R_PARISC_GPREL16WF       = 94;   { 16 bits GP-rel. address.  }
  {$EXTERNALSYM R_PARISC_GPREL16WF}
  R_PARISC_GPREL16DF       = 95;   { 16 bits GP-rel. address.  }
  {$EXTERNALSYM R_PARISC_GPREL16DF}
  R_PARISC_LTOFF64         = 96;   { 64 bits LT-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF64}
  R_PARISC_LTOFF14WR       = 99;   { LT-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_LTOFF14WR}
  R_PARISC_LTOFF14DR       = 100;  { LT-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_LTOFF14DR}
  R_PARISC_LTOFF16F        = 101;  { 16 bits LT-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF16F}
  R_PARISC_LTOFF16WF       = 102;  { 16 bits LT-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF16WF}
  R_PARISC_LTOFF16DF       = 103;  { 16 bits LT-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF16DF}
  R_PARISC_SECREL64        = 104;  { 64 bits section rel. address.  }
  {$EXTERNALSYM R_PARISC_SECREL64}
  R_PARISC_SEGREL64        = 112;  { 64 bits segment rel. address.  }
  {$EXTERNALSYM R_PARISC_SEGREL64}
  R_PARISC_PLTOFF14WR      = 115;  { PLT-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_PLTOFF14WR}
  R_PARISC_PLTOFF14DR      = 116;  { PLT-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_PLTOFF14DR}
  R_PARISC_PLTOFF16F       = 117;  { 16 bits LT-rel. address.  }
  {$EXTERNALSYM R_PARISC_PLTOFF16F}
  R_PARISC_PLTOFF16WF      = 118;  { 16 bits PLT-rel. address.  }
  {$EXTERNALSYM R_PARISC_PLTOFF16WF}
  R_PARISC_PLTOFF16DF      = 119;  { 16 bits PLT-rel. address.  }
  {$EXTERNALSYM R_PARISC_PLTOFF16DF}
  R_PARISC_LTOFF_FPTR64    = 120;  { 64 bits LT-rel. function ptr.  }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR64}
  R_PARISC_LTOFF_FPTR14WR  = 123;  { LT-rel. fct. ptr., right 14 bits. }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR14WR}
  R_PARISC_LTOFF_FPTR14DR  = 124;  { LT-rel. fct. ptr., right 14 bits. }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR14DR}
  R_PARISC_LTOFF_FPTR16F   = 125;  { 16 bits LT-rel. function ptr.  }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR16F}
  R_PARISC_LTOFF_FPTR16WF  = 126;  { 16 bits LT-rel. function ptr.  }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR16WF}
  R_PARISC_LTOFF_FPTR16DF  = 127;  { 16 bits LT-rel. function ptr.  }
  {$EXTERNALSYM R_PARISC_LTOFF_FPTR16DF}
  R_PARISC_LORESERVE       = 128;
  {$EXTERNALSYM R_PARISC_LORESERVE}
  R_PARISC_COPY            = 128;  { Copy relocation.  }
  {$EXTERNALSYM R_PARISC_COPY}
  R_PARISC_IPLT            = 129;  { Dynamic reloc, imported PLT }
  {$EXTERNALSYM R_PARISC_IPLT}
  R_PARISC_EPLT            = 130;  { Dynamic reloc, exported PLT }
  {$EXTERNALSYM R_PARISC_EPLT}
  R_PARISC_TPREL32         = 153;  { 32 bits TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_TPREL32}
  R_PARISC_TPREL21L        = 154;  { TP-rel. address, left 21 bits.  }
  {$EXTERNALSYM R_PARISC_TPREL21L}
  R_PARISC_TPREL14R        = 158;  { TP-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_TPREL14R}
  R_PARISC_LTOFF_TP21L     = 162;  { LT-TP-rel. address, left 21 bits. }
  {$EXTERNALSYM R_PARISC_LTOFF_TP21L}
  R_PARISC_LTOFF_TP14R     = 166;  { LT-TP-rel. address, right 14 bits.}
  {$EXTERNALSYM R_PARISC_LTOFF_TP14R}
  R_PARISC_LTOFF_TP14F     = 167;  { 14 bits LT-TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF_TP14F}
  R_PARISC_TPREL64         = 216;  { 64 bits TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_TPREL64}
  R_PARISC_TPREL14WR       = 219;  { TP-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_TPREL14WR}
  R_PARISC_TPREL14DR       = 220;  { TP-rel. address, right 14 bits.  }
  {$EXTERNALSYM R_PARISC_TPREL14DR}
  R_PARISC_TPREL16F        = 221;  { 16 bits TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_TPREL16F}
  R_PARISC_TPREL16WF       = 222;  { 16 bits TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_TPREL16WF}
  R_PARISC_TPREL16DF       = 223;  { 16 bits TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_TPREL16DF}
  R_PARISC_LTOFF_TP64      = 224;  { 64 bits LT-TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF_TP64}
  R_PARISC_LTOFF_TP14WR    = 227;  { LT-TP-rel. address, right 14 bits.}
  {$EXTERNALSYM R_PARISC_LTOFF_TP14WR}
  R_PARISC_LTOFF_TP14DR    = 228;  { LT-TP-rel. address, right 14 bits.}
  {$EXTERNALSYM R_PARISC_LTOFF_TP14DR}
  R_PARISC_LTOFF_TP16F     = 229;  { 16 bits LT-TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF_TP16F}
  R_PARISC_LTOFF_TP16WF    = 230;  { 16 bits LT-TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF_TP16WF}
  R_PARISC_LTOFF_TP16DF    = 231;  { 16 bits LT-TP-rel. address.  }
  {$EXTERNALSYM R_PARISC_LTOFF_TP16DF}
  R_PARISC_HIRESERVE       = 255;
  {$EXTERNALSYM R_PARISC_HIRESERVE}

{ Legal values for p_type field of Elf32_Phdr/Elf64_Phdr.  }

  PT_HP_TLS            = (PT_LOOS + $0);
  {$EXTERNALSYM PT_HP_TLS}
  PT_HP_CORE_NONE      = (PT_LOOS + $1);
  {$EXTERNALSYM PT_HP_CORE_NONE}
  PT_HP_CORE_VERSION   = (PT_LOOS + $2);
  {$EXTERNALSYM PT_HP_CORE_VERSION}
  PT_HP_CORE_KERNEL    = (PT_LOOS + $3);
  {$EXTERNALSYM PT_HP_CORE_KERNEL}
  PT_HP_CORE_COMM      = (PT_LOOS + $4);
  {$EXTERNALSYM PT_HP_CORE_COMM}
  PT_HP_CORE_PROC      = (PT_LOOS + $5);
  {$EXTERNALSYM PT_HP_CORE_PROC}
  PT_HP_CORE_LOADABLE  = (PT_LOOS + $6);
  {$EXTERNALSYM PT_HP_CORE_LOADABLE}
  PT_HP_CORE_STACK     = (PT_LOOS + $7);
  {$EXTERNALSYM PT_HP_CORE_STACK}
  PT_HP_CORE_SHM       = (PT_LOOS + $8);
  {$EXTERNALSYM PT_HP_CORE_SHM}
  PT_HP_CORE_MMF       = (PT_LOOS + $9);
  {$EXTERNALSYM PT_HP_CORE_MMF}
  PT_HP_PARALLEL       = (PT_LOOS + $10);
  {$EXTERNALSYM PT_HP_PARALLEL}
  PT_HP_FASTBIND       = (PT_LOOS + $11);
  {$EXTERNALSYM PT_HP_FASTBIND}
  PT_HP_OPT_ANNOT      = (PT_LOOS + $12);
  {$EXTERNALSYM PT_HP_OPT_ANNOT}
  PT_HP_HSL_ANNOT      = (PT_LOOS + $13);
  {$EXTERNALSYM PT_HP_HSL_ANNOT}
  PT_HP_STACK          = (PT_LOOS + $14);
  {$EXTERNALSYM PT_HP_STACK}

  PT_PARISC_ARCHEXT = $70000000;
  {$EXTERNALSYM PT_PARISC_ARCHEXT}
  PT_PARISC_UNWIND  = $70000001;
  {$EXTERNALSYM PT_PARISC_UNWIND}

{ Legal values for p_flags field of Elf32_Phdr/Elf64_Phdr.  }

  PF_PARISC_SBP     = $08000000;
  {$EXTERNALSYM PF_PARISC_SBP}

  PF_HP_PAGE_SIZE   = $00100000;
  {$EXTERNALSYM PF_HP_PAGE_SIZE}
  PF_HP_FAR_SHARED  = $00200000;
  {$EXTERNALSYM PF_HP_FAR_SHARED}
  PF_HP_NEAR_SHARED = $00400000;
  {$EXTERNALSYM PF_HP_NEAR_SHARED}
  PF_HP_CODE        = $01000000;
  {$EXTERNALSYM PF_HP_CODE}
  PF_HP_MODIFY      = $02000000;
  {$EXTERNALSYM PF_HP_MODIFY}
  PF_HP_LAZYSWAP    = $04000000;
  {$EXTERNALSYM PF_HP_LAZYSWAP}
  PF_HP_SBP         = $08000000;
  {$EXTERNALSYM PF_HP_SBP}


{ Alpha specific definitions.  }

{ Legal values for e_flags field of Elf64_Ehdr.  }

  EF_ALPHA_32BIT    = 1; { All addresses must be < 2GB.  }
  {$EXTERNALSYM EF_ALPHA_32BIT}
  EF_ALPHA_CANRELAX = 2; { Relocations for relaxing exist.  }
  {$EXTERNALSYM EF_ALPHA_CANRELAX}

{ Legal values for sh_type field of Elf64_Shdr.  }

{ These two are primarily concerned with ECOFF debugging info.  }
  SHT_ALPHA_DEBUG   = $70000001;
  {$EXTERNALSYM SHT_ALPHA_DEBUG}
  SHT_ALPHA_REGINFO = $70000002;
  {$EXTERNALSYM SHT_ALPHA_REGINFO}

{ Legal values for sh_flags field of Elf64_Shdr.  }

  SHF_ALPHA_GPREL = $10000000;
  {$EXTERNALSYM SHF_ALPHA_GPREL}

{ Legal values for st_other field of Elf64_Sym.  }
  STO_ALPHA_NOPV       = $80;  { No PV required.  }
  {$EXTERNALSYM STO_ALPHA_NOPV}
  STO_ALPHA_STD_GPLOAD = $88;  { PV only used for initial ldgp.  }
  {$EXTERNALSYM STO_ALPHA_STD_GPLOAD}

{ Alpha relocs.  }

  R_ALPHA_NONE             = 0;   { No reloc }
  {$EXTERNALSYM R_ALPHA_NONE}
  R_ALPHA_REFLONG          = 1;   { Direct 32 bit }
  {$EXTERNALSYM R_ALPHA_REFLONG}
  R_ALPHA_REFQUAD          = 2;   { Direct 64 bit }
  {$EXTERNALSYM R_ALPHA_REFQUAD}
  R_ALPHA_GPREL32          = 3;   { GP relative 32 bit }
  {$EXTERNALSYM R_ALPHA_GPREL32}
  R_ALPHA_LITERAL          = 4;   { GP relative 16 bit w/optimization }
  {$EXTERNALSYM R_ALPHA_LITERAL}
  R_ALPHA_LITUSE           = 5;   { Optimization hint for LITERAL }
  {$EXTERNALSYM R_ALPHA_LITUSE}
  R_ALPHA_GPDISP           = 6;   { Add displacement to GP }
  {$EXTERNALSYM R_ALPHA_GPDISP}
  R_ALPHA_BRADDR           = 7;   { PC+4 relative 23 bit shifted }
  {$EXTERNALSYM R_ALPHA_BRADDR}
  R_ALPHA_HINT             = 8;   { PC+4 relative 16 bit shifted }
  {$EXTERNALSYM R_ALPHA_HINT}
  R_ALPHA_SREL16           = 9;   { PC relative 16 bit }
  {$EXTERNALSYM R_ALPHA_SREL16}
  R_ALPHA_SREL32           = 10;  { PC relative 32 bit }
  {$EXTERNALSYM R_ALPHA_SREL32}
  R_ALPHA_SREL64           = 11;  { PC relative 64 bit }
  {$EXTERNALSYM R_ALPHA_SREL64}
  R_ALPHA_OP_PUSH          = 12;  { OP stack push }
  {$EXTERNALSYM R_ALPHA_OP_PUSH}
  R_ALPHA_OP_STORE         = 13;  { OP stack pop and store }
  {$EXTERNALSYM R_ALPHA_OP_STORE}
  R_ALPHA_OP_PSUB          = 14;  { OP stack subtract }
  {$EXTERNALSYM R_ALPHA_OP_PSUB}
  R_ALPHA_OP_PRSHIFT       = 15;  { OP stack right shift }
  {$EXTERNALSYM R_ALPHA_OP_PRSHIFT}
  R_ALPHA_GPVALUE          = 16;
  {$EXTERNALSYM R_ALPHA_GPVALUE}
  R_ALPHA_GPRELHIGH        = 17;
  {$EXTERNALSYM R_ALPHA_GPRELHIGH}
  R_ALPHA_GPRELLOW         = 18;
  {$EXTERNALSYM R_ALPHA_GPRELLOW}
  R_ALPHA_IMMED_GP_16      = 19;
  {$EXTERNALSYM R_ALPHA_IMMED_GP_16}
  R_ALPHA_IMMED_GP_HI32    = 20;
  {$EXTERNALSYM R_ALPHA_IMMED_GP_HI32}
  R_ALPHA_IMMED_SCN_HI32   = 21;
  {$EXTERNALSYM R_ALPHA_IMMED_SCN_HI32}
  R_ALPHA_IMMED_BR_HI32    = 22;
  {$EXTERNALSYM R_ALPHA_IMMED_BR_HI32}
  R_ALPHA_IMMED_LO32       = 23;
  {$EXTERNALSYM R_ALPHA_IMMED_LO32}
  R_ALPHA_COPY             = 24;  { Copy symbol at runtime }
  {$EXTERNALSYM R_ALPHA_COPY}
  R_ALPHA_GLOB_DAT         = 25;  { Create GOT entry }
  {$EXTERNALSYM R_ALPHA_GLOB_DAT}
  R_ALPHA_JMP_SLOT         = 26;  { Create PLT entry }
  {$EXTERNALSYM R_ALPHA_JMP_SLOT}
  R_ALPHA_RELATIVE         = 27;  { Adjust by program base }
  {$EXTERNALSYM R_ALPHA_RELATIVE}
{ Keep this the last entry.  }
  R_ALPHA_NUM              = 28;
  {$EXTERNALSYM R_ALPHA_NUM}


{ PowerPC specific declarations }

{ Values for Elf32/64_Ehdr.e_flags.  }
  EF_PPC_EMB = $80000000;  { PowerPC embedded flag }
  {$EXTERNALSYM EF_PPC_EMB}

{ Cygnus local bits below }
  EF_PPC_RELOCATABLE     = $00010000;  { PowerPC -mrelocatable flag}
  {$EXTERNALSYM EF_PPC_RELOCATABLE}
  EF_PPC_RELOCATABLE_LIB = $00008000;  { PowerPC -mrelocatable-lib flag }
  {$EXTERNALSYM EF_PPC_RELOCATABLE_LIB}

{ PowerPC relocations defined by the ABIs }
  R_PPC_NONE            = 0;
  {$EXTERNALSYM R_PPC_NONE}
  R_PPC_ADDR32          = 1;     { 32bit absolute address }
  {$EXTERNALSYM R_PPC_ADDR32}
  R_PPC_ADDR24          = 2;     { 26bit address, 2 bits ignored.  }
  {$EXTERNALSYM R_PPC_ADDR24}
  R_PPC_ADDR16          = 3;     { 16bit absolute address }
  {$EXTERNALSYM R_PPC_ADDR16}
  R_PPC_ADDR16_LO       = 4;     { lower 16bit of absolute address }
  {$EXTERNALSYM R_PPC_ADDR16_LO}
  R_PPC_ADDR16_HI       = 5;     { high 16bit of absolute address }
  {$EXTERNALSYM R_PPC_ADDR16_HI}
  R_PPC_ADDR16_HA       = 6;     { adjusted high 16bit }
  {$EXTERNALSYM R_PPC_ADDR16_HA}
  R_PPC_ADDR14          = 7;     { 16bit address, 2 bits ignored }
  {$EXTERNALSYM R_PPC_ADDR14}
  R_PPC_ADDR14_BRTAKEN  = 8;
  {$EXTERNALSYM R_PPC_ADDR14_BRTAKEN}
  R_PPC_ADDR14_BRNTAKEN = 9;
  {$EXTERNALSYM R_PPC_ADDR14_BRNTAKEN}
  R_PPC_REL24           = 10;    { PC relative 26 bit }
  {$EXTERNALSYM R_PPC_REL24}
  R_PPC_REL14           = 11;    { PC relative 16 bit }
  {$EXTERNALSYM R_PPC_REL14}
  R_PPC_REL14_BRTAKEN   = 12;
  {$EXTERNALSYM R_PPC_REL14_BRTAKEN}
  R_PPC_REL14_BRNTAKEN  = 13;
  {$EXTERNALSYM R_PPC_REL14_BRNTAKEN}
  R_PPC_GOT16           = 14;
  {$EXTERNALSYM R_PPC_GOT16}
  R_PPC_GOT16_LO        = 15;
  {$EXTERNALSYM R_PPC_GOT16_LO}
  R_PPC_GOT16_HI        = 16;
  {$EXTERNALSYM R_PPC_GOT16_HI}
  R_PPC_GOT16_HA        = 17;
  {$EXTERNALSYM R_PPC_GOT16_HA}
  R_PPC_PLTREL24        = 18;
  {$EXTERNALSYM R_PPC_PLTREL24}
  R_PPC_COPY            = 19;
  {$EXTERNALSYM R_PPC_COPY}
  R_PPC_GLOB_DAT        = 20;
  {$EXTERNALSYM R_PPC_GLOB_DAT}
  R_PPC_JMP_SLOT        = 21;
  {$EXTERNALSYM R_PPC_JMP_SLOT}
  R_PPC_RELATIVE        = 22;
  {$EXTERNALSYM R_PPC_RELATIVE}
  R_PPC_LOCAL24PC       = 23;
  {$EXTERNALSYM R_PPC_LOCAL24PC}
  R_PPC_UADDR32         = 24;
  {$EXTERNALSYM R_PPC_UADDR32}
  R_PPC_UADDR16         = 25;
  {$EXTERNALSYM R_PPC_UADDR16}
  R_PPC_REL32           = 26;
  {$EXTERNALSYM R_PPC_REL32}
  R_PPC_PLT32           = 27;
  {$EXTERNALSYM R_PPC_PLT32}
  R_PPC_PLTREL32        = 28;
  {$EXTERNALSYM R_PPC_PLTREL32}
  R_PPC_PLT16_LO        = 29;
  {$EXTERNALSYM R_PPC_PLT16_LO}
  R_PPC_PLT16_HI        = 30;
  {$EXTERNALSYM R_PPC_PLT16_HI}
  R_PPC_PLT16_HA        = 31;
  {$EXTERNALSYM R_PPC_PLT16_HA}
  R_PPC_SDAREL16        = 32;
  {$EXTERNALSYM R_PPC_SDAREL16}
  R_PPC_SECTOFF         = 33;
  {$EXTERNALSYM R_PPC_SECTOFF}
  R_PPC_SECTOFF_LO      = 34;
  {$EXTERNALSYM R_PPC_SECTOFF_LO}
  R_PPC_SECTOFF_HI      = 35;
  {$EXTERNALSYM R_PPC_SECTOFF_HI}
  R_PPC_SECTOFF_HA      = 36;
  {$EXTERNALSYM R_PPC_SECTOFF_HA}
{ Keep this the last entry.  }
  R_PPC_NUM             = 37;
  {$EXTERNALSYM R_PPC_NUM}

{ The remaining relocs are from the Embedded ELF ABI, and are not
   in the SVR4 ELF ABI.  }
  R_PPC_EMB_NADDR32     = 101;
  {$EXTERNALSYM R_PPC_EMB_NADDR32}
  R_PPC_EMB_NADDR16     = 102;
  {$EXTERNALSYM R_PPC_EMB_NADDR16}
  R_PPC_EMB_NADDR16_LO  = 103;
  {$EXTERNALSYM R_PPC_EMB_NADDR16_LO}
  R_PPC_EMB_NADDR16_HI  = 104;
  {$EXTERNALSYM R_PPC_EMB_NADDR16_HI}
  R_PPC_EMB_NADDR16_HA  = 105;
  {$EXTERNALSYM R_PPC_EMB_NADDR16_HA}
  R_PPC_EMB_SDAI16      = 106;
  {$EXTERNALSYM R_PPC_EMB_SDAI16}
  R_PPC_EMB_SDA2I16     = 107;
  {$EXTERNALSYM R_PPC_EMB_SDA2I16}
  R_PPC_EMB_SDA2REL     = 108;
  {$EXTERNALSYM R_PPC_EMB_SDA2REL}
  R_PPC_EMB_SDA21       = 109;   { 16 bit offset in SDA }
  {$EXTERNALSYM R_PPC_EMB_SDA21}
  R_PPC_EMB_MRKREF      = 110;
  {$EXTERNALSYM R_PPC_EMB_MRKREF}
  R_PPC_EMB_RELSEC16    = 111;
  {$EXTERNALSYM R_PPC_EMB_RELSEC16}
  R_PPC_EMB_RELST_LO    = 112;
  {$EXTERNALSYM R_PPC_EMB_RELST_LO}
  R_PPC_EMB_RELST_HI    = 113;
  {$EXTERNALSYM R_PPC_EMB_RELST_HI}
  R_PPC_EMB_RELST_HA    = 114;
  {$EXTERNALSYM R_PPC_EMB_RELST_HA}
  R_PPC_EMB_BIT_FLD     = 115;
  {$EXTERNALSYM R_PPC_EMB_BIT_FLD}
  R_PPC_EMB_RELSDA      = 116;   { 16 bit relative offset in SDA }
  {$EXTERNALSYM R_PPC_EMB_RELSDA}

{ Diab tool relocations.  }
  R_PPC_DIAB_SDA21_LO   = 180;   { like EMB_SDA21, but lower 16 bit }
  {$EXTERNALSYM R_PPC_DIAB_SDA21_LO}
  R_PPC_DIAB_SDA21_HI   = 181;   { like EMB_SDA21, but high 16 bit }
  {$EXTERNALSYM R_PPC_DIAB_SDA21_HI}
  R_PPC_DIAB_SDA21_HA   = 182;   { like EMB_SDA21, adjusted high 16 }
  {$EXTERNALSYM R_PPC_DIAB_SDA21_HA}
  R_PPC_DIAB_RELSDA_LO  = 183;   { like EMB_RELSDA, but lower 16 bit }
  {$EXTERNALSYM R_PPC_DIAB_RELSDA_LO}
  R_PPC_DIAB_RELSDA_HI  = 184;   { like EMB_RELSDA, but high 16 bit }
  {$EXTERNALSYM R_PPC_DIAB_RELSDA_HI}
  R_PPC_DIAB_RELSDA_HA  = 185;   { like EMB_RELSDA, adjusted high 16 }
  {$EXTERNALSYM R_PPC_DIAB_RELSDA_HA}

{ This is a phony reloc to handle any old fashioned TOC16 references
   that may still be in object files.  }
  R_PPC_TOC16           = 255;
  {$EXTERNALSYM R_PPC_TOC16}


{ ARM specific declarations }

{ Processor specific flags for the ELF header e_flags field.  }
  EF_ARM_RELEXEC    = $01;
  {$EXTERNALSYM EF_ARM_RELEXEC}
  EF_ARM_HASENTRY   = $02;
  {$EXTERNALSYM EF_ARM_HASENTRY}
  EF_ARM_INTERWORK  = $04;
  {$EXTERNALSYM EF_ARM_INTERWORK}
  EF_ARM_APCS_26    = $08;
  {$EXTERNALSYM EF_ARM_APCS_26}
  EF_ARM_APCS_FLOAT = $10;
  {$EXTERNALSYM EF_ARM_APCS_FLOAT}
  EF_ARM_PIC        = $20;
  {$EXTERNALSYM EF_ARM_PIC}
  EF_ALIGN8         = $40;   { 8-bit structure alignment is in use }
  {$EXTERNALSYM EF_ALIGN8}
  EF_NEW_ABI        = $80;
  {$EXTERNALSYM EF_NEW_ABI}
  EF_OLD_ABI        = $100;
  {$EXTERNALSYM EF_OLD_ABI}

{ Additional symbol types for Thumb }
  STT_ARM_TFUNC     = $d;
  {$EXTERNALSYM STT_ARM_TFUNC}

{ ARM-specific values for sh_flags }
  SHF_ARM_ENTRYSECT = $10000000;  { Section contains an entry point }
  {$EXTERNALSYM SHF_ARM_ENTRYSECT}
  SHF_ARM_COMDEF    = $80000000;  { Section may be multiply defined
                                    in the input to a link step }
  {$EXTERNALSYM SHF_ARM_COMDEF}

{ ARM-specific program header flags }
  PF_ARM_SB = $10000000;  { Segment contains the location
                            addressed by the static base }
  {$EXTERNALSYM PF_ARM_SB}

{ ARM relocs.  }
  R_ARM_NONE           = 0;     { No reloc }
  {$EXTERNALSYM R_ARM_NONE}
  R_ARM_PC24           = 1;     { PC relative 26 bit branch }
  {$EXTERNALSYM R_ARM_PC24}
  R_ARM_ABS32          = 2;     { Direct 32 bit  }
  {$EXTERNALSYM R_ARM_ABS32}
  R_ARM_REL32          = 3;     { PC relative 32 bit }
  {$EXTERNALSYM R_ARM_REL32}
  R_ARM_PC13           = 4;
  {$EXTERNALSYM R_ARM_PC13}
  R_ARM_ABS16          = 5;     { Direct 16 bit }
  {$EXTERNALSYM R_ARM_ABS16}
  R_ARM_ABS12          = 6;     { Direct 12 bit }
  {$EXTERNALSYM R_ARM_ABS12}
  R_ARM_THM_ABS5       = 7;
  {$EXTERNALSYM R_ARM_THM_ABS5}
  R_ARM_ABS8           = 8;     { Direct 8 bit }
  {$EXTERNALSYM R_ARM_ABS8}
  R_ARM_SBREL32        = 9;
  {$EXTERNALSYM R_ARM_SBREL32}
  R_ARM_THM_PC22       = 10;
  {$EXTERNALSYM R_ARM_THM_PC22}
  R_ARM_THM_PC8        = 11;
  {$EXTERNALSYM R_ARM_THM_PC8}
  R_ARM_AMP_VCALL9     = 12;
  {$EXTERNALSYM R_ARM_AMP_VCALL9}
  R_ARM_SWI24          = 13;
  {$EXTERNALSYM R_ARM_SWI24}
  R_ARM_THM_SWI8       = 14;
  {$EXTERNALSYM R_ARM_THM_SWI8}
  R_ARM_XPC25          = 15;
  {$EXTERNALSYM R_ARM_XPC25}
  R_ARM_THM_XPC22      = 16;
  {$EXTERNALSYM R_ARM_THM_XPC22}
  R_ARM_COPY           = 20;    { Copy symbol at runtime }
  {$EXTERNALSYM R_ARM_COPY}
  R_ARM_GLOB_DAT       = 21;    { Create GOT entry }
  {$EXTERNALSYM R_ARM_GLOB_DAT}
  R_ARM_JUMP_SLOT      = 22;    { Create PLT entry }
  {$EXTERNALSYM R_ARM_JUMP_SLOT}
  R_ARM_RELATIVE       = 23;    { Adjust by program base }
  {$EXTERNALSYM R_ARM_RELATIVE}
  R_ARM_GOTOFF         = 24;    { 32 bit offset to GOT }
  {$EXTERNALSYM R_ARM_GOTOFF}
  R_ARM_GOTPC          = 25;    { 32 bit PC relative offset to GOT }
  {$EXTERNALSYM R_ARM_GOTPC}
  R_ARM_GOT32          = 26;    { 32 bit GOT entry }
  {$EXTERNALSYM R_ARM_GOT32}
  R_ARM_PLT32          = 27;    { 32 bit PLT address }
  {$EXTERNALSYM R_ARM_PLT32}
  R_ARM_GNU_VTENTRY    = 100;
  {$EXTERNALSYM R_ARM_GNU_VTENTRY}
  R_ARM_GNU_VTINHERIT  = 101;
  {$EXTERNALSYM R_ARM_GNU_VTINHERIT}
  R_ARM_THM_PC11       = 102;   { thumb unconditional branch }
  {$EXTERNALSYM R_ARM_THM_PC11}
  R_ARM_THM_PC9        = 103;   { thumb conditional branch }
  {$EXTERNALSYM R_ARM_THM_PC9}
  R_ARM_RXPC25         = 249;
  {$EXTERNALSYM R_ARM_RXPC25}
  R_ARM_RSBREL32       = 250;
  {$EXTERNALSYM R_ARM_RSBREL32}
  R_ARM_THM_RPC22      = 251;
  {$EXTERNALSYM R_ARM_THM_RPC22}
  R_ARM_RREL32         = 252;
  {$EXTERNALSYM R_ARM_RREL32}
  R_ARM_RABS22         = 253;
  {$EXTERNALSYM R_ARM_RABS22}
  R_ARM_RPC24          = 254;
  {$EXTERNALSYM R_ARM_RPC24}
  R_ARM_RBASE          = 255;
  {$EXTERNALSYM R_ARM_RBASE}
{ Keep this the last entry.  }
  R_ARM_NUM            = 256;
  {$EXTERNALSYM R_ARM_NUM}

{ IA-64 specific declarations.  }

{ Processor specific flags for the Ehdr e_flags field.  }
  EF_IA_64_MASKOS = $0000000f;  { os-specific flags }
  {$EXTERNALSYM EF_IA_64_MASKOS}
  EF_IA_64_ABI64  = $00000010;  { 64-bit ABI }
  {$EXTERNALSYM EF_IA_64_ABI64}
  EF_IA_64_ARCH   = $ff000000;  { arch. version mask }
  {$EXTERNALSYM EF_IA_64_ARCH}

{ Processor specific values for the Phdr p_type field.  }
  PT_IA_64_ARCHEXT = (PT_LOPROC + 0);  { arch extension bits }
  {$EXTERNALSYM PT_IA_64_ARCHEXT}
  PT_IA_64_UNWIND  = (PT_LOPROC + 1);  { ia64 unwind bits }
  {$EXTERNALSYM PT_IA_64_UNWIND}

{ Processor specific flags for the Phdr p_flags field.  }
  PF_IA_64_NORECOV = $80000000;  { spec insns w/o recovery }
  {$EXTERNALSYM PF_IA_64_NORECOV}

{ Processor specific values for the Shdr sh_type field.  }
  SHT_IA_64_EXT    = (SHT_LOPROC + 0);  { extension bits }
  {$EXTERNALSYM SHT_IA_64_EXT}
  SHT_IA_64_UNWIND = (SHT_LOPROC + 1);  { unwind bits }
  {$EXTERNALSYM SHT_IA_64_UNWIND}

{ Processor specific flags for the Shdr sh_flags field.  }
  SHF_IA_64_SHORT   = $10000000;  { section near gp }
  {$EXTERNALSYM SHF_IA_64_SHORT}
  SHF_IA_64_NORECOV = $20000000;  { spec insns w/o recovery }
  {$EXTERNALSYM SHF_IA_64_NORECOV}

{ Processor specific values for the Dyn d_tag field.  }
  DT_IA_64_PLT_RESERVE = (DT_LOPROC + 0);
  {$EXTERNALSYM DT_IA_64_PLT_RESERVE}
  DT_IA_64_NUM = 1;
  {$EXTERNALSYM DT_IA_64_NUM}

{ IA-64 relocations.  }
  R_IA64_NONE             = $00;  { none }
  {$EXTERNALSYM R_IA64_NONE}
  R_IA64_IMM14            = $21;  { symbol + addend, add imm14 }
  {$EXTERNALSYM R_IA64_IMM14}
  R_IA64_IMM22            = $22;  { symbol + addend, add imm22 }
  {$EXTERNALSYM R_IA64_IMM22}
  R_IA64_IMM64            = $23;  { symbol + addend, mov imm64 }
  {$EXTERNALSYM R_IA64_IMM64}
  R_IA64_DIR32MSB         = $24;  { symbol + addend, data4 MSB }
  {$EXTERNALSYM R_IA64_DIR32MSB}
  R_IA64_DIR32LSB         = $25;  { symbol + addend, data4 LSB }
  {$EXTERNALSYM R_IA64_DIR32LSB}
  R_IA64_DIR64MSB         = $26;  { symbol + addend, data8 MSB }
  {$EXTERNALSYM R_IA64_DIR64MSB}
  R_IA64_DIR64LSB         = $27;  { symbol + addend, data8 LSB }
  {$EXTERNALSYM R_IA64_DIR64LSB}
  R_IA64_GPREL22          = $2a;  { @gprel(sym + add), add imm22 }
  {$EXTERNALSYM R_IA64_GPREL22}
  R_IA64_GPREL64I         = $2b;  { @gprel(sym + add), mov imm64 }
  {$EXTERNALSYM R_IA64_GPREL64I}
  R_IA64_GPREL32MSB       = $2c;  { @gprel(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_GPREL32MSB}
  R_IA64_GPREL32LSB       = $2d;  { @gprel(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_GPREL32LSB}
  R_IA64_GPREL64MSB       = $2e;  { @gprel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_GPREL64MSB}
  R_IA64_GPREL64LSB       = $2f;  { @gprel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_GPREL64LSB}
  R_IA64_LTOFF22          = $32;  { @ltoff(sym + add), add imm22 }
  {$EXTERNALSYM R_IA64_LTOFF22}
  R_IA64_LTOFF64I         = $33;  { @ltoff(sym + add), mov imm64 }
  {$EXTERNALSYM R_IA64_LTOFF64I}
  R_IA64_PLTOFF22         = $3a;  { @pltoff(sym + add), add imm22 }
  {$EXTERNALSYM R_IA64_PLTOFF22}
  R_IA64_PLTOFF64I        = $3b;  { @pltoff(sym + add), mov imm64 }
  {$EXTERNALSYM R_IA64_PLTOFF64I}
  R_IA64_PLTOFF64MSB      = $3e;  { @pltoff(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_PLTOFF64MSB}
  R_IA64_PLTOFF64LSB      = $3f;  { @pltoff(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_PLTOFF64LSB}
  R_IA64_FPTR64I          = $43;  { @fptr(sym + add), mov imm64 }
  {$EXTERNALSYM R_IA64_FPTR64I}
  R_IA64_FPTR32MSB        = $44;  { @fptr(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_FPTR32MSB}
  R_IA64_FPTR32LSB        = $45;  { @fptr(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_FPTR32LSB}
  R_IA64_FPTR64MSB        = $46;  { @fptr(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_FPTR64MSB}
  R_IA64_FPTR64LSB        = $47;  { @fptr(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_FPTR64LSB}
  R_IA64_PCREL60B         = $48;  { @pcrel(sym + add), brl }
  {$EXTERNALSYM R_IA64_PCREL60B}
  R_IA64_PCREL21B         = $49;  { @pcrel(sym + add), ptb, call }
  {$EXTERNALSYM R_IA64_PCREL21B}
  R_IA64_PCREL21M         = $4a;  { @pcrel(sym + add), chk.s }
  {$EXTERNALSYM R_IA64_PCREL21M}
  R_IA64_PCREL21F         = $4b;  { @pcrel(sym + add), fchkf }
  {$EXTERNALSYM R_IA64_PCREL21F}
  R_IA64_PCREL32MSB       = $4c;  { @pcrel(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_PCREL32MSB}
  R_IA64_PCREL32LSB       = $4d;  { @pcrel(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_PCREL32LSB}
  R_IA64_PCREL64MSB       = $4e;  { @pcrel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_PCREL64MSB}
  R_IA64_PCREL64LSB       = $4f;  { @pcrel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_PCREL64LSB}
  R_IA64_LTOFF_FPTR22     = $52;  { @ltoff(@fptr(s+a)), imm22 }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR22}
  R_IA64_LTOFF_FPTR64I    = $53;  { @ltoff(@fptr(s+a)), imm64 }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR64I}
  R_IA64_LTOFF_FPTR32MSB  = $54;  { @ltoff(@fptr(s+a)), data4 MSB }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR32MSB}
  R_IA64_LTOFF_FPTR32LSB  = $55;  { @ltoff(@fptr(s+a)), data4 LSB }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR32LSB}
  R_IA64_LTOFF_FPTR64MSB  = $56;  { @ltoff(@fptr(s+a)), data8 MSB }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR64MSB}
  R_IA64_LTOFF_FPTR64LSB  = $57;  { @ltoff(@fptr(s+a)), data8 LSB }
  {$EXTERNALSYM R_IA64_LTOFF_FPTR64LSB}
  R_IA64_SEGREL32MSB      = $5c;  { @segrel(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_SEGREL32MSB}
  R_IA64_SEGREL32LSB      = $5d;  { @segrel(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_SEGREL32LSB}
  R_IA64_SEGREL64MSB      = $5e;  { @segrel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_SEGREL64MSB}
  R_IA64_SEGREL64LSB      = $5f;  { @segrel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_SEGREL64LSB}
  R_IA64_SECREL32MSB      = $64;  { @secrel(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_SECREL32MSB}
  R_IA64_SECREL32LSB      = $65;  { @secrel(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_SECREL32LSB}
  R_IA64_SECREL64MSB      = $66;  { @secrel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_SECREL64MSB}
  R_IA64_SECREL64LSB      = $67;  { @secrel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_SECREL64LSB}
  R_IA64_REL32MSB         = $6c;  { data 4 + REL }
  {$EXTERNALSYM R_IA64_REL32MSB}
  R_IA64_REL32LSB         = $6d;  { data 4 + REL }
  {$EXTERNALSYM R_IA64_REL32LSB}
  R_IA64_REL64MSB         = $6e;  { data 8 + REL }
  {$EXTERNALSYM R_IA64_REL64MSB}
  R_IA64_REL64LSB         = $6f;  { data 8 + REL }
  {$EXTERNALSYM R_IA64_REL64LSB}
  R_IA64_LTV32MSB         = $74;  { symbol + addend, data4 MSB }
  {$EXTERNALSYM R_IA64_LTV32MSB}
  R_IA64_LTV32LSB         = $75;  { symbol + addend, data4 LSB }
  {$EXTERNALSYM R_IA64_LTV32LSB}
  R_IA64_LTV64MSB         = $76;  { symbol + addend, data8 MSB }
  {$EXTERNALSYM R_IA64_LTV64MSB}
  R_IA64_LTV64LSB         = $77;  { symbol + addend, data8 LSB }
  {$EXTERNALSYM R_IA64_LTV64LSB}
  R_IA64_PCREL21BI        = $79;  { @pcrel(sym + add), 21bit inst }
  {$EXTERNALSYM R_IA64_PCREL21BI}
  R_IA64_PCREL22          = $7a;  { @pcrel(sym + add), 22bit inst }
  {$EXTERNALSYM R_IA64_PCREL22}
  R_IA64_PCREL64I         = $7b;  { @pcrel(sym + add), 64bit inst }
  {$EXTERNALSYM R_IA64_PCREL64I}
  R_IA64_IPLTMSB          = $80;  { dynamic reloc, imported PLT, MSB }
  {$EXTERNALSYM R_IA64_IPLTMSB}
  R_IA64_IPLTLSB          = $81;  { dynamic reloc, imported PLT, LSB }
  {$EXTERNALSYM R_IA64_IPLTLSB}
  R_IA64_COPY             = $84;  { copy relocation }
  {$EXTERNALSYM R_IA64_COPY}
  R_IA64_SUB              = $85;  { Addend and symbol difference }
  {$EXTERNALSYM R_IA64_SUB}
  R_IA64_LTOFF22X         = $86;  { LTOFF22, relaxable.  }
  {$EXTERNALSYM R_IA64_LTOFF22X}
  R_IA64_LDXMOV           = $87;  { Use of LTOFF22X.  }
  {$EXTERNALSYM R_IA64_LDXMOV}
  R_IA64_TPREL14          = $91;  { @tprel(sym + add), imm14 }
  {$EXTERNALSYM R_IA64_TPREL14}
  R_IA64_TPREL22          = $92;  { @tprel(sym + add), imm22 }
  {$EXTERNALSYM R_IA64_TPREL22}
  R_IA64_TPREL64I         = $93;  { @tprel(sym + add), imm64 }
  {$EXTERNALSYM R_IA64_TPREL64I}
  R_IA64_TPREL64MSB       = $96;  { @tprel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_TPREL64MSB}
  R_IA64_TPREL64LSB       = $97;  { @tprel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_TPREL64LSB}
  R_IA64_LTOFF_TPREL22    = $9a;  { @ltoff(@tprel(s+a)), imm2 }
  {$EXTERNALSYM R_IA64_LTOFF_TPREL22}
  R_IA64_DTPMOD64MSB      = $a6;  { @dtpmod(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_DTPMOD64MSB}
  R_IA64_DTPMOD64LSB      = $a7;  { @dtpmod(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_DTPMOD64LSB}
  R_IA64_LTOFF_DTPMOD22   = $aa;  { @ltoff(@dtpmod(sym + add)), imm22 }
  {$EXTERNALSYM R_IA64_LTOFF_DTPMOD22}
  R_IA64_DTPREL14         = $b1;  { @dtprel(sym + add), imm14 }
  {$EXTERNALSYM R_IA64_DTPREL14}
  R_IA64_DTPREL22         = $b2;  { @dtprel(sym + add), imm22 }
  {$EXTERNALSYM R_IA64_DTPREL22}
  R_IA64_DTPREL64I        = $b3;  { @dtprel(sym + add), imm64 }
  {$EXTERNALSYM R_IA64_DTPREL64I}
  R_IA64_DTPREL32MSB      = $b4;  { @dtprel(sym + add), data4 MSB }
  {$EXTERNALSYM R_IA64_DTPREL32MSB}
  R_IA64_DTPREL32LSB      = $b5;  { @dtprel(sym + add), data4 LSB }
  {$EXTERNALSYM R_IA64_DTPREL32LSB}
  R_IA64_DTPREL64MSB      = $b6;  { @dtprel(sym + add), data8 MSB }
  {$EXTERNALSYM R_IA64_DTPREL64MSB}
  R_IA64_DTPREL64LSB      = $b7;  { @dtprel(sym + add), data8 LSB }
  {$EXTERNALSYM R_IA64_DTPREL64LSB}
  R_IA64_LTOFF_DTPREL22   = $ba;  { @ltoff(@dtprel(s+a)), imm22 }
  {$EXTERNALSYM R_IA64_LTOFF_DTPREL22}

{ SH specific declarations }

{ SH relocs.  }
  R_SH_NONE          = 0;
  {$EXTERNALSYM R_SH_NONE}
  R_SH_DIR32         = 1;
  {$EXTERNALSYM R_SH_DIR32}
  R_SH_REL32         = 2;
  {$EXTERNALSYM R_SH_REL32}
  R_SH_DIR8WPN       = 3;
  {$EXTERNALSYM R_SH_DIR8WPN}
  R_SH_IND12W        = 4;
  {$EXTERNALSYM R_SH_IND12W}
  R_SH_DIR8WPL       = 5;
  {$EXTERNALSYM R_SH_DIR8WPL}
  R_SH_DIR8WPZ       = 6;
  {$EXTERNALSYM R_SH_DIR8WPZ}
  R_SH_DIR8BP        = 7;
  {$EXTERNALSYM R_SH_DIR8BP}
  R_SH_DIR8W         = 8;
  {$EXTERNALSYM R_SH_DIR8W}
  R_SH_DIR8L         = 9;
  {$EXTERNALSYM R_SH_DIR8L}
  R_SH_SWITCH16      = 25;
  {$EXTERNALSYM R_SH_SWITCH16}
  R_SH_SWITCH32      = 26;
  {$EXTERNALSYM R_SH_SWITCH32}
  R_SH_USES          = 27;
  {$EXTERNALSYM R_SH_USES}
  R_SH_COUNT         = 28;
  {$EXTERNALSYM R_SH_COUNT}
  R_SH_ALIGN         = 29;
  {$EXTERNALSYM R_SH_ALIGN}
  R_SH_CODE          = 30;
  {$EXTERNALSYM R_SH_CODE}
  R_SH_DATA          = 31;
  {$EXTERNALSYM R_SH_DATA}
  R_SH_LABEL         = 32;
  {$EXTERNALSYM R_SH_LABEL}
  R_SH_SWITCH8       = 33;
  {$EXTERNALSYM R_SH_SWITCH8}
  R_SH_GNU_VTINHERIT = 34;
  {$EXTERNALSYM R_SH_GNU_VTINHERIT}
  R_SH_GNU_VTENTRY   = 35;
  {$EXTERNALSYM R_SH_GNU_VTENTRY}
  R_SH_GOT32         = 160;
  {$EXTERNALSYM R_SH_GOT32}
  R_SH_PLT32         = 161;
  {$EXTERNALSYM R_SH_PLT32}
  R_SH_COPY          = 162;
  {$EXTERNALSYM R_SH_COPY}
  R_SH_GLOB_DAT      = 163;
  {$EXTERNALSYM R_SH_GLOB_DAT}
  R_SH_JMP_SLOT      = 164;
  {$EXTERNALSYM R_SH_JMP_SLOT}
  R_SH_RELATIVE      = 165;
  {$EXTERNALSYM R_SH_RELATIVE}
  R_SH_GOTOFF        = 166;
  {$EXTERNALSYM R_SH_GOTOFF}
  R_SH_GOTPC         = 167;
  {$EXTERNALSYM R_SH_GOTPC}
{ Keep this the last entry.  }
  R_SH_NUM           = 256;
  {$EXTERNALSYM R_SH_NUM}

{ Additional s390 relocs }

  R_390_NONE     = 0;              { No reloc.  }
  {$EXTERNALSYM R_390_NONE}
  R_390_8        = 1;              { Direct 8 bit.  }
  {$EXTERNALSYM R_390_8}
  R_390_12       = 2;              { Direct 12 bit.  }
  {$EXTERNALSYM R_390_12}
  R_390_16       = 3;              { Direct 16 bit.  }
  {$EXTERNALSYM R_390_16}
  R_390_32       = 4;              { Direct 32 bit.  }
  {$EXTERNALSYM R_390_32}
  R_390_PC32     = 5;              { PC relative 32 bit.  }
  {$EXTERNALSYM R_390_PC32}
  R_390_GOT12    = 6;              { 12 bit GOT offset.  }
  {$EXTERNALSYM R_390_GOT12}
  R_390_GOT32    = 7;              { 32 bit GOT offset.  }
  {$EXTERNALSYM R_390_GOT32}
  R_390_PLT32    = 8;              { 32 bit PC relative PLT address.  }
  {$EXTERNALSYM R_390_PLT32}
  R_390_COPY     = 9;              { Copy symbol at runtime.  }
  {$EXTERNALSYM R_390_COPY}
  R_390_GLOB_DAT = 10;             { Create GOT entry.  }
  {$EXTERNALSYM R_390_GLOB_DAT}
  R_390_JMP_SLOT = 11;             { Create PLT entry.  }
  {$EXTERNALSYM R_390_JMP_SLOT}
  R_390_RELATIVE = 12;             { Adjust by program base.  }
  {$EXTERNALSYM R_390_RELATIVE}
  R_390_GOTOFF   = 13;             { 32 bit offset to GOT.  }
  {$EXTERNALSYM R_390_GOTOFF}
  R_390_GOTPC    = 14;             { 32 bit PC relative offset to GOT.  }
  {$EXTERNALSYM R_390_GOTPC}
  R_390_GOT16    = 15;             { 16 bit GOT offset.  }
  {$EXTERNALSYM R_390_GOT16}
  R_390_PC16     = 16;             { PC relative 16 bit.  }
  {$EXTERNALSYM R_390_PC16}
  R_390_PC16DBL  = 17;             { PC relative 16 bit shifted by 1.  }
  {$EXTERNALSYM R_390_PC16DBL}
  R_390_PLT16DBL = 18;             { 32 bit PC rel. PLT shifted by 1.  }
  {$EXTERNALSYM R_390_PLT16DBL}
{ Keep this the last entry.  }
  R_390_NUM      = 19;
  {$EXTERNALSYM R_390_NUM}


implementation

function ELF32_ST_BIND(const val: Byte): Byte;
begin
  Result := val shr 4;
end;

function ELF32_ST_TYPE(const val: Byte): Byte;
begin
  Result := val and $0f;
end;

function ELF32_ST_INFO(const bind, __type: Byte): Byte;
begin
  Result := (bind shl 4) or (__type and $f);
end;

function ELF64_ST_BIND(const val: Byte): Byte;
begin
  Result := ELF32_ST_BIND(val);
end;

function ELF64_ST_TYPE(const val: Byte): Byte;
begin
  Result := ELF32_ST_TYPE(val);
end;

function ELF64_ST_INFO(const bind, __type: Byte): Byte;
begin
  Result := ELF32_ST_INFO(bind, __type);
end;

function ELF32_ST_VISIBILITY(const o: Byte): Byte;
begin
  Result := o and $03;
end;

function ELF64_ST_VISIBILITY(const o: Byte): Byte;
begin
  Result := ELF32_ST_VISIBILITY(o);
end;

function ELF32_R_SYM(const val: Elf32_Word): Elf32_Word;
begin
  Result := val shr 8;
end;

function ELF32_R_TYPE(const val: Elf32_Word): Elf32_Word;
begin
  Result := val and $ff;
end;

function ELF32_R_INFO(const sym, __type: Elf32_Word): Elf32_Word;
begin
  Result := (sym shr 8) or (__type and $ff);
end;

function ELF64_R_SYM(const val: Elf64_Xword): Elf64_Xword;
begin
  Result := val shr 32;
end;

function ELF64_R_TYPE(const val: Elf64_Xword): Elf64_Xword;
begin
  Result := val and $ffffffff;
end;

function ELF64_R_INFO(const sym, __type: Elf64_Xword): Elf64_Xword;
begin
  Result := (sym shl 32) or __type;
end;

function DT_VERSIONTAGIDX(const tag: LongWord): LongWord;
begin
  Result := DT_VERNEEDNUM - tag; { Reverse order! }
end;

function DT_EXTRATAGIDX(const tag: Elf32_Word): Elf32_Word;
begin
  // DT_EXTRATAGIDX(tag)	((Elf32_Word)-((Elf32_Sword) (tag) <<1>>1)-1)
  Result := Elf32_Word(-(Elf32_Sword(tag) shl 1 shr 1) - 1);
end;

function ELF32_M_SYM(const info: Elf32_Word): Elf32_Word;
begin
  Result := info shr 8;
end;

function ELF32_M_SIZE(const info: Elf32_Word): Byte;
begin
  Result := info and $FF;
end;

function ELF32_M_INFO(const sym: Elf32_Word; const size: Byte): Elf32_Word;
begin
  Result := (sym shl 8) or Size;
end;

function ELF64_M_SYM(const info: Elf64_Xword): Elf64_Xword;
begin
  Result := info shr 8;
end;

function ELF64_M_SIZE(const info: Elf64_Xword): Byte;
begin
  Result := info and $FF;
end;

function ELF64_M_INFO(const sym: Elf64_Xword; const size: Byte): Elf64_Xword;
begin
  Result := (sym shl 8) or Size;
end;

end.
