/* **************************************************************************
 *                                                                          *
 * arith.c --- Arithmetic encoders.                                         *
 *                                                                          *
 * Copyright (c) 2001 Borland Software Corporation.                         *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

#include	"arith.h"

static	int		lenTab[16] =
{
	1,		/*	0000	*/
	2,      /*	0001	*/
	1,      /*	0010	*/
	3,      /*	0011	*/
	1,      /*	0100	*/
	2,		/*	0101	*/
	1,		/*	0110	*/
	4,		/*	0111	*/

	1,		/*	1000	*/
	2,		/*	1001	*/
	1,		/*	1010	*/
	3,		/*	1011	*/
	1,		/*	1100	*/
	2,		/*	1101	*/
	1,		/*	1110	*/
	5,		/*	1111	*/
};

static	unsigned char	shiftTab[16] =
{
	25,		/*	0000	*/
	18,      /*	0001	*/
	25,      /*	0010	*/
	11,      /*	0011	*/
	25,      /*	0100	*/
	18,		/*	0101	*/
	25,		/*	0110	*/
	4,		/*	0111	*/

	25,		/*	1000	*/
	18,		/*	1001	*/
	25,		/*	1010	*/
	11,		/*	1011	*/
	25,		/*	1100	*/
	18,		/*	1101	*/
	25,		/*	1110	*/
	0,		/*	1111	*/
};

unsigned char *Arith_WrUnsigned(unsigned char *d, unsigned long u)
{
	if	(u <= (1UL<<7)-1)
	{
		*d = u<<1;
		return(d + 1);
	}
	else if	(u <= (1UL<<14)-1)
	{
		*(unsigned short *)d = (u<<2) + 1;
		return(d + 2);
	}
	else if	(u <= (1UL<<21)-1)
	{
		*(unsigned long *)d = (u<<3) + 3;
		return(d + 3);
	}
	else if	(u <= (1UL<<28)-1)
	{
		*(unsigned long *)d = (u<<4) + 7;
		return(d + 4);
	}
	else
	{
		*d = 0x0F;
		*(unsigned long *)(d+1) = u;
		return(d + 5);
	}
}

unsigned char *Arith_WrSigned(unsigned char *d, long s)
{
	if	((-1L<<6) <= s && s <= (1L<<6)-1)
	{
		*d = s<<1;
		return(d + 1);
	}
	else if	((-1L<<13) <= s && s <= (1L<<13)-1)
	{
		*(unsigned short *)d = (s<<2) + 1;
		return(d + 2);
	}
	else if	((-1L<<20) <= s && s <= (1L<<20)-1)
	{
		*(unsigned long *)d = (s<<3) + 3;
		return(d + 3);
	}
	else if	((-1L<<27) <= s && s <= (1L<<27)-1)
	{
		*(unsigned long *)d = (s<<4) + 7;
		return(d + 4);
	}
	else
	{
		*d = 0x0F;
		*(unsigned long *)(d+1) = s;
		return(d + 5);
	}
}

unsigned long Arith_RdUnsigned(unsigned char **pSrc)
{
	int				temp;
	unsigned char *	src;

	src = *pSrc;
	temp = *src & 0xf;
	src += lenTab[temp];
	*pSrc = src;
	return *((unsigned long *)src - 1) >> shiftTab[temp];
}

long Arith_RdSigned(unsigned char **pSrc)
{
	int				temp;
	unsigned char *	src;

	src = *pSrc;
	temp = *src & 0xf;
	src += lenTab[temp];
	*pSrc = src;
	return *((long *)src - 1) >> shiftTab[temp];
}

