/* **************************************************************************
 *                                                                          *
 * arith.h --- Arithmetic encoders.                                         *
 *                                                                          *
 * Copyright (c) 2001-2002 Borland Software Corporation.                    *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

#ifndef	ARITH_H
#define	ARITH_H

/*
	Arithmetic Encoders

	The routines here attempt to find efficient encodings for generic
	arithmetic quantities, both in terms of space and in terms of speed
	for encoding and decoding.

	N.B.:  The encoders are not guaranteed to shrink the data.  There are
	cases where it can expand a 4 byte quantity to 5 bytes when encoded.
	Don't use them for a completely arbitrary case, and expect them to help
	you.  Make sure that your compressed stream is a win for your application.
*/

#ifdef __cplusplus
extern "C" {
#endif

/*
	These are the values that are the maximum number of bytes that
	Arith will put out for a signed/unsigned quantity.
*/
#define ARITH_MAX_COMPRESSED_SIGNED_SIZE	5
#define ARITH_MAX_COMPRESSED_UNSIGNED_SIZE	5

unsigned char *Arith_WrUnsigned(unsigned char *d, unsigned long u);
unsigned long Arith_RdUnsigned(unsigned char **pSrc);

unsigned char *Arith_WrSigned(unsigned char *d, long u);
long Arith_RdSigned(unsigned char **pSrc);

#ifdef __cplusplus
}
#endif

#endif
