/* ************************************************************************* *
 *                                                                           *
 * delphiuw.c --- Delphi Unwinder                                            *
 *                                                                           *
 * Copyright (c) 1994-2002 Borland Software Corporation                      *
 *                                                                           *
 * This file may be distributed under the terms of the GNU General Public    *
 * License (GPL) version 2 as published by the Free Software Foundation and  *
 * appearing at http://www.borland.com/kylix/gpl.html.                       *
 *                                                                           *
 * Licensees holding a valid Borland No-Nonsense License for this Software   *
 * may use this file in accordance with such license, which appears in the   *
 * file license.txt that came with this software.                            *
 *                                                                           *
 * ************************************************************************ */


#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<assert.h>

#include	"diag.h"
#include	"arith.h"
#include	"unwind.h"
#include	"unwindfi.h"
#include	"uwutil.h"

// Turn this off for shipping.  See objout.c in dcc32 for more info.
//#define	GEN_EXC_NAMES

#ifdef	DO_STATISTICS
static	int				RunningCodeSizeTotal;
static	int				RunningExcSizeTotal;
static	int				RunningUnwindSizeTotal;
static	int				RunningRangeSizeTotal;
static  int				RunningProcCountTotal;
static	int				RunningExcProcTotal;
static	int				RunningFramedProcTotal;
static	int				RunningPotentialSavings;
static	int				RunningNonFramedProcUSize;
static	int				RunningSmallProcCount;
static	int				RunningSmallProcRanges;
#endif

//#ifdef	__linux__
typedef unsigned long DWORD;
//#endif

//static void ReverseFrameInfo(const unsigned char *Src, unsigned char *Buffer, int BuffSize)
//{
//	Src -= BuffSize;
//	while	(BuffSize)
//		Buffer[--BuffSize] = *Src++;
//}

static int DecodeFrameInfo(unsigned char *Buffer, ExceptionFrameInfo *FrameInfo)
{
	unsigned char *	p;
	int				i;
	unsigned char *	Start;

	memset(FrameInfo, 0, sizeof(*FrameInfo));

	Start = p = Buffer;
	
	/*
		The flags byte is guaranteed.  Everything else is conditional.
	*/
	FrameInfo->Flags = *p++;

	if	(FrameInfo->Flags & UFI_F_CALLEEPOP)
		FrameInfo->CalleePopAmount = Arith_RdUnsigned(&p);

	if	(FrameInfo->Flags & UFI_F_SROFFS)
	{
		if	(FrameInfo->Flags & UFI_F_HASFRAME)
			FrameInfo->SavedRegsEBPOffs = Arith_RdUnsigned(&p);
		else
			FrameInfo->OffsetToPush = Arith_RdUnsigned(&p);
	}

	if	(FrameInfo->Flags & UFI_F_HASRANGES)
		FrameInfo->RangeCount = Arith_RdUnsigned(&p);

	if	(!(FrameInfo->Flags & UFI_F_HASFRAME) || (FrameInfo->Flags & UFI_F_NESTEDHANDLER))
	{
		FrameInfo->UnwindInfoSize = Arith_RdUnsigned(&p);
		FrameInfo->ProcLen = Arith_RdUnsigned(&p);
	}
	else if	(FrameInfo->Flags & UFI_F_HASRANGES)
	{
		FrameInfo->ProcLen = Arith_RdUnsigned(&p);
	}

	return p - Start;
}

static void *DecodeReversedFrameInfo(unsigned char *EndOfProcedure, ExceptionFrameInfo *FrameInfo)
{
	unsigned char	Buffer[ARITH_MAX_COMPRESSED_UNSIGNED_SIZE * 6 + 1];
	unsigned char *	p;
	int				i;
//	unsigned char *	Start;

	/*
		The frame info is compressed at the end of the unwind info and
		code ranges.  It is reversed, too, since we don't know the size
		of the info.  So we have to copy the bytes, reverse them and
		decompress them.  We copy enough bytes to make sure that we got
		all the compressed data.  This is safe, since there is always
		enough data on the other side of the frame info to make up for
		overruns.
	*/
	p = EndOfProcedure - sizeof(Buffer);
	i = sizeof(Buffer);
	while	(i)
		Buffer[--i] = *p++;
//	Start = p = Buffer;

	return (void *)(EndOfProcedure - DecodeFrameInfo(Buffer, FrameInfo));
}

#ifdef	DO_STATISTICS
static	void SplitName(const char *Name, char *BaseName, char *MemberName)
{
	char *	Member;
	Member = strchr(Name, '.');
	if	(Member)
	{
		strncpy(BaseName, Name, Member - Name);
		BaseName[Member - Name] = 0;
		strcpy(MemberName, Member + 1);
	}
	else
	{
		BaseName[0] = 0;
		strcpy(MemberName, Name);
	}
}

static void DumpStatistics(void *Context)
{
	int				UnitCount;
	PCMapUnit *		Units;
	int				i;
	char *			TextStart;
	PCMapHeader *	Header;
	int				CodeSizeTotal;
	int				ExcSizeTotal;
	int				UnwindSizeTotal;
	int				RangeSizeTotal;
	int				ProcCountTotal;
	char			LastBaseName[4096];
	char			BaseName[4096];
	char			MemberName[4096];

	LastBaseName[0] = 0;
	BaseName[0] = 0;

	Header = (PCMapHeader *)Context;

	TextStart = (char *)Context - Header->CodeSegmentSize;
	printf("Image Range: %08x, %08x\n", TextStart, Context);
	CodeSizeTotal = ExcSizeTotal = UnwindSizeTotal = RangeSizeTotal = 0;
	ProcCountTotal = 0;
	
	UnitCount = Header->UnitCount;
	Units = (PCMapUnit *)(Header + 1);
	printf("%-25s %-4s %-4s/%-7s (%-4s/%-7s, %-4s/%-7s)\n",
			"Procedure", "Code ", "Exc ", "%", "Unwd", "%", "Rngs", "%");
	for	(i = 0; i < UnitCount; i++)
	{
		PCMapProc *	Procs;
		int			ProcCount;
		int			j;
		Procs = (PCMapProc *)(((char *)Context) + Units[i].ProcInfoOffs);
		ProcCount = Units[i].ProcCount;
		ProcCountTotal += ProcCount;
		RunningProcCountTotal += ProcCount;
		for	(j = 0; j < ProcCount; j++)
		{
			ExceptionFrameInfo 		FrameInfo;
			int *					pNameLen;
			char *					Name;
			int						ExcRangeSize;
			int						ExcInfoSize;
			unsigned char *			EndOfUnwindInfo;
			unsigned char *			End;
			int						ProcLen;
			
			End = (unsigned char *)(TextStart + Procs[j].End);
			EndOfUnwindInfo = DecodeReversedFrameInfo(End, &FrameInfo);

			pNameLen = ((int *)EndOfUnwindInfo) - 1;
			Name = ((char *)pNameLen) - *pNameLen;
			SplitName(Name, BaseName, MemberName);
			if	(strcmp(BaseName, LastBaseName))
			{
				strcpy(LastBaseName, BaseName);
				if	(*BaseName)
					printf("%s\n", BaseName);
			}
			if	 (*BaseName)
				strcpy(BaseName, "  ");
			//      name procsize exc size/% (unwind size/%, range size/%)
			ExcRangeSize = FrameInfo.RangeCount * sizeof(ExceptionRangeOffsets);
			RunningExcProcTotal += FrameInfo.RangeCount;

			if	(!FrameInfo.RangeCount && (FrameInfo.Flags & UFI_F_HASFRAME))
				RunningFramedProcTotal++;
			if	(!FrameInfo.RangeCount && !(FrameInfo.Flags & UFI_F_HASFRAME))
				RunningNonFramedProcUSize += FrameInfo.UnwindInfoSize;
			if	(FrameInfo.RangeCount && (FrameInfo.Flags & UFI_F_HASFRAME))
				RunningPotentialSavings += FrameInfo.UnwindInfoSize;

			ExcInfoSize = (End - EndOfUnwindInfo) + FrameInfo.UnwindInfoSize + ExcRangeSize;

			if	(FrameInfo.ProcLen)
				ProcLen = FrameInfo.ProcLen;
			else
				ProcLen = Procs[j].End - Procs[j].Start - ExcInfoSize;
			if	(ProcLen <= 256 && FrameInfo.RangeCount)
			{
				RunningSmallProcCount++;
				RunningSmallProcRanges += FrameInfo.RangeCount;
			}

			CodeSizeTotal += ProcLen;
			ExcSizeTotal += ExcInfoSize;
			UnwindSizeTotal += FrameInfo.UnwindInfoSize;
			RangeSizeTotal += ExcRangeSize;
			
			RunningCodeSizeTotal += ProcLen;
			RunningExcSizeTotal += ExcInfoSize;
			RunningUnwindSizeTotal += FrameInfo.UnwindInfoSize;
			RunningRangeSizeTotal += ExcRangeSize;
			
			printf("%s%-*s %4d %4d/%5.2f (%4d/%5.2f, %4d/%5.2f)\n",
				   BaseName,
				   *BaseName ? 23 : 25,
				   MemberName,
				   ProcLen,
				   ExcInfoSize,
				   (float)ExcInfoSize / (float)ProcLen,
				   FrameInfo.UnwindInfoSize,
				   (float)FrameInfo.UnwindInfoSize / (float)ExcInfoSize * 100.0f,
				   ExcRangeSize,
				   (float)ExcRangeSize / (float)ExcInfoSize * 100.0f);
		}
	}
	printf("Totals: C(%d) E(%d) (%4.2f), U(%d) R(%d) (%4.2f, %4.2f), PC(%d)\n",
	       CodeSizeTotal,
	       ExcSizeTotal,
	       (float)ExcSizeTotal / (float)CodeSizeTotal,
	       UnwindSizeTotal,
	       RangeSizeTotal,
	       (float)UnwindSizeTotal / (float)ExcSizeTotal,
	       (float)RangeSizeTotal / (float)ExcSizeTotal,
	       ProcCountTotal);
	printf("RTotals: C(%d) E(%d) (%4.2f), U(%d) R(%d) (%4.2f, %4.2f), PC(%d)\n",
	       RunningCodeSizeTotal,
	       RunningExcSizeTotal,
	       (float)RunningExcSizeTotal / (float)RunningCodeSizeTotal,
	       RunningUnwindSizeTotal,
	       RunningRangeSizeTotal,
	       (float)RunningUnwindSizeTotal / (float)RunningExcSizeTotal,
	       (float)RunningRangeSizeTotal / (float)RunningExcSizeTotal,
	       RunningProcCountTotal);
	printf("Procs with exception handling: %d (%4.2f%% of total)\n",
		   RunningExcProcTotal, (float)RunningExcProcTotal / (float)RunningProcCountTotal * 100.0f);
	printf("Procs without exception handling, but with frames: %d (%4.2f%% of total)\n",
		   RunningFramedProcTotal, (float)RunningFramedProcTotal / (float)RunningProcCountTotal * 100.0f);
	printf("Procs without exception handling, no frames, unwind size: %d (%4.2f%% of total)\n",
		   RunningNonFramedProcUSize, (float)RunningNonFramedProcUSize / (float)RunningCodeSizeTotal * 100.0f);
	printf("Potential unwind space savings: %d (%4.2f%% of total => new percentage of %4.2f%%)\n",
			RunningPotentialSavings,
			(float)RunningPotentialSavings / (float)RunningExcSizeTotal * 100.0f,
			(float)(RunningExcSizeTotal - RunningPotentialSavings) / (float)RunningCodeSizeTotal * 100.0f);
	printf("Small procs with exception handling: %d (%d total ranges)\n",
		   RunningSmallProcCount, RunningSmallProcRanges);
}
#endif

static	ExceptionRangeOffsets *	FindRange(
	unsigned long 			ProcRelOffs,
	ExceptionRangeOffsets *	Ranges,
	int						RangeCount)
{
	int	i;

	/*
		Note the shift of the inclusion fence here:  The end is included,
		and the start is excluded.  This may cause a problem for OS exception
		dispatching if the faulting instruction is a one byte instruction at
		the end of the code range.
	*/
	for	(i = 0; i < RangeCount; i++)
		if	(ProcRelOffs > Ranges[i].Start && ProcRelOffs <= Ranges[i].End)
			return &Ranges[i];

	return NULL;
}

static	ExcRangeCPP *	FindRangeCPP(
	unsigned long 	ProcRelOffs,
	ExcRangeCPP *	Ranges,
	int				RangeCount)
{
	int	i;

	/*
		Note the shift of the inclusion fence here:  The end is included,
		and the start is excluded.  This may cause a problem for OS exception
		dispatching if the faulting instruction is a one byte instruction at
		the end of the code range.
	*/
//	DEBUG_PRINT("unwind.control.search",
//				("Looking for 0x%x in %d ranges\n", ProcRelOffs, RangeCount));
	for	(i = 0; i < RangeCount; i++)
	{
//		DEBUG_PRINT("unwind.control.search",
//					("  checking 0x%x - 0x%x\n", Ranges[i].Start, Ranges[i].End));
		if	(ProcRelOffs > Ranges[i].Start && ProcRelOffs <= Ranges[i].End)
		{
//			DEBUG_PRINT("unwind.control.search", ("    found\n"));
			return &Ranges[i];
		}
	}

	return NULL;
}

typedef	struct	DelphiDecoder
{
	unsigned char *	UnwindInfo;
	unsigned char *	EndUnwindInfo;
	int				DeltaESP;
	int				DeltaOffs;
	unsigned long	CurOffs;
	unsigned long	ProcStart;		//debugging
}	DelphiDecoder;

static	void	InitDelphiDecoder(
	DelphiDecoder *	Decoder,
	unsigned char *	Info,
	int				InfoSize,
	unsigned long ProcStart)
{
	Decoder->UnwindInfo = Info;
	Decoder->EndUnwindInfo = Info + InfoSize;
	Decoder->DeltaESP = 0;
	Decoder->DeltaOffs = 0;
	Decoder->CurOffs = 0;
	Decoder->ProcStart = ProcStart;
}

static	int	DecodeTo(
	DelphiDecoder *	Decoder,
	unsigned long TargetOffs)
{
	int	Res;
	int	DeltaESP;
	int	DeltaOffs;
	unsigned long	NewOffs;

	Res = 0;
	DeltaESP = Decoder->DeltaESP;
	DeltaOffs = Decoder->DeltaOffs;
	NewOffs = Decoder->CurOffs;
	DEBUG_CONTINUE("unwind.frame.decode", ("DecodeTo:\n"));
	while	(NewOffs + DeltaOffs <= TargetOffs &&
			 Decoder->UnwindInfo < Decoder->EndUnwindInfo)
	{
		NewOffs += DeltaOffs;
		Res += DeltaESP;
		DeltaESP = Arith_RdSigned(&Decoder->UnwindInfo);
		DeltaOffs = Arith_RdUnsigned(&Decoder->UnwindInfo);
		DEBUG_CONTINUE("unwind.frame.decode", ("%08x (%4d): DeltaEsp = %d\n",
											Decoder->ProcStart + NewOffs,
											NewOffs,
											Res));
	}
	Decoder->DeltaESP = DeltaESP;
	Decoder->DeltaOffs = DeltaOffs;
	Decoder->CurOffs = NewOffs;
	return Res;
}

static	int	BitCount(unsigned int Bits)
{
	int	Count = 0;

	while	(Bits)
	{
		Count += Bits & 1;
		Bits = Bits >> 1;
	}
	return Count;
}

/*
	The two structures below are borrowed from unwind.c.  This is improper, but this file
	was split from unwind.c late in the game.  The proper thing to do is to expand the SetGR/GetGR
	interface in unwind.c's implementation to accept the needed registers, and then we go
	through the APIs in this file.
*/
#ifdef	__IA32__
#define MAX_EXTENSION     512
typedef unsigned long	ULONG;
typedef unsigned char	UBYTE;
#define SIZE_FP_REGISTERS      80
typedef struct FP_SAVE
{
    ULONG   Control;
    ULONG   Status;
    ULONG   Tag;
    ULONG   ErrorOff;
    ULONG   ErrorSel;
    ULONG   DataOff;
    ULONG   DataSel;
    UBYTE   Registers[SIZE_FP_REGISTERS];
    ULONG   Cr0NpxState;
}	FP_SAVE;

/*
	Don't rearrange _Unwind_Context arbitrarily!  There's ASM code here which
	manipulates this directly.
*/
struct	_Unwind_Context
{
    ULONG   Eip;
    ULONG   Ebp;
    ULONG   Esp;
    ULONG   Ebx;
    ULONG   Dr0;
    ULONG   Dr1;
    ULONG   Dr2;
    ULONG   Dr3;
    ULONG   Dr6;
    ULONG   Dr7;
    FP_SAVE FloatSave;
    ULONG   SegGs;
    ULONG   SegFs;
    ULONG   SegEs;
    ULONG   SegDs;
    ULONG   Edi;
    ULONG   Esi;
    ULONG   Edx;
    ULONG   Ecx;
    ULONG   Eax;
    ULONG   SegCs;
    ULONG   EFlags;
    ULONG   SegSs;
    UBYTE   ExtendedRegisters[MAX_EXTENSION];
};
#endif

static	void	UnwindDelphiFrame(
	_Unwind_Context *	Context,
	unsigned char *		Info,
	ExceptionFrameInfo *FrameInfo,
	unsigned long 		ProcRelOffs,
	unsigned long		ProcStartAddr,
	unsigned long		StopAddr)
{
	unsigned long	PushRegsOffs;
	unsigned char	FrameFlags;
	int				ESPToPush;
	int				ESPToStopAddr;
	int				ESPToEIP;
	int				SavedRegs;
	DelphiDecoder	Decoder;
	int				CalleePopAmount;

	FrameFlags = FrameInfo->Flags;


#pragma message ("Use context APIs instead of direct manipulation")
	DEBUG_PRINT("unwind.frame.info", ("UnwindDelphiFrame (to %p)\n", StopAddr));
	DEBUG_CONTINUE("unwind.frame.info", ("    Unwind Info Size: %d bytes.", FrameInfo->UnwindInfoSize));
	DEBUG_CODE("unwind.frame.info", {
		DEBUG_CLASS("unwind.frame.info");
		if	(FrameFlags & UFI_F_HASFRAME)
		{
			DEBUG_CONTINUE_(("  EBP frame present\n"));
		}
		else
		{
			DEBUG_CONTINUE_(("  No frame present\n"));	
		}
				});
	DEBUG_CONTINUE("unwind.frame.info", ("    Pushed regs:"));
	SavedRegs = FrameInfo->Flags & (UFI_F_REG_EBX |
									UFI_F_REG_ESI |
									UFI_F_REG_EDI |
									UFI_F_REG_EBP);

	DEBUG_CODE("unwind.frame.info", {
		DEBUG_CLASS("unwind.frame.info");
		if	(SavedRegs)
		{
			if	(FrameFlags & UFI_F_REG_EBX) DEBUG_CONTINUE_((" EBX"));
			if	(FrameFlags & UFI_F_REG_ESI) DEBUG_CONTINUE_((" ESI"));
			if	(FrameFlags & UFI_F_REG_EDI) DEBUG_CONTINUE_((" EDI"));
			if	(FrameFlags & UFI_F_REG_EBP) DEBUG_CONTINUE_((" EBP"));
			if	(FrameInfo->Flags & UFI_F_HASFRAME)
			{
				DEBUG_CONTINUE_(("\n    EBP rel offs to saved regs: %08x\n", FrameInfo->SavedRegsEBPOffs));
			}
			else
			{
				DEBUG_CONTINUE_(("\n    Offset to push %08x\n", FrameInfo->OffsetToPush));
			}
		}
		else
		{
			DEBUG_CONTINUE_((" None\n"));
		}
		});

	CalleePopAmount = FrameInfo->CalleePopAmount;
	DEBUG_CONTINUE("unwind.frame.info", ("    Callee pops %d bytes.\n", CalleePopAmount));

	/*
		Under some circumstances, there is no unwind information whatsoever
		present.  This happens if there is a BP frame.  Also, if we have a frame,
		and we want to unwind this frame completely, we don't have to bother
		decoding the unwind info, we can just ditch the stack frame
		immediately.  If we have UFI_F_NESTEDHANDLER, and we're doing a partial
		unwind, we have to use the unwind info, because otherwise we might toss
		a crucial return address for a finally block if we just unwind the stack
		back to the saved regs push point.
	*/

#pragma message ("Optimize this code next time round")
/*
	We're not properly testing here to see if we really need the unwind info.
	We can still optimize the case where there is unwind info, but we are completely
	unwinding the frame, and there is an EBP frame.  Be careful, though - SavedRegs
	will have UFI_F_REG_EBP set in this case, which isn't really right, because
	it's overloaded to UFI_F_NESTEDHANDLER.
*/
	if	((FrameInfo->Flags & (UFI_F_HASFRAME | UFI_F_NESTEDHANDLER)) == UFI_F_HASFRAME)
	{
		/*
			We need to retrieve EBX, ESI and EDI, if necessary.  If we're
			unwinding the function completely, we toss the whole frame,
			otherwise we just reset ESP to the end of the prologue sequence.
		*/
		if	(!StopAddr)
		{
			if	(SavedRegs)
			{
				DWORD *	SavedRegsPtr;
	
				SavedRegsPtr = (DWORD *)(Context->Ebp + FrameInfo->SavedRegsEBPOffs);
				
				if	(SavedRegs & UFI_F_REG_EDI)
					Context->Edi = *SavedRegsPtr++;
				if	(SavedRegs & UFI_F_REG_ESI)
					Context->Esi = *SavedRegsPtr++;
				if	(SavedRegs & UFI_F_REG_EBX)
					Context->Ebx = *SavedRegsPtr++;
			}
			Context->Eip = *(DWORD *)(Context->Ebp + 4);
			Context->Esp = Context->Ebp + 4 + 4 + FrameInfo->CalleePopAmount;
			Context->Ebp = *((DWORD *)Context->Ebp);
		}
		else
		{
			DEBUG_CONTINUE("unwind.frame.info", ("    Resetting ESP to prologue (%p - %x - %x).\n", Context->Ebp, -FrameInfo->SavedRegsEBPOffs, BitCount(SavedRegs) * 4));
			Context->Esp = Context->Ebp + FrameInfo->SavedRegsEBPOffs - BitCount(SavedRegs) * 4;
		}
		return;
	}

	PushRegsOffs = FrameInfo->OffsetToPush;

	assert(!SavedRegs || (ProcRelOffs > PushRegsOffs));

	InitDelphiDecoder(&Decoder, Info, FrameInfo->UnwindInfoSize, ProcStartAddr);

	if	(SavedRegs)
	{
		DWORD *	SavedRegsPtr;

		/*
			Start munching the offset info until we pass the pushed regs.
		*/
		ESPToPush = DecodeTo(&Decoder, PushRegsOffs);

		/*
			If we're only unwinding partially, then we take the
			further step of finding the change in the stack across the
			entire range up to the stopping point provided us.
		*/
		if	(PushRegsOffs != StopAddr)
			ESPToStopAddr = DecodeTo(&Decoder, StopAddr) + ESPToPush;
		else
			ESPToStopAddr = ESPToPush;

		/*
			Now we read until we get to the current execution pointer,
			and we'll have all the stack mods we need.
		*/
		ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);

		/*
			Blow away the stack to the point where entered the protected range,
			then restore the saved regs, and do the rest of the cleanup.
		*/
		if	(!StopAddr)
		{
			SavedRegsPtr = (DWORD *)(Context->Esp - (ESPToEIP + (ESPToStopAddr - ESPToPush)));
			if	((SavedRegs & UFI_F_REG_EBP) && !(FrameInfo->Flags & UFI_F_HASFRAME))
				Context->Ebp = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_EDI)
				Context->Edi = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_ESI)
				Context->Esi = *SavedRegsPtr++;
			if	(SavedRegs & UFI_F_REG_EBX)
				Context->Ebx = *SavedRegsPtr++;
			Context->Esp -= ESPToStopAddr + ESPToEIP;
			Context->Eip = *((DWORD *)Context->Esp);
			Context->Esp += 4 + CalleePopAmount; // ditch call and any stack parms
			if	(FrameFlags & UFI_F_HASFRAME)
				Context->Ebp = *((DWORD *)Context->Ebp);
		}
		else
		{
			Context->Esp -= ESPToEIP;
		}
	}
	else
	{
		if	(StopAddr)
		{
			ESPToStopAddr = DecodeTo(&Decoder, StopAddr);
			ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);
			Context->Esp -= ESPToEIP;
		}
		else
		{
			/*
				No saved regs.  We just gather ESP mods.
			*/
			ESPToEIP = DecodeTo(&Decoder, ProcRelOffs);
	
			/*
				Got the ESP mods.  Now we undo them in the context, and
				update the EIP.
			*/
			Context->Esp -= ESPToEIP;
			Context->Eip = *((DWORD *)Context->Esp);
			Context->Esp += 4 + CalleePopAmount; // ditch call and any stack parms
			if	(FrameFlags & UFI_F_HASFRAME)
				Context->Ebp = *((DWORD *)Context->Ebp);
		}
	}
}

static	_Unwind_Reason_Code	UNWIND_CC DelphiPersonalityFn(
	int version,
	_Unwind_Action actions,
	_Unwind_Exception *exceptionObject,
	_Unwind_Context *context,
	void *ImplementationData)
{
	unsigned long *			BPChain;
	int						ExcInfoSize;
	ExceptionFrameInfo 		FrameInfo;
	unsigned char *			End;
	unsigned char *			EndOfUnwindInfo;
	unsigned long			ProcRelOffs;
	unsigned long			ProcStartAddr;
	int						i;
	ExceptionRangeOffsets *	Ranges;
	ExceptionRangeOffsets *	Range;
	unsigned char *			UnwindInfo;
	_Unwind_Context 		NewContext;
#ifdef	GEN_EXC_NAMES
	int *					pNameLen;
	char *					Name;
#endif

	// exceptionObject may be NULL if the IDE goes to ask us if we have a closest handler.
	// Otherwise, it better be a Delphi exception object.  If it's a CPP exception,
	// we really don't want it.
//	if (actions & _UA_SEARCH_PHASE &&
//		exceptionObject &&
//		exceptionObject->exception_class == UW_EXC_CLASS_BORLANDCPP)
//	{
//		return _URC_CONTINUE_UNWIND;
//	}
//	assert(!exceptionObject || (exceptionObject->exception_class == UW_EXC_CLASS_BORLANDDELPHI));

	NewContext = *context;
	End = (unsigned char *)ImplementationData;
	EndOfUnwindInfo = DecodeReversedFrameInfo(End, &FrameInfo);
#ifdef	GEN_EXC_NAMES
	pNameLen = ((int *)EndOfUnwindInfo) - 1;
	Name = ((char *)pNameLen) - *pNameLen;

	DEBUG_PRINT("unwind.control.phases", ("  Found %s", Name));
#else
	DEBUG_PRINT("unwind.control.phases", ("  Found %p", End - FrameInfo.ProcLen));
#endif

#ifdef	GEN_EXC_NAMES
	ProcStartAddr = ((unsigned long)EndOfUnwindInfo) - FrameInfo.ProcLen - 4 - *pNameLen -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount -
									FrameInfo.UnwindInfoSize;
#else
	ProcStartAddr = ((unsigned long)EndOfUnwindInfo) - FrameInfo.ProcLen -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount -
									FrameInfo.UnwindInfoSize;
#endif
	ProcRelOffs = _BorUnwind_GetIP(context) - ProcStartAddr;
#ifdef	GEN_EXC_NAMES
	Ranges = (ExceptionRangeOffsets *)((char *)EndOfUnwindInfo - *pNameLen - 4 - 
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount);
#else
	Ranges = (ExceptionRangeOffsets *)((char *)EndOfUnwindInfo -
									sizeof(ExceptionRangeOffsets) * FrameInfo.RangeCount);
#endif
	UnwindInfo = (unsigned char *)Ranges - FrameInfo.UnwindInfoSize;

	BPChain = (unsigned long *)_BorUnwind_GetGR(context, _UW_REG_EBP);
	if	(actions & _UA_SEARCH_PHASE)
	{
		/*
			Now we have to dig down into the local frame's ranges, and 
			see which part of the fn really wants this thing.
		*/
		Range = FindRange(ProcRelOffs, Ranges, FrameInfo.RangeCount);
		if	(Range)
		{
			/*
				Someone here wants the exception.  Fine.
			*/
			DEBUG_CONTINUE("unwind.control.phases",
						(" - I want the exception - %08x in [%08x, %08x)\n",
						_BorUnwind_GetIP(context),
						ProcStartAddr + Range->Start,
						ProcStartAddr + Range->End));
			return _URC_HANDLER_FOUND;
		}

		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}
	if	(actions & _UA_CLEANUP_PHASE)
	{
		if	(actions & _UA_HANDLER_FRAME)
		{
			/*
				Unwind the current frame, back to the start of the current range
			*/
			Range = FindRange(ProcRelOffs, Ranges, FrameInfo.RangeCount);
			assert(Range);

			DEBUG_CONTINUE("unwind.control.phases", (" - transfer control now\n"));
			UnwindDelphiFrame(context,
							  UnwindInfo,
							  &FrameInfo,
							  ProcRelOffs,
							  ProcStartAddr,
							  Range->Start);
			/*
				Got the handler.  Install it.
			*/
			_BorUnwind_SetIP(context, ProcStartAddr + Range->Handler);
			_BorUnwind_SetGR(context, _UW_REG_EBX, _BorUnwind_FindGOT(_BorUnwind_GetIP(context)));
			_BorUnwind_SetGR(context, _UW_REG_EAX, (_Unwind_RegVal)exceptionObject->private_1);
			_BorUnwind_SetGR(context, _UW_REG_EDX, (_Unwind_RegVal)exceptionObject->private_2);
			_BorUnwind_SetGR(context, _UW_REG_ECX, (_Unwind_RegVal)exceptionObject->exception_class);
			DEBUG_PRINT("unwind.control.phases",
						("Transferring control with new EIP = %08x, ESP = %08x, EBP = %08x\n",
						context->Eip, context->Esp, context->Ebp));
			return _URC_INSTALL_CONTEXT;
		}
		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}

	DEBUG_PRINT("unwind.control.phases",
				("Entered unwind with        EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));
	UnwindDelphiFrame(context,
					  UnwindInfo,
					  &FrameInfo,
					  ProcRelOffs,
					  ProcStartAddr,
					  0);
	DEBUG_PRINT("unwind.control.phases",
				("Continuing unwind with new EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));
	return _URC_CONTINUE_UNWIND;
}

static CPPhandlerProc CPPhandler;

UNWIND_PUBLIC void  UNWIND_CC _BorUnwind_SetCPPHandler(void *handler)
{
	CPPhandler = handler;
}

#define ALIGN(x,s) (((x) + (s) - 1) & ~((s) - 1))

static _Unwind_RegVal CopyArg(_Unwind_Context *context, void *arg, int size)
{
	void *pStack;
	pStack = (void *)_BorUnwind_GetGR(context, _UW_REG_ESP);
	pStack = (void *)(((char *)pStack) - ALIGN(size, sizeof(_Unwind_RegVal)));
	memcpy(pStack, arg, size);
	_BorUnwind_SetGR(context, _UW_REG_ESP, (_Unwind_RegVal)pStack);
	return (_Unwind_RegVal)pStack;
}

static void PushArg(_Unwind_Context *context, void *arg)
{
	void **pStack;

	pStack = (void **)_BorUnwind_GetGR(context, _UW_REG_ESP);
	pStack--;
	*pStack = arg;
	_BorUnwind_SetGR(context, _UW_REG_ESP, (_Unwind_RegVal)pStack);
}

static	_Unwind_Reason_Code	UNWIND_CC BCBPersonalityFn(
	int version,
	_Unwind_Action actions,
	_Unwind_Exception *exceptionObject,
	_Unwind_Context *context,
	void *ImplementationData)
{
	unsigned long *			BPChain;
	int						ExcInfoSize;
	ExceptionFrameInfo 		FrameInfo;
	void *					CtxtTable = 0;
	unsigned long			ProcRelOffs = 0;
	unsigned long			ProcStartAddr = 0;
	int						i;
	ExcRangeCPP *			Ranges;
	ExcRangeCPP *			Range;
	unsigned char *			UnwindInfo;
	unsigned char *			EndOfUnwindInfo;
	_Unwind_Reason_Code		Res;
#ifdef	GEN_EXC_NAMES
	int						NameLen;
	char *					Name;
#endif

	UnwindInfo = (unsigned char *)ImplementationData;
	EndOfUnwindInfo = UnwindInfo +
					  DecodeFrameInfo(UnwindInfo, &FrameInfo) +
					  FrameInfo.UnwindInfoSize;
	UnwindInfo = EndOfUnwindInfo - FrameInfo.UnwindInfoSize;
#ifdef	GEN_EXC_NAMES
	Name = (char *)EndOfUnwindInfo;
	NameLen = strlen(Name) + 1;
	Ranges = (ExcRangeCPP *)(Name + NameLen);

	DEBUG_PRINT("unwind.control.phases", ("  Found %s", Name));
#else
	Ranges = (ExcRangeCPP *)EndOfUnwindInfo;
	DEBUG_PRINT("unwind.control.phases", ("  Found %p", ProcStartAddr));
#endif

	//  We could get rid of the unconditional presence of the procedure
	//  start symbol here, with some more work in the compiler.  That's
	//  a space optimization that we should consider for later.

	// Pointer to the function is after the ranges

	ProcStartAddr = *(unsigned long *)(Ranges + FrameInfo.RangeCount);
	ProcRelOffs = _BorUnwind_GetIP(context) - ProcStartAddr;

	if (FrameInfo.RangeCount)
	{
		
		// Pointer to the C++ context table is after the pointer to the function
		
		CtxtTable = *(((void **)(Ranges + FrameInfo.RangeCount)) + 1);
	}

	BPChain = (unsigned long *)_BorUnwind_GetGR(context, _UW_REG_EBP);
	if	(actions & _UA_SEARCH_PHASE)
	{
		/*
			Now we have to dig down into the local frame's ranges, and 
			see which part of the fn really wants this thing.
		*/
		Range = FindRangeCPP(ProcRelOffs, Ranges, FrameInfo.RangeCount);
		if	(Range)
		{
			/*
				Someone here wants the exception.  Fine.
			*/
			DEBUG_CONTINUE("unwind.control.phases",
						(" - I want the exception - %08x in [%08x, %08x) dtc = %d, ctxoffs = %d\n",
						_BorUnwind_GetIP(context),
						ProcStartAddr + Range->Start,
						ProcStartAddr + Range->End,
						Range->DtorCnt, Range->Context));

			Res = CPPhandler(actions,
							 exceptionObject,
							 Range->Context,
							 Range->DtorCnt,
							 CtxtTable,
							 context);
			if (Res != _URC_CONTINUE_UNWIND)
				return Res;
		}

		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}
	if	(actions & _UA_CLEANUP_PHASE)
	{
		Range = FindRangeCPP(ProcRelOffs, Ranges, FrameInfo.RangeCount);
		if	(actions & _UA_HANDLER_FRAME)
		{
			_Unwind_RegVal argPtr;
			/*
				Unwind the current frame, back to the start of the current range
			*/
			assert(Range);

			DEBUG_CONTINUE("unwind.control.phases", (" - transfer control now\n"));
			UnwindDelphiFrame(context,
							  UnwindInfo,
							  &FrameInfo,
							  ProcRelOffs,
							  ProcStartAddr,
							  Range->Start);
			/*
				Got the handler.  Install it.
			*/
			argPtr = CopyArg(context, (void *)exceptionObject, sizeof(*exceptionObject));
			PushArg(context, (void *)context);
			PushArg(context, (void *)CtxtTable);
			PushArg(context, (void *)Range->DtorCnt);
			PushArg(context, (void *)Range->Context);
			PushArg(context, (void *)argPtr);
			PushArg(context, (void *)actions);
			PushArg(context, (void *)0); // return addr!
			_BorUnwind_SetIP(context, (_Unwind_RegVal)CPPhandler);
			_BorUnwind_SetGR(context, _UW_REG_EBX, _BorUnwind_FindGOT(_BorUnwind_GetIP(context)));

			DEBUG_PRINT("unwind.control.phases",
						("Transferring control with new EIP = %08x, ESP = %08x, EBP = %08x\n",
						context->Eip, context->Esp, context->Ebp));
			return _URC_INSTALL_CONTEXT;
		}
#ifdef GEN_EXC_NAMES
		DEBUG_PRINT("unwind.cleanup", ("Unwind cleanup for %s\n", Name));
#endif
		if (FrameInfo.RangeCount)
			CPPhandler(actions,
					   exceptionObject,
					   Range->Context,
					   Range->DtorCnt,
					   CtxtTable,
					   context);
#ifdef GEN_EXC_NAMES
		DEBUG_PRINT("unwind.cleanup", ("Done with cleanup for %s\n", Name));
#endif
		DEBUG_CONTINUE("unwind.control.phases", ("\n"));
	}

	DEBUG_PRINT("unwind.control.phases",
				("Entered unwind with        EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));
	UnwindDelphiFrame(context,
					  UnwindInfo,
					  &FrameInfo,
					  ProcRelOffs,
					  ProcStartAddr,
					  0);
	DEBUG_PRINT("unwind.control.phases",
				("Continuing unwind with new EIP = %08x, ESP = %08x, EBP = %08x\n",
				context->Eip, context->Esp, context->Ebp));

	if (_BorUnwind_GetUnhandledExceptionFilter())
	{
		if (ProcStartAddr == _BorUnwind_GetUEFOuterProc())
			(_BorUnwind_GetUnhandledExceptionFilter())();
	}

	return _URC_CONTINUE_UNWIND;
}

UNWIND_PUBLIC __personality_routine	UNWIND_CC _BorUnwind_BCBLookup(
	_Unwind_codePtr Addr,
	void *			Context,
	void **			pImplementationData)
{
	int				ProcCount;
	PCMapProcCPP *	Procs;
	int				i;
	char *			TextStart;
	long			AddrOffs;
	PCMapHeader *	Header;
	int				high, low, mid;
	PCMapProcCPP *	elt;

	Header = (PCMapHeader *)Context;

if	(Header->Signature != PCMAPHEADER_SIGNATURE2)
{
    printf("Signature = %x\n", Header->Signature);
	printf("FATAL: version mismatch in unwind.so.  Maybe you need to update the compiler/unwind.so?\n");
	assert(!"Have to fail here");
	Header = (PCMapHeader *)((char *)Context + 1);
//if	(Header->Signature != PCMAPHEADER_SIGNATURE2)
//{
//    printf("Signature = %x\n", Header->Signature);
//	assert(!"Have to fail here");
//}
}
	TextStart = (char *)Context - Header->CodeSegmentSize;
	ProcCount = Header->UnitCount;

	AddrOffs = Addr - (unsigned long)TextStart;
	DEBUG_PRINT("unwind.control.search", ("Searching for %p in %d procs\n", AddrOffs, ProcCount));
	Procs = (PCMapProcCPP *)(Header + 1);

	/*
  char  *kmin, *probe;
  int i, j;

  kmin = (char *) base;
  while (nelem > 0){
    i = nelem >> 1;
    probe = kmin + i * width;
    j = (*fcmp)(key, probe);
    if (j == 0)
      return(probe);
    else if (j < 0)
      nelem = i;
    else  {
      kmin = probe + width;
      nelem = nelem - i - 1;
    }
  }
  */

	while	(ProcCount > 0)
	{
		i = ProcCount >> 2;
		elt = Procs + i;
	    DEBUG_PRINT("unwind.control.search", ("Checking in proc [%p, %p]\n",
						  elt->Start, elt->End));
		if	(AddrOffs >= elt->Start && AddrOffs < elt->End)
		{
			// Found the proc.
			if (elt->ImplementationData != -1)
			{
				if	(pImplementationData)
					*pImplementationData = (void *)(TextStart + elt->ImplementationData);
				return BCBPersonalityFn;
			}
			else
			{
				if	(pImplementationData)
					*pImplementationData = (void *)(TextStart + elt->End);
				return DelphiPersonalityFn;
			}
		}
		else if (AddrOffs < elt->Start)
		{
			ProcCount = i;
		}
		else
		{
			Procs = elt + 1;
			ProcCount = ProcCount - i - 1;
		}
	}
	return NULL;
}

UNWIND_PUBLIC __personality_routine	UNWIND_CC _BorUnwind_DelphiLookup(
	_Unwind_codePtr Addr,
	void *			Context,
	void **			pImplementationData)
{
	int				UnitCount;
	PCMapUnit *		Units;
	int				i;
	char *			TextStart;
	long			AddrOffs;
	PCMapHeader *	Header;

	Header = (PCMapHeader *)Context;

if	(Header->Signature != PCMAPHEADER_SIGNATURE)
{
	printf("FATAL: version mismatch in unwind.so.  Maybe you need to update the compiler/unwind.so?\n");
	assert(!"Have to fail here");
}

	TextStart = (char *)Context - Header->CodeSegmentSize;
	UnitCount = Header->UnitCount;
	Units = (PCMapUnit *)(Header + 1);
	AddrOffs = Addr - (unsigned long)TextStart;
	DEBUG_PRINT("unwind.control.search", ("Searching for %p\n", AddrOffs));
	for	(i = 0; i < UnitCount; i++)
	{
	    DEBUG_PRINT("unwind.control.search", ("Checking in unit [%p, %p]\n",
						  Units[i].Start, Units[i].End));
		if	(AddrOffs >= Units[i].Start && AddrOffs < Units[i].End)
		{
			// Found it.  Now find the proc.
			PCMapProc *	Procs;
			int			ProcCount;
			int			j;
			Procs = (PCMapProc *)(((char *)Context) + Units[i].ProcInfoOffs);
			ProcCount = Units[i].ProcCount;
			for	(j = 0; j < ProcCount; j++)
			{
			    DEBUG_PRINT("unwind.control.search", ("  Checking in proc [%p, %p]\n",
						  Procs[j].Start, Procs[j].End));
				if	(AddrOffs >= Procs[j].Start && AddrOffs < Procs[j].End)
				{
					// Found the proc.
					if	(pImplementationData)
						*pImplementationData = (void *)(TextStart + Procs[j].End);
					return DelphiPersonalityFn;
				}
			}
		}
	}
	return NULL;
}

int _BorUnwind_DelphiForeignUnwindHook(struct _Unwind_Context *	context)
{
	unsigned char *p;
	unsigned char *sig;

	DEBUG_PRINT("unwind.control.phases", ("foreign unwind\n"));

	/*
		Recognize a Borland Delphi SOAP thunk and adjust the stack.
	*/

	p = (unsigned char *)_BorUnwind_GetGR(context, _UW_REG_EIP);
	/* Check for RET and RET nn instructions */
	if (*p == 0xc3)
	{
		sig = p + 1;
	}
	else
	{
		if (*p == 0xc2)
		{
			sig = p + 3;
		}
		else
		{
			return 0;
		}
	}

	/*
		Is there a signature here that makes this a Borland thunk that
		we know how to unwind?  The signature is generated to the
		thunk by the Pascal SOAP RTL.  Signature value taken from
		the SOAP RTL.
	*/
	if (*(unsigned short *)sig == 0xfbad)
	{
		int add = 0;
		unsigned long newEIP;
		unsigned long *stackPtr;

		/*
			Signature checked out.  Unwind what we now know is a SOAP
			thunk.
		*/

		if (*p == 0xc2)
			add = *(unsigned short *)(p + 1);
		stackPtr = (unsigned long *)_BorUnwind_GetGR(context, _UW_REG_ESP);
		newEIP = *stackPtr;

		/*
			We adjust ESP by a return address, plus any bytes the RET
			instruction may have been planning on removing.
		*/
		stackPtr += (add >> 2) + 1;

		_BorUnwind_SetGR(context, _UW_REG_ESP, (_Unwind_RegVal)stackPtr);
		_BorUnwind_SetGR(context, _UW_REG_EIP, (_Unwind_RegVal)newEIP);
		return 1;
	}

	return 0;
}

UNWIND_PUBLIC void	UNWIND_CC _BorUnwind_DumpUnwindInfo(void *UnwindInfo)
{
	ExceptionFrameInfo	FrameInfo;
	int 				FrameInfoLen;
	const char *		name;
	int					SavedRegs;
	unsigned char		FrameFlags;

	FrameInfoLen = DecodeFrameInfo(UnwindInfo, &FrameInfo);
//	nameLen = (int *)((char *)UnwindInfo + FrameInfoLen);
	name = (char *)UnwindInfo + FrameInfoLen;
	printf("Name: %s\n", name);
	FrameFlags = FrameInfo.Flags;
	if	(FrameFlags & UFI_F_HASFRAME)
		printf(" EBP frame present\n");
	else
		printf(" No frame present\n");
	SavedRegs = FrameFlags & (UFI_F_REG_EBX |
							  UFI_F_REG_ESI |
							  UFI_F_REG_EDI |
							  UFI_F_REG_EBP);
	if	(SavedRegs)
	{
		printf(" Saved regs: ");
		if	(FrameFlags & UFI_F_REG_EBX) printf(" EBX");
		if	(FrameFlags & UFI_F_REG_ESI) printf(" ESI");
		if	(FrameFlags & UFI_F_REG_EDI) printf(" EDI");
		if	(FrameFlags & UFI_F_REG_EBP) printf(" EBP");
		printf("\n");
		if	(FrameFlags & UFI_F_HASFRAME)
			printf("   EBP rel offs to saved regs: %08x\n", FrameInfo.SavedRegsEBPOffs);
		else
			printf("   Offset to push %08x\n", FrameInfo.OffsetToPush);
	}
	else
	{
		printf(" No saved regs\n");
	}
}

#ifdef	INTERNAL_DEBUGGING
UNWIND_PUBLIC void	UNWIND_CC _BorUnwind_CrawlStack(void)
{
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_RegVal			CatchIP;

	GetContext(&Context);

	DEBUG_PRINT("unwind.control.crawl", ("CrawlStack\n"));

	for	(;;)
	{
		ExceptionFrameInfo 		FrameInfo;
		unsigned char *			End;
		unsigned char *			EndOfUnwindInfo;
		int *					pNameLen;
		char *					Name;

		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.crawl", (" Unable to find personality\n"));
			return;
		}

		EndOfUnwindInfo = DecodeReversedFrameInfo((unsigned char *)PersonalityData, &FrameInfo);
		pNameLen = ((int *)EndOfUnwindInfo) - 1;
		Name = ((char *)pNameLen) - *pNameLen;
		DEBUG_PRINT("unwind.control.crawl", ("%s\n", Name));

		if ((Personality)(_UNWIND_VERSION,
					  _UA_SEARCH_PHASE,
					  NULL,
					  &Context,
					  PersonalityData) == _URC_HANDLER_FOUND)
		{
			DEBUG_PRINT("unwind.control.crawl", ("--- Handler found; end of trace ---\n"));
			return;
		}
	}
}
#endif

