/* **************************************************************************
 *                                                                          *
 * unwind.c                                                                 *
 *                                                                          *
 * Copyright (c) 2002 Borland Software Corporation.                         *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

/* unwind.c --- Exception stack unwind routines as defined by the
        C++ ABI for IA-64:Exception Handling specification.  

	Note: this routine does not follow the "glue code" mechanisms
	    used by other util-common functions.
	    The implementations of assert, debug print, error handling 
	    and memory allocation used within this module are assumed to
	    be sufficient for all clients.  */

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<assert.h>
#ifdef	__linux__
#include	<signal.h>
#endif

#include	"diag.h"
#include	"arith.h"
#include	"unwind.h"
#include	"unwindfi.h"

/*
	The following debug classes are supported here:

"unwind.control.phases"
"unwind.search.image"
"unwind.frame.decode"
"unwind.frame.info"
*/

typedef struct ForeignUnwindHookList
{
    struct ForeignUnwindHookList *	Next;
    _BorUnwind_ForeignUnwindProc	Proc;
}	ForeignUnwindHookList;

static ForeignUnwindHookList 	ForeignUnwindHooks;

/*
    This implementation is limited for now to allow only one hook.  Easy to
    extend later.
*/

UNWIND_PUBLIC int _BorUnwind_SetForeignUnwindHook(_BorUnwind_ForeignUnwindProc fn)
{
    if (ForeignUnwindHooks.Proc)
	return 0;
    ForeignUnwindHooks.Proc = fn;
}

UNWIND_PUBLIC void _BorUnwind_ClearForeignUnwindHook(_BorUnwind_ForeignUnwindProc fn)
{
    assert(ForeignUnwindHooks.Proc == fn);
    ForeignUnwindHooks.Proc = 0;
}

typedef	struct	LookupList
{
	_Unwind_IPLookupFn	Func;
	_Unwind_codePtr		Start;
	_Unwind_codePtr		End;
	_Unwind_codePtr		GOT;
	void *				Context;
	struct LookupList *	Next;
	struct LookupList *	Prev;
}	LookupList;

static	LookupList *	LookList;

static _Unwind_codePtr _BorUnwind_UEFOuterProc;
static void			(*_BorUnwind_UnhandledExceptionFilter)(void);


static	int				RunningCodeSizeTotal;
static	int				RunningExcSizeTotal;
static	int				RunningUnwindSizeTotal;
static	int				RunningRangeSizeTotal;
static  int				RunningProcCountTotal;
static	int				RunningExcProcTotal;
static	int				RunningFramedProcTotal;
static	int				RunningPotentialSavings;
static	int				RunningNonFramedProcUSize;
static	int				RunningSmallProcCount;
static	int				RunningSmallProcRanges;

static __personality_routine	FindPersonality(
	struct _Unwind_Context *	CPUContext,
	_Unwind_codePtr 			IP,
	void **						pImplementationData,
	int							DestroySignatures);


UNWIND_PUBLIC int UNWIND_CC _BorUnwind_RegisterIPLookup(
	_Unwind_IPLookupFn 	LookupFn,
	_Unwind_codePtr 	Start,
	_Unwind_codePtr 	End,
	void *				Context,
	_Unwind_codePtr		GOT)
{
	LookupList *	NewEntry;

#if	DEBUG_CODE_ON
	{
		static int initialized = 0;
		if	(!initialized)
		{
			DEBUG_INIT();
			initialized = 1;
		}
	}
#endif

//	printf("register! %p-%p, %p\n", Start, End, Context);
//DumpStatistics(Context);
	 DEBUG_PRINT("unwind.register", ("registered %p-%p\n", Start, End));

	NewEntry = (LookupList *) malloc(sizeof(*NewEntry));
	if	(!NewEntry)
		return 0;

	NewEntry->Func = LookupFn;
	NewEntry->Start = Start;
	NewEntry->End = End;
	NewEntry->GOT = GOT;
	NewEntry->Context = Context;

	if	(LookList)
	{
		NewEntry->Next = LookList;
		NewEntry->Prev = LookList->Prev;
		LookList = NewEntry;
		LookList->Prev->Next = NewEntry;
		LookList->Next->Prev = NewEntry;
	}
	else
	{
		LookList = NewEntry;
		NewEntry->Next = NewEntry;
		NewEntry->Prev = NewEntry;
	}
	return 1;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_UnregisterIPLookup(_Unwind_codePtr Start)
{
	LookupList *	Temp;

	assert(LookList);
	Temp = LookList;
	do
	{
		if	(Temp->Start == Start)
		{
			if	(LookList == Temp)
			{
				LookList = Temp->Next;
				if	(LookList == Temp)
					LookList = NULL;
			}
			Temp->Next->Prev = Temp->Prev;
			Temp->Prev->Next = Temp->Next;
			free(Temp);
			break;
		}
		Temp = Temp->Next;
	}	while	(Temp != LookList);
}

UNWIND_PUBLIC _Unwind_codePtr UNWIND_CC _BorUnwind_FindGOT(_Unwind_codePtr IP)
{
	LookupList *			Temp;

	if	(LookList)
	{
		Temp = LookList;
		do
		{
			if	(IP >= Temp->Start && IP < Temp->End)
			{
				DEBUG_PRINT("unwind.control.phases", ("GOT for EIP %p is %p\n", IP, Temp->GOT));
				return Temp->GOT;
			}
			Temp = Temp->Next;
		}	while	(Temp != LookList);
	}
	return 0;
}

#ifdef	__IA32__
#define MAXIMUM_SUPPORTED_EXTENSION     512
#define	DWORD unsigned long
#define BYTE unsigned char
#define SIZE_OF_80387_REGISTERS      80
typedef struct _FLOATING_SAVE_AREA {
    DWORD   ControlWord;
    DWORD   StatusWord;
    DWORD   TagWord;
    DWORD   ErrorOffset;
    DWORD   ErrorSelector;
    DWORD   DataOffset;
    DWORD   DataSelector;
    BYTE    RegisterArea[SIZE_OF_80387_REGISTERS];
    DWORD   Cr0NpxState;
} FLOATING_SAVE_AREA;
struct	_Unwind_Context
{
    DWORD   Eip;	// +0
    DWORD   Ebp;	// +4
    DWORD   Esp;	// +8
    DWORD   Ebx;	// +12
    DWORD   Edi;	// +16
    DWORD   Esi;	// +20
    DWORD   Edx;	// +24
    DWORD   Ecx;	// +28
    DWORD   Eax;	// +32

    DWORD   Dr0;
    DWORD   Dr1;
    DWORD   Dr2;
    DWORD   Dr3;
    DWORD   Dr6;
    DWORD   Dr7;
    FLOATING_SAVE_AREA FloatSave;
    DWORD   SegGs;
    DWORD   SegFs;
    DWORD   SegEs;
    DWORD   SegDs;
    DWORD   SegCs;
    DWORD   EFlags;
//    DWORD   Esp;
    DWORD   SegSs;
    BYTE    ExtendedRegisters[MAXIMUM_SUPPORTED_EXTENSION];
};

typedef	struct	ForeignCallInfo
{
	unsigned int	Signature;
	unsigned int	SavedEBP;
	unsigned int	ParamBytes;
}	ForeignCallInfo;

static int	CheckAlternatives(struct _Unwind_Context *CPUContext, int DestroySignatures)
{
	_Unwind_codePtr *	StackPtr;
	ForeignUnwindHookList *	Hooks;

	/*
		We couldn't find the given address in the PC map.  We will try looking
		at user installed filters first.  If that fails, then we'll try looking
		up the stack for the magic markers that indicate we made a foreign
		call.  If we find one, we ditch all the stack up to that point, reset
		the EIP in the context, and restart the search from there.  If we
		don't find one, we bail out completely.
	*/
	DEBUG_PRINT("unwind.search.image", ("   Checking alternatives\n"));
	
	Hooks = &ForeignUnwindHooks;
	while (Hooks)
	{
		DEBUG_PRINT("unwind.control.phases", ("searching hooks\n"));
		if (Hooks->Proc)
			if ((Hooks->Proc)(CPUContext))
				return 1;
		Hooks = Hooks->Next;
	}

	/*
		This should really be done by using the ForeignUnwindHooks interface,
		but there is a bug in the Delphi linker's obj importer that prevents us
		from properly handling function pointer relocations, so we have to
		do this for the short term, and do the proper fix in an update.
	*/
	if (_BorUnwind_DelphiForeignUnwindHook(CPUContext))
		return 1;

	for	(StackPtr = (_Unwind_codePtr *)CPUContext->Esp;
		 *StackPtr != UFI_TOPOFSTACK;
		 StackPtr++)
	{
		if	(*StackPtr == UFI_STACK_MARKER)
		{
			unsigned			ParamBytes;
			_Unwind_codePtr		RetAddr;
			ForeignCallInfo *	Info;

			if	(DestroySignatures)
			{
				DEBUG_PRINT("unwind.search.image", ("   Found marker, destroying signature\n"));
				*StackPtr = *StackPtr + 1;
			}

			Info = (ForeignCallInfo *)StackPtr;

			ParamBytes = Info->ParamBytes & ~0x80000000;
			// Better be dword aligned!
			DEBUG_PRINT("unwind.search.image", ("   Found marker, Info->ParamBytes = %08x\n", Info->ParamBytes));
			assert(!(Info->ParamBytes & 3));
			RetAddr = *(_Unwind_codePtr *)(((char *)Info) - ParamBytes - 4);
			CPUContext->Eip = RetAddr;
			CPUContext->Ebp = Info->SavedEBP;
			if	(Info->ParamBytes & 0x80000000)
				CPUContext->Esp = ((DWORD)Info);
			else
				CPUContext->Esp = ((DWORD)Info) - ParamBytes;
			DEBUG_PRINT("unwind.search.image", ("   Found marker, RA(0x%08x) EBP(0x%08x) ESP(0x%08x)\n",
				CPUContext->Eip, CPUContext->Ebp, CPUContext->Esp));
			return 1;
		}
	}
	DEBUG_PRINT("unwind.search.image", ("   No markers found\n"));
	return 0;
}

static __personality_routine	FindPersonality(
	struct _Unwind_Context *	CPUContext,
	_Unwind_codePtr 			IP,
	void **						pImplementationData,
	int							DestroySignatures)
{
	LookupList *			Temp;
	__personality_routine	Personality;

	DEBUG_PRINT("unwind.search.image", ("   Looking for %x\n", IP));
restart:
	for	(;;)
	{
		if	(LookList)
		{
			Temp = LookList;
			do
			{
//				DEBUG_PRINT("unwind.search.image", ("   Looking for %x in (%x, %x)\n",
//													IP, Temp->Start, Temp->End));
				if	(IP >= Temp->Start && IP < Temp->End)
				{
					for	(;;)
					{
						Personality = (Temp->Func)(IP, Temp->Context, pImplementationData);
						if	(Personality)
						{
							LookList = Temp;
//							DEBUG_PRINT("unwind.search.image", ("   Found personality %p\n", Personality));
							return Personality;
						}
						if	(!CPUContext || !CheckAlternatives(CPUContext, DestroySignatures))
							return NULL;
						IP = CPUContext->Eip;
						goto restart;
					}
				}
				Temp = Temp->Next;
			}	while	(Temp != LookList);
		}
		if	(!CPUContext || !CheckAlternatives(CPUContext, DestroySignatures))
			break;
		IP = CPUContext->Eip;
	}
	return NULL;
}

#ifdef	__WIN32__

static __declspec(naked)	__cdecl void GetContext(struct _Unwind_Context *Context)
{
	asm
	{
		push	eax
		push	ecx
		mov		eax, dword ptr [esp + 12]	// Get Context
		mov		ecx, dword ptr [ebp + 4]	// EIP in foreign RTL
		mov		dword ptr [eax], ecx		// save EIP in context
		mov		ecx, [ebp]					// EBP in foreign RTL
		mov		dword ptr [eax + 4], ecx	// save EBP in context
		mov		ecx, ebp					// ESP of parent, at the frame
		add		ecx, 4 + 4					// account for EBP, and RA to foreign RTL
		mov		dword ptr [eax + 8], ecx	// save calced ESP in context
//		mov		dword ptr [eax + 8], esp	// ESP
		pop		ecx
		pop		eax
		ret
	}
#pragma message ("have to do more here")
}
#else // __WIN32__
static void GetContext(struct _Unwind_Context * Context)
{
#ifdef __GNUC__
	asm("movl (%ebp), %eax");			// get parent EBP
//	asm("movl (%eax), %eax");			// One more 
	asm("movl 4(%eax), %edx");			// RA from parent to foreign RTL
	asm("movl 8(%ebp), %ecx");			// ECX gets ptr to context
	asm("movl %edx, (%ecx)");			// save user EIP
//	asm("lea %eax + 4 + 4, %edx");		// Calculate ESP in foreign RTL
	asm("movl %eax, %edx");				// Calculate ESP in foreign RTL
	asm("add $8, %edx");
	asm("movl (%eax), %eax");			// get foreign EBP
	asm("movl %eax, 4(%ecx)");			// save foreign EBP in context
	asm("movl %edx, 8(%ecx)");			// save foreign ESP in context
//	asm("movl %esp, 8(%ecx)");			// save 'user' ESP
#endif
#ifdef __BORLANDC__
	asm
	{
		mov eax, [ebp]			// get parent EBP
		mov	edx, [eax+4]			// RA from parent to foreign RTL
		mov	ecx, [ebp+8]			// ECX gets ptr to context
		mov [ecx], edx			// save user EIP
		mov edx, eax				// Calculate ESP in foreign RTL
		add edx, 8
		mov eax, [eax]			// get foreign EBP
		mov [ecx+4], eax			// save foreign EBP in context
		mov [ecx+8], edx			// save foreign ESP in context
	}
#endif
}
#endif // __WIN32__
#endif // __IA32__

UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetIP(const struct _Unwind_Context *Context)
{
	return Context->Eip;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetIP(struct _Unwind_Context *Context, _Unwind_RegVal NewValue)
{
	Context->Eip = NewValue;
}

UNWIND_PUBLIC _Unwind_RegVal UNWIND_CC _BorUnwind_GetGR(const struct _Unwind_Context *Context, int Index)
{
	switch	(Index)
	{
	case	_UW_REG_EBP:
		return Context->Ebp;
		break;
	case	_UW_REG_ESP:
		return Context->Esp;
		break;
	case	_UW_REG_EBX:
		return Context->Ebx;
		break;
	case	_UW_REG_EIP:
		return Context->Eip;
		break;
	default:
		assert(!"Not implemented");
	}
	return 0;
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_SetGR(
	struct _Unwind_Context *	Context,
	int 						Index,
	_Unwind_RegVal 				NewValue)
{
	switch	(Index)
	{
	case	_UW_REG_EBP:
		Context->Ebp = NewValue;
		break;
	case	_UW_REG_ESP:
		Context->Esp = NewValue;
		break;
	case	_UW_REG_EBX:
		Context->Ebx = NewValue;
		break;
	case	_UW_REG_EAX:
		Context->Eax = NewValue;
		break;
	case	_UW_REG_EDX:
		Context->Edx = NewValue;
		break;
	case	_UW_REG_ECX:
		Context->Ecx = NewValue;
		break;
	case	_UW_REG_EIP:
		Context->Eip = NewValue;
		break;
	default:
		assert(!"Not implemented");
	}
}

static	void	InstallAndResume(
	_Unwind_Context *	Context,			// ebp + 8
	_Unwind_Exception *exception_object,	//ebp + 12
	_Unwind_RegVal CatchIP,
	__personality_routine	Personality)
{
#ifdef	__IA32__
#ifdef	__WIN32__
	assert(!"Not implemented");
#else
#ifdef __GNUC__
	asm("movl 8(%ebp), %ecx");		// ecx -> context
	asm("movl 4(%ecx), %ebp");		// ebp -> target EBP
	asm("movl 12(%ecx), %ebx");		// ebx -> target EBX
	asm("movl 16(%ecx), %edi");		// edi -> target EDI
	asm("movl 20(%ecx), %esi");		// esi -> target ESI
	asm("movl 24(%ecx), %edx");		// edx -> target EDX
	asm("movl 32(%ecx), %eax");		// eax -> target EAX
	asm("movl 8(%ecx), %esp");		// esp -> target ESP
	asm("pushl 0(%ecx)");			// push target EIP
	asm("movl 28(%ecx), %ecx");		// ecx -> target ECX
	asm("ret");
#endif
#ifdef __BORLANDC__
	asm
	{
		mov ecx, [ebp+8]		// ecx -> context
		mov ebp, [ecx+4]		// restore ebp
		mov ebx, [ecx+12]		// restore ebx
		mov edi, [ecx+16]		// restore edi
		mov esi, [ecx+20]		// restore esi
		mov edx, [ecx+24]		// restore edx
		mov eax, [ecx+32]		// restore eax
		mov esp, [ecx+8]		// restore esp
		push [ecx+0]			// push target EIP
		mov ecx, [ecx+28]		// restore ecx
		ret						// off we go
	}
#endif // __BORLANDC__
#endif // __WIN32__
#endif // __IA32__
}

UNWIND_PUBLIC void UNWIND_CC _BorUnwind_Resume(_Unwind_Exception *exception_object)
{
	assert(!"Don't quite know how to implement this");
}

#ifdef	__linux__
#ifdef __BORLANDC__
static __declspec(naked) _Unwind_codePtr GetFSChain(void)
#else
static _Unwind_codePtr GetFSChain(void)
#endif
{
#if defined(GEN_IA64)
	// FIXME 
#else
#ifdef __GNUC__
	asm("xor %eax, %eax");
	asm("mov %fs:(%eax), %eax");
#endif
#ifdef __BORLANDC__
	asm
	{
		xor eax, eax
		mov eax, fs:[eax]
		ret
	}
#endif
#endif // GEN_IA64
}

UNWIND_PUBLIC _Unwind_codePtr UNWIND_CC _BorUnwind_ClosestDelphiHandler(
	struct sigcontext *SContext)
{
#if defined(GEN_IA64)
	// FIXME 
#else
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_RegVal			CatchIP;
	_Unwind_Reason_Code		Res;

	DEBUG_PRINT("unwind.control.phases", ("Looking for closest handler\n"));
	
	Context.Ebp = SContext->ebp;
	Context.Esp = SContext->esp;
	Context.Eip = SContext->eip;
	Context.Eax = SContext->eax;
	Context.Ebx = SContext->ebx;
	Context.Ecx = SContext->ecx;
	Context.Edx = SContext->edx;
	Context.Edi = SContext->edi;
	Context.Esi = SContext->esi;

	for	(;;)
	{
		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search closest handler error - no personality\n"));
//			return 0xffffffff;
			return 0;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_SEARCH_PHASE, 
						    0,
						    &Context,
						    PersonalityData);
		if	(Res == _URC_HANDLER_FOUND)
			break;
		if	(Res != _URC_CONTINUE_UNWIND)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search closest handler error - strange personality return: %d\n", Res));
//			return 0xffffffff;
			return 0;
		}
	}
//	return Context.Esp;
	return (Context.Esp < GetFSChain()) ? 1 : 0;
#endif // GEN_IA64
}
#endif // __linux__

UNWIND_PUBLIC _Unwind_Reason_Code UNWIND_CC _BorUnwind_RaiseException(
	_Unwind_Exception *exception_object)
{
	__personality_routine	Personality;
	void *					PersonalityData;
	_Unwind_Context			Context;
	_Unwind_Context			SaveContext;
	_Unwind_RegVal			CatchIP;
	_Unwind_Reason_Code		Res;

	GetContext(&Context);
	SaveContext = Context;

	DEBUG_PRINT("unwind.control.phases", ("RaiseException\n"));
	DEBUG_PRINT("unwind.control.phases", ("Search Phase:\n"));

	for	(;;)
	{
		CatchIP = _BorUnwind_GetIP(&Context);
		Personality = FindPersonality(&Context, CatchIP, &PersonalityData, 0);
		if	(!Personality)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search fatal phase 1 error - no personality\n"));
			return _URC_FATAL_PHASE1_ERROR;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_SEARCH_PHASE, 
						    exception_object,
						    &Context,
						    PersonalityData);
		if	(Res == _URC_HANDLER_FOUND)
			break;
		if	(Res != _URC_CONTINUE_UNWIND)
		{
			DEBUG_PRINT("unwind.control.phases", (" Search fatal phase 1 error - strange personality return\n"));
			return _URC_FATAL_PHASE1_ERROR;
		}
	}

	/*
		Found a handler.  Restart, telling frames to clean themselves up.
	*/
	
	DEBUG_PRINT("unwind.control.phases", ("Cleanup Phase:\n"));
	while	(_BorUnwind_GetIP(&SaveContext) != CatchIP)
	{
		Personality = FindPersonality(&SaveContext, _BorUnwind_GetIP(&SaveContext), &PersonalityData, 1);
		if	(!Personality)
		{
			return _URC_FATAL_PHASE2_ERROR;
		}
		Res = (Personality)(_UNWIND_VERSION,
						    _UA_CLEANUP_PHASE,
						    exception_object,
						    &SaveContext,
						    PersonalityData);
		if	(Res == _URC_INSTALL_CONTEXT)
		{
			/*
				Hoo boy.  The phase 2 handler wants to be running without a net.
				OK.  Install the context, after saving off some data that we'll
				need on a resume.
			*/
			InstallAndResume(&SaveContext, exception_object, CatchIP, Personality);
			assert(!"How is Resume supposed to work?");
		}
		if	(Res != _URC_CONTINUE_UNWIND)
			return _URC_FATAL_PHASE2_ERROR;

	}
	/*
		We should now be at the frame that wanted the exception.
	*/
	Personality = FindPersonality(&SaveContext, _BorUnwind_GetIP(&SaveContext), &PersonalityData, 1);
	if	(!Personality)
		return _URC_FATAL_PHASE2_ERROR;
	Res = (Personality)(_UNWIND_VERSION,
					    _UA_CLEANUP_PHASE | _UA_HANDLER_FRAME,
					    exception_object,
					    &SaveContext,
					    PersonalityData);
	if	(Res != _URC_INSTALL_CONTEXT)
		return _URC_FATAL_PHASE2_ERROR;

	/*
		We've come to the point where the personality routine for the catch
		frame has run its cleanup phase.  We require it to install a context,
		and continue execution at this point.  This is currently not resumeable.
	*/
#ifndef __GNUC__
#pragma message ("Should this be resumeable?")
#endif
	InstallAndResume(&SaveContext, exception_object, 0, NULL);
	assert(!"Never should get here");
	return _URC_NO_REASON;
}


UNWIND_PUBLIC int _BorUnwind_AddressIsInPCMap(_Unwind_codePtr Addr)
{
	if	(FindPersonality(NULL, Addr, NULL, 0))
	{
		DEBUG_PRINT("unwind.control.phases", ("Address %p is in PC map\n", Addr));
		return 1;
	}
	DEBUG_PRINT("unwind.control.phases", ("Address %p is not in PC map\n", Addr));
	return 0;
}

UNWIND_PUBLIC void * _BorUnwind_GetFuncUnwindInfo(_Unwind_codePtr Addr)
{
	void *res;

	if	(FindPersonality(NULL, Addr, &res, 0))
		return res;
	return 0;
}

UNWIND_PUBLIC void _BorUnwind_SetUnhandledExceptionFilter(void *OutermostProc,
							  void (*handler)(void))
{
    _BorUnwind_UEFOuterProc = (_Unwind_codePtr)OutermostProc;
    _BorUnwind_UnhandledExceptionFilter = handler;
}

_Unwind_codePtr _BorUnwind_GetUEFOuterProc(void)
{
	return _BorUnwind_UEFOuterProc;
}

_BorUnwind_UEFHandler _BorUnwind_GetUnhandledExceptionFilter(void)
{
	return _BorUnwind_UnhandledExceptionFilter;
}

