/* **************************************************************************
 *                                                                          *
 * unwindfi.h                                                               *
 *                                                                          *
 * Copyright (c) 2002 Borland Software Corporation.                         *
 *                                                                          *
 * This file may be distributed and/or modified under the terms of the GNU  *
 * Lesser General Public License (LGPL) version 2.1 as published by the     *
 * Free Software Foundation and appearing at                                *
 * http://www.borland.com/kylix/lgpl.html.                                  *
 *                                                                          *
 * Licensees holding a valid Borland No-Nonsense License for this Software  *
 * may  use this file in accordance with such license, which appears in the *
 * file license.txt that came with this Software.                           *
 *                                                                          *
 * *************************************************************************/

/* unwindfi.h --- Unwind Frame Info.  
	This includes the descriptors for the Delphi PC map,
	plus the frame unwind info and local PC map.
*/

#ifndef	UNWINDFI_H
#define	UNWINDFI_H

#ifdef __cplusplus
extern "C" {
#endif

#define	UFI_F_REG_EBX		0x01	/* EBX is saved */
#define	UFI_F_REG_ESI		0x02	/* ESI is saved */
#define	UFI_F_REG_EDI		0x04	/* EDI is saved */
#define	UFI_F_REG_EBP		0x08	/* EBP is saved */
#define UFI_F_HASFRAME		0x10	/* proc has an EBP frame */
#define UFI_F_HASRANGES		0x20	/* proc has any protected ranges at all */
#define UFI_F_SROFFS		0x40	/* saved regs offset was stored */
#define UFI_F_CALLEEPOP		0x80	/* proc has callee pop amount */
#define UFI_F_NESTEDHANDLER 0x08	/* overloading REG_EBP - means that unwind info is present */

/*
	If you change this structure, you must change ReverseFrameInfo in objout.c!
*/
typedef	struct ExceptionFrameInfo
{
	unsigned char	Flags;				/* UFI_F_* above */
	unsigned int	SavedRegsEBPOffs;
	unsigned int	OffsetToPush;
	unsigned int	RangeCount;
	unsigned int	UnwindInfoSize;
	unsigned int	ProcLen;
	unsigned int	CalleePopAmount;
}	ExceptionFrameInfo;

typedef	struct ExceptionRangeOffsets
{
	unsigned int	Start;
	unsigned int	End;
	int	Handler;
}	ExceptionRangeOffsets;

typedef	struct ExcRangeCPP
{
	unsigned int	Start;
	unsigned int	End;
	unsigned		Context;
	int				DtorCnt;
}	ExcRangeCPP;


typedef	struct	PCMapUnit
{
	long	Start;
	long	End;
	int		ProcCount;
	long	ProcInfoOffs;
}	PCMapUnit;

typedef	struct	PCMapProc
{
	long	Start;
	long	End;
}	PCMapProc;

typedef	struct	PCMapProcCPP
{
	long	Start;
	long	End;
	long	ImplementationData;
}	PCMapProcCPP;

/*
	Frank Borland Delphi Exceptions version 0x00.0x07
*/
#define	PCMAPHEADER_SIGNATURE	0xFBDE0007

typedef	struct	PCMapHeader
{
	unsigned long	Signature;
	long			CodeSegmentSize;
	int				UnitCount;
}	PCMapHeader;

/*
    PCMAPHEADER_SIGNATURE2 means that the UnitCount is interpreted as
    a ProcCount.
*/
#define	PCMAPHEADER_SIGNATURE2	0xFBDE0008

#define	UFI_TOPOFSTACK		0xBE00EF00
#define	UFI_STACK_MARKER	0xBEEFFEED

/*
	This should *not* be public here.  It should be encapsulated in the next
	release by the RTL, in the form of a call out to the unwinder foreign
	hook interfaces.
*/
int _BorUnwind_DelphiForeignUnwindHook(struct _Unwind_Context *	CPUContext);

#ifdef __cplusplus
}
#endif
#endif
