/*-----------------------------------------------------------------------*
 * filename - crtlvcl.cpp
 * Copyright (c) 1997, 2002 Borland Software Corporation
 *
 * function(s)
 *
 *   __CRTL_VCL_Init
 *   __CRTL_VCL_Exit
 *
 *   Both functions are called from the startup code.  Init happens
 *   very early and Exit happens very late.
 *
 *   These functions complete the initialization and are called later
 *   during the process
 *
 *   InitVCL
 *   ExitVCL
 *
 *
 *-----------------------------------------------------------------------*/
/*
 *      C/C++ Run Time Library - Version 10.0
 *
 *      Copyright (c) 1997, 2000, 2002 by Inprise Corporation
 *      All Rights Reserved.
 *
 */

/* 1.1        */

#pragma option -V?-

#ifdef _Windows
#include <windows.h>
#endif
#include <_defs.h>
#ifdef _Windows
#include "delphimm.h"
#endif

#ifdef __linux__
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include "unwind.h"
#define PC_MAPPED_EXCEPTIONS

// link in required libraries
#pragma link "libpthread.so"
#pragma link "libdl.so"
#endif

#ifdef _Windows
#define PACKAGE __declspec(PACKAGE)
#else
#define PACKAGE
#endif

namespace System {
    extern PACKAGE void* MainInstance;
    extern PACKAGE bool      IsMultiThread;

#ifdef __linux__
    extern PACKAGE int ArgCount;
    extern PACKAGE char** ArgValues;
#endif


struct TUnwinder
{
    void *RaiseException;
    void *RegisterIPLookup;
    void *UnregisterIPLookup;
    void *DelphiLookup;
    void *ClosestHandler;
};

    extern void __fastcall initialization(void);
    extern void __fastcall InitUnwinder(void);
    extern void __fastcall SetUnwinder(const TUnwinder &u);
    extern bool __fastcall IsUnwinderSet();
};

namespace Sysinit
{
    extern int TlsIndex;
    extern bool ModuleIsCpp;
    extern void *__fastcall GetThisModuleHandle();
    extern void VclInit(bool isDLL, bool isPkg, int hinst, bool isGui);
    extern void VclExit();
};

extern "C"
{

void _RTLENTRY __CRTL_VCL_Sharemem (void);
#ifdef _Windows
extern unsigned char __isDLL;
extern unsigned char __isVCLPackage;
#else
unsigned char __isDLL;
unsigned char __isVCLPackage;
#endif

void _RTLENTRY __CRTL_VCLLIB_Linkage (void) {}

void _RTLENTRY __CRTL_VCL_Init (void)
{
    System::TUnwinder uw;

    // We must set the VCL copy of hInstance (called MainInstance).
    // The startup code has placed this value into the EDX register for us.

#ifdef _Windows
    // Always set MainInstance in an EXE, and only set it in a DLL if it
    // hasn't been set.  Also, never set it in a PACKAGE.
    unsigned int hinst = _EDX;
    if (!__isVCLPackage)
	if ((!__isDLL) || (__isDLL && !System::MainInstance))
	  System::MainInstance = (HINSTANCE) hinst;
#else
    typedef void *HINSTANCE;
    void *hinst = Sysinit::GetThisModuleHandle();

    if (__isVCLPackage)
	__isDLL = 1;
    else if (void *hexe = dlopen(0, RTLD_LAZY))
    {
	if (hinst != hexe)
	    __isDLL = 1;
	dlclose(hexe);
    }

    // On linux, never set MainInstance in a dll
    if (!__isDLL)
	  System::MainInstance = (HINSTANCE) hinst;
#endif

    // We must tell the Pascal RTL to enforce multithread issues in its
    // memory allocation routines, since we are in the multi-threaded RTL
    // on the C/C++ side.

    System::IsMultiThread = true;

#ifdef __linux__
    extern int __TLS_index;

    // Workaround a linker wart:  both of these symbols will get aliased
    // to the same address, but we need to touch them both to ensure
    // they (and more importantly their corresponding module) get
    // linked. 
    //
    // this should be fixed by now.
    Sysinit::TlsIndex = __TLS_index;
    
#endif

#ifdef __linux__
    // we need to initialize System's copy of argc & argv
    if (!__isDLL)
    {
	extern int _argc;
	extern char ** _argv;
	System::ArgCount = _argc;
	System::ArgValues = _argv;
    }
#endif

#ifdef __linux__
    Sysinit::VclInit(__isDLL, __isVCLPackage, (long) hinst, false);
#endif

    // Now initialize the Pascal RTL
    System::initialization();

#ifdef PC_MAPPED_EXCEPTIONS
    // Set the unwinder in an executable, and in a non-package dll if it
    // hasn't already been set.
    if (!__isDLL || (!__isVCLPackage && !System::IsUnwinderSet()))
    {
	uw.RaiseException = (void *)_BorUnwind_RaiseException;
	uw.RegisterIPLookup = (void *)_BorUnwind_RegisterIPLookup;
	uw.UnregisterIPLookup = (void *)_BorUnwind_UnregisterIPLookup;
	uw.DelphiLookup = (void *)_BorUnwind_DelphiLookup;
	uw.ClosestHandler = 0; // IDE specific thing
	System::SetUnwinder(uw);
    }
#endif

#if defined(_BUILDRTLDLL) && !defined(__WIN32__)
    __CRTL_VCL_Sharemem();
#endif
}

void _RTLENTRY __CRTL_VCL_Exit (void)
{
#ifdef _Windows
// This cannot be preformed now, since it prematurely unloads the memory
// manager DLL.  If the RTLDLL is being used, the RTL cannot free its memory
// without jumping into the non-existant memory manager DLL.  Therefore SYSTEM
// is initialized but never finalized.

//  #if defined(_BUILDRTLDLL)
//    System::Finalization();
//  #endif
#else
    Sysinit::VclExit();
    //? System::Finalization();
#endif
}

//
// These functions start up and shut down VCL at the appropriate times.
//

#ifdef __linux__
#pragma startup __CRTL_VCL_Init 28
#pragma exit  __CRTL_VCL_Exit 31
#else

void __stdcall __InitVCL(void);
void __stdcall __ExitVCL(void);


#pragma startup __InitVCL 28     // Initialization of early VCL routines.
				 // Priority 28 is two eariler than what DCC32
				 // uses for .PAS Units' Initialization
				 // sections (which defaults to 30)

				 // User's can now make a #29 priority startup
				 // routine that will set IsLibrary = false
				 // for statically linked DLLs that use VCL
				 // Packages.

#pragma exit  __ExitVCL 31       // Finalization of VCL routines.
				 // Priority 31 is one eariler than what DCC32
				 // uses for .PAS Units' Finalization sections
				 // (which defaults to 30)
#endif


} // extern "C"

