//  dstring.cpp - support for delphi strings in cpp
//  1.1
//  Copyright (c) 1997, 2002 Borland Software Corporation

#define __DSTRING_INLINE

#include <SysUtils.hpp>
#include <StrHlpr.hpp>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
#include <wchar.h>
#include "dstring.h"

static inline void AtomicIncrement(int *x)
{
#if defined(_Windows)
  #if sizeof(int) != sizeof(LONG)
    #error
  #endif
  InterlockedIncrement((LONG*)x);
#else
  Sysutils::InterlockedIncrement(*x);
#endif
}

namespace System
{
  void  __fastcall AnsiString::ThrowIfOutOfRange(int idx) const
  {
    if (idx < 1 || idx > Length())    // NOTE: AnsiString are 1-based !!
      throw ERangeError(Sysconst_SRangeError);
  }

  void __fastcall AnsiString::InitStr(const char* src, int len)
  {
    Data = 0;
    if (len)
    {
      if (len < 0 || len == INT_MAX)
        len = INT_MAX - 1;
      SetLength(len);
      memcpy(Data, src, len);
    }
  }
  
  __fastcall AnsiString::AnsiString(const char* src)
  {
    InitStr(src, src ? strlen(src) : 0);
  }

  __fastcall AnsiString::AnsiString(const AnsiString &src)
  {
    Data = src.Data;
    if (Data)
    {
      StrRec &r = GetRec();
      AtomicIncrement(&r.refCnt);
    }
  }

  __fastcall AnsiString::AnsiString(const WideString &src)
    : Data(0)
  {
    AnsiFromWide(*this, src);
  }


  __fastcall AnsiString::AnsiString(const char* src, unsigned int len)
  {
    InitStr(src, len);
  }

  __fastcall AnsiString::AnsiString(const System::WideChar* src)
    : Data(0)
  {
    System::WideCharToStrVar(const_cast<System::WideChar*>(src), *this);
  }

#if !defined(WIDECHAR_IS_WCHAR)
  __fastcall AnsiString::AnsiString(const wchar_t* src)
    : Data(0)
  {
    WideString ws(src);
    *this = ws;
  }
#endif


  __fastcall AnsiString::AnsiString(char src)
    : Data(0)
  {
    sprintf("%c", src);
  }

  __fastcall AnsiString::AnsiString(short src)
    : Data(0)
  {
    sprintf("%hd", src);
  }

  __fastcall AnsiString::AnsiString(unsigned short src)
    : Data(0)
  {
    sprintf("%hu", src);
  }

  __fastcall AnsiString::AnsiString(int src)
    : Data(0)
  {
    sprintf("%i", src);
  }

  __fastcall AnsiString::AnsiString(unsigned int src)
    : Data(0)
  {
    sprintf("%u", src);
  }

  __fastcall AnsiString::AnsiString(long src)
    : Data(0)
  {
    sprintf("%ld", src);
  }

  __fastcall AnsiString::AnsiString(unsigned long src)
    : Data(0)
  {
    sprintf("%lu", src);
  }

  __fastcall AnsiString::AnsiString(__int64 src)
    : Data(0)
  {
    sprintf("%Li", src);
  }

  __fastcall AnsiString::AnsiString(unsigned __int64 src)
    : Data(0)
  {
    sprintf("%Lu", src);
  }

  __fastcall AnsiString::AnsiString(double src)
    : Data(0)
  {
    *this = Sysutils::FloatToStr(src);
  }

  __fastcall AnsiString::~AnsiString()
  {
    if (Data)
      Strhlpr::AnsiFree(*this);
  }

  AnsiString& __fastcall AnsiString::operator=(const char* src)
  {
    InitStr(src, src ? strlen(src) : 0);
    return *this;
  }

  AnsiString& __fastcall AnsiString::operator=(const AnsiString& src)
  {
    if (this != &src)
      Strhlpr::AnsiAssign(*this, const_cast<AnsiString&>(src));
    return *this;
  }

  AnsiString& __fastcall AnsiString::operator=(const WideString& src)
  {
    AnsiFromWide(*this, src);
    return *this;
  }

  AnsiString& __fastcall AnsiString::operator+=(const WideString& src)
  {
    return *this += AnsiString(src);
  }
  AnsiString& __fastcall AnsiString::operator+=(const AnsiString& src)
  {
    AnsiAppend(*this, src);
    return *this;
  }

  AnsiString __fastcall operator+(const AnsiString& lhs, const AnsiString& rhs)
  {
    return AnsiCat(lhs, rhs);
  }


  int __fastcall AnsiString::AnsiCompare(const AnsiString& rhs) const
  {
    return Sysutils::AnsiCompareStr(*this, rhs);
  }

  int __fastcall AnsiString::AnsiCompareIC(const AnsiString& rhs) const
  {
    return Sysutils::AnsiCompareText(*this, rhs);
  }
  
  bool __fastcall operator==(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) == 0;
  }
  
  bool __fastcall operator!=(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) != 0;
  }
  
  bool __fastcall operator<(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) < 0;
  }
  
  bool __fastcall operator>(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) > 0;
  }
  
  bool __fastcall operator<=(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) <= 0;
  }
  
  bool __fastcall operator>=(const AnsiString& lhs, const AnsiString& rhs)
  {
    return Sysutils::CompareStr(lhs, rhs) >= 0;
  }


  // Return AnsiString of count chars of value ch
  AnsiString __fastcall AnsiString::StringOfChar(char ch, int count)
  {
    AnsiString tmp;
    tmp.SetLength(count);
    memset(tmp.Data, ch, count);
    return tmp;
  }

  AnsiString __fastcall AnsiString::LoadStr(int ident)
  {
    return Sysutils::LoadStr(ident);
  }

#ifdef _Windows
  AnsiString& __fastcall AnsiString::LoadString(HINSTANCE hInstance, int id)
  {
    HRSRC resHdl = ::FindResource(hInstance, MAKEINTRESOURCE(id/16+1), RT_STRING);
    if (resHdl)
    {
      HGLOBAL gblHdl = ::LoadResource(hInstance, resHdl);
      if (gblHdl)
      {
        WCHAR* resData = (WCHAR*)::LockResource(gblHdl);
        if (resData)
        {
          unsigned int len;
          for (int cnt = id % 16; len = *resData++, cnt--; resData += len)
            ;
          if (len != 0)
          {
            SetLength(len);
            len = ::WideCharToMultiByte(CP_ACP, 0, resData, len, Data, len, 0, 0);
            if (len > 0)
              Data[len] = 0;
          }
          /* Unnecessary in Win32
          ::UnlockResource(gblHdl);
          */
        }
        ::FreeResource(resHdl);
      }
    }

    return *this;
  }

  AnsiString __fastcall AnsiString::LoadStr(HINSTANCE hInstance, int ident)
  {
    AnsiString str;
    str.LoadString(hInstance, ident);
    return str;
  }
#endif


  AnsiString __fastcall AnsiString::FmtLoadStr(int ident, const TVarRec* args, int size)
  {
    return Sysutils::FmtLoadStr(ident, args, size);
  }

  AnsiString __fastcall AnsiString::Format(const AnsiString& format, const TVarRec *args, int size)
  {
    return Sysutils::Format(format, args, size);
  }

  AnsiString __fastcall AnsiString::FormatFloat(const AnsiString& format, const long double& value)
  {
    return Sysutils::FormatFloat(format, value);
  }


#undef vprintf
#undef printf
#undef sprintf

  int __cdecl AnsiString::vprintf(const char* format, va_list paramList)
  {
    int size = vsnprintf(NULL, 0, format, paramList);
    SetLength(size);
    return vsnprintf(Data, size + 1, format, paramList);
  }

  int __cdecl AnsiString::cat_vprintf(const char* format, va_list paramList)
  {
    int size = vsnprintf(NULL, 0, format, paramList);
    int len = Length();
    SetLength(len + size);
    return vsnprintf(Data + len, size + 1, format, paramList);
  }


  int __cdecl AnsiString::printf(const char* format, ...)
  {
    int rc;
    va_list paramList;
    va_start(paramList, format);
    rc = vprintf(format, paramList);
    va_end(paramList);
    return rc;
  }

  AnsiString& __cdecl AnsiString::sprintf(const char* format, ...)
  {
    va_list paramList;
    va_start(paramList, format);
    vprintf(format, paramList);
    va_end(paramList);
    return *this;
  }


  int __cdecl AnsiString::cat_printf(const char* format, ...)
  {
    int rc;
    va_list paramList;
    va_start(paramList, format);
    rc = cat_vprintf(format, paramList);
    va_end(paramList);
    return rc;
  }

  AnsiString& __cdecl AnsiString::cat_sprintf(const char* format, ...)
  {
    va_list paramList;
    va_start(paramList, format);
    cat_vprintf(format, paramList);
    va_end(paramList);
    return *this;
  }


  AnsiString __fastcall AnsiString::FloatToStrF(long double value, TStringFloatFormat format, int precision, int digits)
  {
#if 0
    //!JK this code requires TStringFloatFormat and TFloatFormat to be identical
#endif
    return Sysutils::FloatToStrF(value, TFloatFormat(format), precision, digits);
  }

  AnsiString __fastcall AnsiString::IntToHex(int value, int digits)
  {
    return Sysutils::IntToHex(value, digits);
  }

  AnsiString __fastcall AnsiString::CurrToStr(Currency value)
  {
    return ::CurrToStr(value);
  }

  AnsiString __fastcall AnsiString::CurrToStrF(Currency value, TStringFloatFormat format, int digits)
  {
#if 0
    //!JK this code requires TStringFloatFormat and TFloatFormat to be identical
#endif
    return Sysutils::CurrToStrF(value, TFloatFormat(format), digits);
  }

  int __fastcall AnsiString::WideCharBufSize() const
  {
    // return size of buffer required to call System::WideChar()
    return (Length() + 1) * sizeof(System::WideChar);
  }

  System::WideChar* __fastcall AnsiString::WideChar(System::WideChar* dest, int destSize) const
  {
    System::StringToWideChar(*this, dest, destSize);
    return dest;
  }

  AnsiString& __fastcall AnsiString::Unique()
  {
    System::UniqueString(*this);
    return *this;
  }

  AnsiString& __fastcall AnsiString::Insert(const AnsiString& src, int index)
  {
    Strhlpr::AnsiInsert(*this, src, index);
    return *this;
  }

  AnsiString& __fastcall AnsiString::Delete(int index, int count)
  {
    Strhlpr::AnsiDelete(*this, index, count);
    return *this;
  }

  AnsiString& __fastcall AnsiString::SetLength(int newLength)
  {
    Strhlpr::AnsiSetLength(*this, newLength);
    return *this;
  }

  int __fastcall AnsiString::Pos(const AnsiString& subStr) const
  {
    return Strhlpr::AnsiPos(*this, subStr);
  }

  AnsiString __fastcall AnsiString::LowerCase() const
  {
    return Sysutils::AnsiLowerCase(*this);
  }

  AnsiString __fastcall AnsiString::UpperCase() const
  {
    return Sysutils::AnsiUpperCase(*this);
  }

  AnsiString __fastcall AnsiString::Trim() const
  {
    return Sysutils::Trim(*this);
  }

  AnsiString __fastcall AnsiString::TrimLeft() const
  {
    return Sysutils::TrimLeft(*this);
  }

  AnsiString __fastcall AnsiString::TrimRight() const
  {
    return Sysutils::TrimRight(*this);
  }

  AnsiString __fastcall AnsiString::SubString(int index, int count) const
  {
    return Strhlpr::AnsiCopy(*this, index, count);
  }

  int __fastcall AnsiString::ToInt() const
  {
    return Sysutils::StrToInt(*this);
  }

  int __fastcall AnsiString::ToIntDef(int defaultValue) const
  {
    return Sysutils::StrToIntDef(*this, defaultValue);
  }

  double __fastcall AnsiString::ToDouble() const
  {
#if 0
    //!JK change this to return a long double when the compiler properly
    //!JK promotes/demotes long doubles/doubles
#endif
    return Sysutils::StrToFloat(*this);
  }

  bool __fastcall AnsiString::IsDelimiter(const AnsiString& delimiters, int index) const
  {
    return Sysutils::IsDelimiter(delimiters, *this, index);
  }

  int __fastcall AnsiString::AnsiPos(const AnsiString& subStr) const
  {
    return Sysutils::AnsiPos(subStr, *this);
  }

  bool __fastcall AnsiString::IsPathDelimiter(int index) const
  {
    return Sysutils::IsPathDelimiter(*this, index);
  }

  int __fastcall AnsiString::LastDelimiter(const AnsiString& delimiters) const
  {
    return Sysutils::LastDelimiter(delimiters, *this);
  }

  AnsiString::TStringMbcsByteType __fastcall AnsiString::ByteType(int index) const
  {
    //!GCD this code requires TStringMbcsByteType and TMbcsByteType to be identical
    return TStringMbcsByteType(Sysutils::ByteType(*this, index));
  }

  bool __fastcall AnsiString::IsLeadByte(int index) const
  {
    return ByteType(index) == mbLeadByte;
  }

  bool __fastcall AnsiString::IsTrailByte(int index) const
  {
    return ByteType(index) == mbTrailByte;
  }

  char* __fastcall AnsiString::AnsiLastChar() const
  {
    return Sysutils::AnsiLastChar(*this);
  }


} // System Namespace
